//
//  EMTI_Data.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth on 07.09.15.
#include "EMTI_Data.hpp"

vector<string> EMTI_Data::split(string str, char delimiter) {
    vector<string> result;
    
    const char* s = str.c_str();
    
    char space = delimiter;
    char tab = 9;
    char endOfLine = '\n';
    char endOfLine2 = '\r';
    
    do
    {
        const char *begin = s;
        
        while(*s != space && *s != tab && *s != endOfLine && *s != endOfLine2 && *s)
            s++;
        
        if (string(begin, s).size() > 0 ) {
            
            string token = string(begin, s);
            
            result.push_back(token);
            //cout << string(begin, s) << "\n";
        }
    } while (0 != *s++);
    
    return result;
}

int EMTI_Data::random_integer(int start, int end) {
    int range = end - start + 1;
    int number;
    
    if (end>start) {
        number = rand() % range + start;
    }
    else {
        number = start;
    }
    
    return number;
}

EMTI_Data::EMTI_Data() {
    
    attributeValues_int = new vector<vector<int>>();
    dataIntVector = new vector<vector<unsigned>>();
    attributeValues_intToString = new vector<map<int, string>>();
    attributeValues_stringToInt = new vector<map<string, int>>();
    
}

EMTI_Data::EMTI_Data(const char* path, char delimiter, vector<int> ignore, int categoryLocation, vector<attributeType> attributeTypes) {
    
    this->attributeTypes = attributeTypes;
    
    attributeValues_int = new vector<vector<int>>();
    dataIntVector = new vector<vector<unsigned>>();
    attributeValues_intToString = new vector<map<int, string>>();
    attributeValues_stringToInt = new vector<map<string, int>>();
    
    ifstream dataFile(path);
    
    if (dataFile.is_open())
    {
        
        string lineString;
        
        if (getline(dataFile, lineString)) {
            
            
            vector<string> stringVector = split(lineString, delimiter);
            
            dataIntVector->push_back(getIntVector(stringVector, ignore));
            
            attributes = (*dataIntVector)[0].size();
            
            
            
            //dataStringVector.push_back(split(lineString, ','));
            
        }
        int entryCounter = 1;
        while ( getline(dataFile, lineString) ) {
            entryCounter++;
            //dataStringVector.push_back(split(lineString, ','));
            //updateStringAttributeValues(dataStringVector.back());
            
            vector<string> stringVector = split(lineString, delimiter);
            
            vector<unsigned> intVector = getIntVector(stringVector, ignore);
            if (intVector.size() == attributes) {
                dataIntVector->push_back(intVector);
            }
            else {
                cout << "Rejected entry number " << entryCounter << " (Wrong size, " << intVector.size() << " vs " << attributes << ")\n";
            }
        }
        
        if (dataIntVector->size() > 0) {
            
            instances = dataIntVector->size();
        }
        
        dataFile.close();
        
    }
    
    /* put the feature to be classified in the back */
    if (categoryLocation != -1) {
        if (dataIntVector->size() > 0 && categoryLocation < (*dataIntVector)[0].size()-1) {
            
            for (int i = 0; i < dataIntVector->size(); i++) {
                
                (*dataIntVector)[i].push_back((*dataIntVector)[i][categoryLocation]);
                (*dataIntVector)[i].erase( (*dataIntVector)[i].begin()+categoryLocation);
            }
            
            attributeValues_intToString->push_back((*attributeValues_intToString)[categoryLocation]);
            attributeValues_intToString->erase( attributeValues_intToString->begin()+categoryLocation);
            
            attributeValues_stringToInt->push_back((*attributeValues_stringToInt)[categoryLocation]);
            attributeValues_stringToInt->erase( attributeValues_stringToInt->begin()+categoryLocation);
            
            attributeValues_int->push_back((*attributeValues_int)[categoryLocation]);
            attributeValues_int->erase( attributeValues_int->begin()+categoryLocation);
            
        }
    }
    
    if (attributeTypes.size() == 0) {
        for (int i = 0; i < attributeValues_intToString->size(); i++) {
            this->attributeTypes.push_back(categorical);
        }
    }
    
}

EMTI_Data::EMTI_Data(const char* path,
                     vector<vector<int>>* attributeValues_int,
                     vector<map<int, string>>* attributeValues_intToString,
                     vector<map<string, int>>* attributeValues_stringToInt,
                     char delimiter, vector<int> ignore, int categoryLocation, vector<attributeType> attributeTypes)
{
    this->attributeTypes = attributeTypes;
    
    this->attributeValues_int = new vector<vector<int>>(*attributeValues_int);
    dataIntVector = new vector<vector<unsigned>>();
    this->attributeValues_intToString = new vector<map<int, string>>(*attributeValues_intToString);
    this->attributeValues_stringToInt = new vector<map<string, int>>(*attributeValues_stringToInt);
    
    /* flip back potential category flips */
    if (categoryLocation != -1) {
        if (categoryLocation < (this->attributeValues_int->size()-1)) {
            
            this->attributeValues_intToString->insert(this->attributeValues_intToString->begin()+categoryLocation, (this->attributeValues_intToString->back()));
            this->attributeValues_intToString->erase(this->attributeValues_intToString->end()-1);
            
            this->attributeValues_stringToInt->insert(this->attributeValues_stringToInt->begin()+categoryLocation, (this->attributeValues_stringToInt->back()));
            this->attributeValues_stringToInt->erase(this->attributeValues_stringToInt->end()-1);
            
            this->attributeValues_int->insert(this->attributeValues_int->begin()+categoryLocation, (this->attributeValues_int->back()));
            this->attributeValues_int->erase(this->attributeValues_int->end()-1);
            
        }
    }

    
    for (int i = 0; i < this->attributeValues_int->size(); i++) {
        currentHighestAttributes.push_back((int)(*this->attributeValues_int)[i].size()-1);
    }
    
    ifstream dataFile(path);
    
    if (dataFile.is_open())
    {
        
        string lineString;
        
        if (getline(dataFile, lineString)) {
            
            
            vector<string> stringVector = split(lineString, delimiter);
            
            dataIntVector->push_back(getIntVector(stringVector, ignore));
            
            attributes = (*dataIntVector)[0].size();
            
            
            
            //dataStringVector.push_back(split(lineString, ','));
            
        }
        int entryCounter = 1;
        while ( getline(dataFile, lineString) ) {
            entryCounter++;
            //dataStringVector.push_back(split(lineString, ','));
            //updateStringAttributeValues(dataStringVector.back());
            
            vector<string> stringVector = split(lineString, delimiter);
            
            vector<unsigned> intVector = getIntVector(stringVector, ignore);
            if (intVector.size() == attributes) {
                dataIntVector->push_back(intVector);
            }
            else {
                cout << "Rejected entry number " << entryCounter << " (Wrong size, " << intVector.size() << " vs " << attributes << ")\n";
            }
        }
        
        if (dataIntVector->size() > 0) {
            
            instances = dataIntVector->size();
        }
        
        dataFile.close();
        
    }
    
    /* put the feature to be classified in the back */
    if (categoryLocation != -1) {
        if (dataIntVector->size() > 0 && categoryLocation < (*dataIntVector)[0].size()-1) {
            
            for (int i = 0; i < dataIntVector->size(); i++) {
                
                (*dataIntVector)[i].push_back((*dataIntVector)[i][categoryLocation]);
                (*dataIntVector)[i].erase( (*dataIntVector)[i].begin()+categoryLocation);
            }
            
            this->attributeValues_intToString->push_back((*this->attributeValues_intToString)[categoryLocation]);
            this->attributeValues_intToString->erase(this->attributeValues_intToString->begin()+categoryLocation);
            
            this->attributeValues_stringToInt->push_back((*this->attributeValues_stringToInt)[categoryLocation]);
            this->attributeValues_stringToInt->erase(this->attributeValues_stringToInt->begin()+categoryLocation);
            
            this->attributeValues_int->push_back((*this->attributeValues_int)[categoryLocation]);
            this->attributeValues_int->erase(this->attributeValues_int->begin()+categoryLocation);
            
        }
    }

    
    if (attributeTypes.size() == 0) {
        for (int i = 0; i < attributeValues_intToString->size(); i++) {
            this->attributeTypes.push_back(categorical);
        }
    }
    
    
}

EMTI_Data::~EMTI_Data() {
    attributeValues_int->clear();
    delete attributeValues_int;
    
    dataIntVector->clear();
    delete dataIntVector;
    
    attributeValues_intToString->clear();
    delete attributeValues_intToString;
    
    attributeValues_stringToInt->clear();
    delete attributeValues_stringToInt;
}

void EMTI_Data::saveToFile(string filePath) {
    
    string output = "/Users/avividi/Dropbox/fordypningsprosjekt_Anders_Fuglseth_2015/formatedData/formatedData.data.txt";
    //string output = filePath;
    std::ofstream myfile;
    myfile.open (output);
    myfile.close();
    
    std::ofstream file;
    file.open (output);
    file.close();
    
    file.open(output, std::ios_base::app);
    
    for (vector<unsigned> entry : *dataIntVector) {
        
        if (entry.size() > 0) {
            file << (*attributeValues_intToString)[0][entry[0]];
            for (int i = 1; i < entry.size(); i++) {
                file << ',' << (*attributeValues_intToString)[i][entry[i]];
            }
            file << "\n";
        }
    }
    file.close();
    
}

vector<vector<unsigned>>* EMTI_Data::getNewTrainingSubset(double start, double end) {
    
    vector<vector<unsigned>>* result;
    
    if (start < 1 && start >= 0 && end <= 1 && end > 0) {
        
        int startIndex = dataIntVector->size() * start;
        int endIndex = dataIntVector->size() * end;
        
        vector<vector<unsigned>>::const_iterator first = dataIntVector->begin() + startIndex;
        vector<vector<unsigned>>::const_iterator last = dataIntVector->begin() + endIndex;
        
        result = new vector<vector<unsigned>>(first, last);
        
    }
    else {
        result = new vector<vector<unsigned>>();
    }
    
    return result;
}

vector<vector<unsigned>>* EMTI_Data::getNewTrainingSubset(int start, int end) {
    
    vector<vector<unsigned>>* result;
    
    if (start < dataIntVector->size() && start >= 0 && end < dataIntVector->size() && end > 0) {
        
        vector<vector<unsigned>>::const_iterator first = dataIntVector->begin() + start;
        vector<vector<unsigned>>::const_iterator last = dataIntVector->begin() + end;
        
        result = new vector<vector<unsigned>>(first, last);
        
    }
    else {
        result = new vector<vector<unsigned>>();
    }
    
    return result;
}

vector<vector<unsigned>>* EMTI_Data::getNewTestSubset(double start, double end) {
    
    vector<vector<unsigned>>* result;
    
    if (start < 1 && start >= 0 && end <= 1 && end > 0) {
        
        int startIndex = dataIntVector->size() * start;
        int endIndex = dataIntVector->size() * end;
        
        vector<vector<unsigned>>::const_iterator first = dataIntVector->begin() + startIndex;
        vector<vector<unsigned>>::const_iterator last = dataIntVector->begin() + endIndex;
        
        result = new vector<vector<unsigned>>(first, last);
        
        for (int i = 0; i < result->size(); i++) { //remove category attribute
            (*result)[i].pop_back();
        }
        
    }
    else {
        result = new vector<vector<unsigned>>();
    }
    
    return result;
}

vector<vector<unsigned>>* EMTI_Data::getNewTestSubset(int start, int end) {
    
    vector<vector<unsigned>>* result;
    
    if (start < dataIntVector->size() && start >= 0 && end < dataIntVector->size() && end > 0) {
        
        vector<vector<unsigned>>::const_iterator first = dataIntVector->begin() + start;
        vector<vector<unsigned>>::const_iterator last = dataIntVector->begin() + end;
        
        result = new vector<vector<unsigned>>(first, last);
        
        for (int i = 0; i < result->size(); i++) { //remove category attribute
            (*result)[i].pop_back();
        }
        
    }
    else {
        result = new vector<vector<unsigned>>();
    }
    
    return result;
}

vector<int>* EMTI_Data::getNewFullAttributeScope () {
    vector<int>* fullAttributeScope = new vector<int>();
    
    for (auto &attribute : *attributeValues_intToString) {
        fullAttributeScope->push_back((int)attribute.size());
    }
    
    return fullAttributeScope;
}

vector<unsigned> EMTI_Data::getIntVector(vector<string> stringVector, vector<int> ignore) {
    vector<unsigned> intVector;
    bool noIgnore;
    int ignoreCount;
    
    ignoreCount = 0;
    for (int i = 0; i < stringVector.size(); i++) {
        noIgnore = true;
        for (int n = 0; n < ignore.size(); n++) {
            if (i == ignore[n]) {
                noIgnore = false;
                ignoreCount++;
            }
        }
        if (noIgnore) {
            intVector.push_back(getIntValueOfStringAttribute(stringVector[i], i-ignoreCount));
        }
        else {
            
        }
    }
    return intVector;
}

int EMTI_Data::getIntValueOfStringAttribute(string s, int index) {
    
    if (attributeValues_stringToInt->size() == index) {
        currentHighestAttributes.push_back(0);
        
        map<string, int> stringToInt;
        stringToInt[s] = 0;
        attributeValues_stringToInt->push_back(stringToInt);
        
        map<int, string> intToString;
        intToString[0] = s;
        attributeValues_intToString->push_back(intToString);
        
        vector<int> ints;
        ints.push_back(0);
        attributeValues_int->push_back(ints);
        
        return 0;
    }
    
    else if ((*attributeValues_stringToInt)[index].count(s) == 0) {
        
        currentHighestAttributes[index]++;
        (*attributeValues_stringToInt)[index][s] = currentHighestAttributes[index];
        
        (*attributeValues_intToString)[index][currentHighestAttributes[index]] = s;
        
        (*attributeValues_int)[index].push_back(currentHighestAttributes[index]);
        
        
        
        
        return currentHighestAttributes[index];
    }
    
    else {
        return (*attributeValues_stringToInt)[index].at(s);
    }
}

void EMTI_Data::randomizeDataOrder() {
    
    /* initialize random seed based on current time*/
    //srand ((int)time(NULL));
    
    vector<vector<unsigned>>* randomizedDataIntVector = new vector<vector<unsigned>>();
    
    while (dataIntVector->size() > 0) {
        
        int randomIndex = random_integer(0, (int)dataIntVector->size()-1);
        
        randomizedDataIntVector->push_back((*dataIntVector)[randomIndex]);
        
        dataIntVector->erase(dataIntVector->begin()+randomIndex);
        
    }
    
    delete dataIntVector;
    dataIntVector = randomizedDataIntVector;
    
}

//void Data::updateStringAttributeValues(vector<string> attributes) {
//
//    for (int i = 0; i < attributes.size(); i++) {
//        attributeStringValues[i].insert(attributes[i]);
//    }
//}

string EMTI_Data::to_string() {
    
    string result = "";
    
    for (int i = 0; i < dataIntVector->size(); i++) {
        for (int j = 0; j < (*dataIntVector)[i].size()-1; j++) {
            result += std::to_string((*dataIntVector)[i][j]) + ",";
        }
        result += std::to_string((*dataIntVector)[i][(*dataIntVector)[i].size()-1]) + "\n";
    }
    
    return result;
}

string EMTI_Data::attributeValues_to_string() {
    
    string result = "";
    
    for (int i = 0; i < attributeValues_intToString->size(); i++) {
        string line = std::to_string(i) + " - ";
        
        for (std::map<int,string>::iterator it = (*attributeValues_intToString)[i].begin(); it!=(*attributeValues_intToString)[i].end(); ++it) {
            line += it->second + ", ";
        }
        result += line.substr(0, line.size()-2) + "\n";
    }
    
    return result;
}