//
//  EMTI_Data.hpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth on 07.09.15.
//
// Class used for parsing, processing and representing data.

#ifndef EMTI_Data_hpp
#define EMTI_Data_hpp

#include <stdio.h>

#include <vector>
#include <map>
#include <iostream>
#include <fstream>

using namespace std;

enum attributeType { /* not used */
    categorical,
    continuous
};

class EMTI_Data {
    
public:
    
    EMTI_Data ();
    /* default constructor */
    
    EMTI_Data (const char* filePath,
               char delimiter = ',',
               vector<int> ignore = vector<int>(), /* which attributes/datapoints should be ignored */
               int categoryLocation = -1, /* which attribute needs to be classified in case it is not the last (default) */
               vector<attributeType> attributeTypes = vector<attributeType>()/* not used */
    );
    /* constructor, reading from file */
    
    EMTI_Data(const char* path,
              vector<vector<int>>* attributeValues_int,
              vector<map<int, string>>* attributeValues_intToString,
              vector<map<string, int>>* attributeValues_stringToInt,
              char delimiter = ',', vector<int> ignore = vector<int>(), int categoryLocation = -1, vector<attributeType> attributeTypes = vector<attributeType>());
    /* constructor, with attributevalues set, reading from file */
    
    ~EMTI_Data();
    
    void saveToFile (string filePath);
    /* save data to file */
    
    string to_string();
    /* get string ouput of all data */
    string attributeValues_to_string();
    /* get string ouput of attribute data */
    
    vector<vector<unsigned>>* getNewTrainingSubset(double start = 0, double end = 1);
    /* get data subset by indices */
    vector<vector<unsigned>>* getNewTrainingSubset(int start = 0, int end = 0);
    /* get data subset by ratio */
    
    vector<vector<unsigned>>* getNewTestSubset(double start = 0, double end = 1);
    /* get data subset by indices without class labels */
    vector<vector<unsigned>>* getNewTestSubset(int start = 0, int end = 0);
    /* get data subset by ratio without class labels */
    
    vector<int>* getNewFullAttributeScope ();
    /* get the (internal numerical representation) list of possible attribute values, including class values */
    
    void randomizeDataOrder();
    /* randomize the order of the data */
    
    
    
    vector<attributeType> attributeTypes;
    /* not used */
    
    vector<vector<unsigned>> *dataIntVector;
    /* where the actual data is stored */
    
    vector<map<int, string>> *attributeValues_intToString;
    /* map from internal numerical representation of attribute values to string (real) representation */
    vector<map<string, int>> *attributeValues_stringToInt;
    /* for convenience, map the other way */
    vector<vector<int>> *attributeValues_int;
    /* the attribute values themselves */
    
    vector<int> currentHighestAttributes;
    /* used when reading data and building attribute maps */
    
    long attributes = 0;
    /* attribute count */
    long instances = 0;
    /* insance count */
    
private:
    vector<string> split(string str, char delimiter = ' ');
    /* split a line of data text */
    
    vector<unsigned> getIntVector(vector<string> stringVector, vector<int> ignore);
    /* get internal representation of a line of data text */
    int getIntValueOfStringAttribute(string s, int index);
    /* get internal representation of an string attribute in a data text */
    
    int random_integer(int start, int end);
    /* get random int */
};

#endif /* EMTI_Data_hpp */
