//
//  EMTI_TreeCreature.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth on 30.10.2015.

#include "EMTI_TreeCreature.hpp"


EMTI_TreeCreature::EMTI_TreeCreature(vector<int>* fullAttributeScope) {
    
    
    this->attributeScope = new vector<int>(*fullAttributeScope);
    
    this->categoryScope = fullAttributeScope->back();
    this->attributeScope->pop_back();
    
}

EMTI_TreeCreature::EMTI_TreeCreature(vector<int>* attributeScope, int categoryScope) {
    
    
    this->attributeScope = new vector<int>(*attributeScope);
    
    this->categoryScope = categoryScope;
    
}


EMTI_TreeCreature::~EMTI_TreeCreature() {
    delete tree;
    delete attributeScope;
}

EMTI_TreeCreature::EMTI_TreeCreature() {
    
}

EMTI_TreeCreature::EMTI_TreeCreature (const EMTI_TreeCreature &b) {
    this->tree = new EMTI_DecisionTree(*b.tree);
    this->attributeScope = new vector<int>(*b.attributeScope);
    this->categoryScope = b.categoryScope;
    this->classificationAccuracy = b.classificationAccuracy;
    this->fitness = b.fitness;
}

void EMTI_TreeCreature::createRandomTree() {
    
    int rootAttributeVal = random_number(0, (int) attributeScope->size()-1);
    
    tree = new EMTI_DecisionTree(rootAttributeVal, (*attributeScope)[rootAttributeVal], categoryScope);
    
}

EMTI_TreeNode* EMTI_TreeCreature::createRandomSubTree(EMTI_TreeNode* parent, int childIndex) {
    
    int branchStartAttributeVal = random_number(0, (int) attributeScope->size()-1);
    
    EMTI_TreeNode* branchStart = new EMTI_TreeNode(parent, parent->depth+1, categoryScope, branchStartAttributeVal, childIndex, false);
    
    for (int i = 0; i < (*attributeScope)[branchStartAttributeVal]; i++) {
        branchStart->children.push_back(new EMTI_TreeNode(branchStart, branchStart->depth+1, categoryScope, -1, i, true));
    }
    
    return branchStart;
}

int EMTI_TreeCreature::mutate (float mutationRate) {
    
    if (mutationRate == 0) return 0;
    int mutatedCount = 0;
    
    
    for (EMTI_TreeIterator i = tree->begin(); i != tree->end(); ++i) {
        
        if (i.node->isEndNode) {
            
        }
        else {
            if (random_number(0, (int)((double)(tree->decisionNodesCount) /mutationRate)) == 0) { //if true, mutation occurs
                
                mutatedCount++;
                
                EMTI_TreeNode* mutateNode = i.node;
                
                int rand1 = random_number(0, (int)attributeScope->size()+1);
                
                if (rand1 == (int)attributeScope->size() && i != tree->begin()) {//CUT BRANCH
                    
                    for (EMTI_TreeNode* child : mutateNode->children) {
                        delete child;
                    }
                    
                    mutateNode->children.clear();
                    
                    mutateNode->isEndNode = true;
                    
                    mutateNode->value = 0;
                    
                    tree->updateNodeCountAndList();//TODO more efficient?
                    
                }
                else if (rand1 == (int)attributeScope->size()+1 && i != tree->begin()) {//IMPLODE / DEPLODE
                    vector<int> decisionChildrenIndices;
                    
                    for (auto& child : mutateNode->children) {
                        if (!child->isEndNode) decisionChildrenIndices.push_back(child->childIndex);
                    }
                    
                    if (decisionChildrenIndices.size() > 0) { //IMPLODE
                        
                        int rand2 = random_number(0, (int)decisionChildrenIndices.size()-1);
                        
                        EMTI_TreeNode* imploder = mutateNode->children[decisionChildrenIndices[rand2]];
                        
                        i.node = imploder;
                        
                        imploder->parent = mutateNode->parent;
                        
                        mutateNode->parent->children[mutateNode->childIndex] = imploder;
                        
                        imploder->childIndex = mutateNode->childIndex;
                        
                        mutateNode->children.erase(mutateNode->children.begin()+decisionChildrenIndices[rand2]);
                        delete mutateNode;
                        
                        mutateNode = imploder;
                        
                        tree->updateNodeCountAndList();//TODO more efficient?
                        
                    }
                    else { //DEPLODE
                        int rand2 = random_number(0, (int)mutateNode->children.size()-1);
                        
                        delete mutateNode->children[rand2];
                        
                        mutateNode->children[rand2] = createRandomSubTree(mutateNode, rand2);
                        
                        tree->updateNodeCountAndList();//TODO more efficient?
                    }
                }
                else { //FLIP
                    rand1 = random_number(0, (int)attributeScope->size()-1);
                    while (rand1 == mutateNode->value && (int)attributeScope->size()-1 > 0) {
                        rand1 = random_number(0, (int)attributeScope->size()-1);
                    }
                    
                    if ((*attributeScope)[rand1] == (*attributeScope)[mutateNode->value]) {
                        
                    }
                    else {
                        bool structuredChanged = false;
                        while ((*attributeScope)[rand1] < mutateNode->children.size()) {
                            
                            int rand2 = random_number(0, (int)mutateNode->children.size()-1);
                            
                            delete mutateNode->children[rand2];
                            
                            mutateNode->children.erase(mutateNode->children.begin() + rand2);
                            
                            structuredChanged = true;
                            
                            
                        }
                        while ((*attributeScope)[rand1] > mutateNode->children.size()) {
                            
                            int rand2 = random_number(0, categoryScope-1);
                            
                            mutateNode->children.push_back(new EMTI_TreeNode(mutateNode, mutateNode->depth+1, categoryScope, rand2, (int)mutateNode->children.size(), true));
                            
                            structuredChanged = true;
                        }
                        if (structuredChanged) tree->updateNodeCountAndList();
                    }
                    mutateNode->value = rand1;
                }
            }
            
        }
        
    }
    return mutatedCount;
}

pair<EMTI_TreeCreature*, EMTI_TreeCreature*> EMTI_TreeCreature::crossover (pair<EMTI_TreeCreature*, EMTI_TreeCreature*> parents, float crossoverRate, vector<int>* fullAttributeScope) {
    
    
    pair <EMTI_TreeCreature*, EMTI_TreeCreature*> children;
    children.first = new EMTI_TreeCreature(fullAttributeScope);
    children.first->tree = new EMTI_DecisionTree(*parents.first->tree);
    
    children.second = new EMTI_TreeCreature(fullAttributeScope);
    children.second->tree = new EMTI_DecisionTree(*parents.second->tree);
    
    if (crossoverRate > random_double(0, 1)) {
        
        int rand1 = random_number(0, parents.first->tree->nodesCount-1);
        EMTI_TreeNode* subtree1 = parents.first->tree->getSubTree(rand1);
        
        int rand2 = random_number(0, parents.second->tree->nodesCount-1);
        EMTI_TreeNode* subtree2 = parents.second->tree->getSubTree(rand2);
        
        children.first->tree->mergeSubTree(subtree2, rand1);
        //children.first->tree->updateNodeCountAndList();
        
        children.second->tree->mergeSubTree(subtree1, rand2);
        //children.second->tree->updateNodeCountAndList();
        
    }
    
    return children;
    
    
    
}

void EMTI_TreeCreature::assignFitness (vector<vector<unsigned>>* trainingSet, int sizeWeight) {
    
    //chrono::high_resolution_clock::time_point t1;
    //chrono::high_resolution_clock::time_point t2;
    
    classificationAccuracy = 0;
    fitness = 0;
    unordered_set<EMTI_TreeNode*> endNodes;
    
    //t1 = chrono::high_resolution_clock::now();
    
    for (auto &entry : *trainingSet) {
        if (entry.size()>0) {
            //        cout << "Entry=(";
            //        for (auto &n : entry) {
            //            cout << n << ",";
            //        }
            //        cout << ") Tree:";
            //        cout << to_string() << " Decided=" << tree->decide(entry) << " vs " << entry[entry.size()-1] << "\n";
            EMTI_TreeNode* endNode = tree->getEndNode(entry);
            
            endNode->endNodeAccuracies[entry[entry.size()-1]]++;
            
            endNodes.insert(endNode);
        }
    }
    
    //t2 = chrono::high_resolution_clock::now();
    //long long duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    //cout << "dd - " << duration << "\n";
    
    
    for (auto &decideNode : endNodes) {
        int bestCategory = 0;
        int best = 0;
        for (int i = 0; i < decideNode->endNodeAccuracies.size(); i++) {
            if (decideNode->endNodeAccuracies[i] > best) {
                best = decideNode->endNodeAccuracies[i];
                bestCategory = i;
            }
        }
        
        decideNode->value = bestCategory;
        classificationAccuracy += best;
        
        /* resetting node traffic so they don't stack up */
        decideNode->clearEndNodeAccuracies();
        
    }
    
    
    fitness = (classificationAccuracy - (float)tree->nodesCount/sizeWeight) / trainingSet->size() - (float)(tree->nodesCount == 1);
    
    if (fitness < 0.0001) fitness = 0.0001;
    
    //cout << "[" << this << "]Fitness asssigned: " << fitness << "\n";
    
}

void EMTI_TreeCreature::setFitness (float fitness) {
    this->fitness = fitness;
}

float EMTI_TreeCreature::getFitness () {
    return fitness;
}

int EMTI_TreeCreature::random_number(int start, int end) {
    int range = end - start + 1;
    int number;
    
    if (end>start) {
        number = rand() % range + start;
    }
    else {
        number = start;
    }
    
    return number;
}

double EMTI_TreeCreature::random_double(double start, double end) {
    double f = (double)rand() / RAND_MAX;
    return start + f * (end - start);
}

string EMTI_TreeCreature::to_string() {
    return "Fitness: " + ::to_string(fitness) + " Accuracy: " + ::to_string(classificationAccuracy) + " Size: " + ::to_string(tree->nodesCount);// + " Tree: " + tree->to_string();
}

void EMTI_TreeCreature::determineEndNodeAccuraciesAndTreeTraffic (vector<vector<unsigned>>* trainingSet) {
    
    /* reset all traffic and EndNodeAccuracies */
    for (EMTI_TreeIterator i = tree->begin(); i != tree->end(); ++i) {
        i.node->traffic = 0;
        i.node->clearEndNodeAccuracies();
    }
    
    //classificationAccuracy = 0;
    fitness = 0;
    
    /* set traffic and EndNodeAccuracies */
    unordered_set<EMTI_TreeNode*> endNodes;
    for (auto &instance : *trainingSet) {
        if (instance.size()>0) {
            
            EMTI_TreeNode* endNode = tree->getEndNode(instance);
            
            if (instance[instance.size()-1] == endNode->value) endNode->endNodeAccuracies[0]++;
            
            //endNode->endNodeAccuracies[entry[entry.size()-1]]++;
            
            endNode->traffic++;
            
            endNodes.insert(endNode);
        }
    }
    
    for (auto &endNode : endNodes) {
        assignAccuracyUpToRoot(endNode);
    }
}

void EMTI_TreeCreature::assignAccuracyUpToRoot (EMTI_TreeNode* n) {
    if (n->isRoot) {
        return;
    }
    else {
        int bestAccuracy = 0;
        for (int i = 0; i < n->endNodeAccuracies.size(); i++) {
            if (n->endNodeAccuracies[i] > bestAccuracy) {
                bestAccuracy = n->endNodeAccuracies[i];
            }
        }
        n->parent->endNodeAccuracies[0] += bestAccuracy;
        n->parent->traffic += n->traffic;
        assignAccuracyUpToRoot(n->parent);
    }
}





