//
//  EMTI_TreePruner.hpp
//  EMTI_0.2_swift
//
//  Created by Anders Nikolai Fuglseth on 18.04.2016.
//

#ifndef EMTI_TreePruner_hpp
#define EMTI_TreePruner_hpp

#include <vector>
#include <unordered_set>
#include <stdio.h>

#include "EMTI_DecisionTree.hpp"

using namespace std;

class EMTI_TreePruner {
public:
    
    static EMTI_DecisionTree* error_complexity_prune (EMTI_DecisionTree* tree, vector<vector<unsigned>>* trainingSet, vector<vector<unsigned>>* validationSet);
    /*Error-Complexity Pruning (Err-comp) Breiman et al. [1984]*/
    
    static EMTI_DecisionTree* minimum_error_prune (EMTI_DecisionTree* tree, vector<vector<unsigned>>* trainingSet);
    /*Minimum-Error Pruning (Min-err) Niblett and Bratko [1986]*/
    
private:
    
    static vector<EMTI_DecisionTree*> getPrunedCandidates(EMTI_DecisionTree* tree, vector<vector<unsigned>>* trainingSet);
    
    static EMTI_DecisionTree* getNewNextCandidate (EMTI_DecisionTree* tree, vector<vector<unsigned>>* trainingSet);
    
    static void registerEntryUpToRoot (EMTI_TreeNode* n, unsigned value);
    
    static double getSubTreeLeaves_errorCost (EMTI_TreeNode* n, long trainingSetSize);
    
    static int getSubTreeLeaveCount(EMTI_TreeNode* n);
    
    
    static double do_prune_min_err (EMTI_TreeNode* n);
    
    static double expected_error_rate (int categoryCount, int nodeTrafficCount, int bestCategoryCount);
    
    
    
};

#endif /* EMTI_TreePruner_hpp */
