//
//  EMTI_autoRun.cpp
//  EMTI_0.2_swift
//
//  Created by Anders Nikolai Fuglseth on 25.04.2016.
//

#include "EMTI_autoRun.hpp"

EMTI_DecisionTree* EMTI_autoRun::autoEMTI (vector<vector<unsigned>>* trainingSet, vector<int>* fullAttributeScope, bool postStabilization_exploration) {
    
    const int populationSize = 100;
    const float mutationRate = 1.0;
    const float crossoverRate = 1.0;
    const int parentSelection = 2;
    const int adultSelection = 2;
    
    const int populationSizeExploration = 150;
    const float mutationRateExploration = 2.0;
    const int parentSelectionExploration = 0;
    const int adultSelectionExploration = 2;
    
    const int k = 5;
    
    string output = "/Users/avividi/Google Drive/masteroppgave/plotdata/plotData.dat";
    std::ofstream myfile;
    myfile.open (output);
    myfile.close();
    
    EMTI_Population* p = new EMTI_Population(
                                             fullAttributeScope, //vector<int>* fullAttributeScope,
                                             trainingSet, //vector<vector<int>* trainingSet
                                             0, //int maxTreeSize (< 0 for unlimited)
                                             populationSize, //int populationSize
                                             1000, //int treeSize_fitnessWeight
                                             mutationRate, //float mutationRate
                                             crossoverRate, //float crossOverRateParameter
                                             false, //bool rejectWorseGeneration
                                             parentSelection, //int parentSelectionType
                                             /*
                                              0 = parent_selection_fitness_proportionate
                                              1 = parent_selection_tournament
                                              2 = parent_selection_sigma_scaling
                                              */
                                             adultSelection, //int adultSelectionType
                                             /*
                                              0 = adult_selection_full_generational_replacement
                                              1 = adult_selection_overproduction
                                              2 = adult_selection_generational_mixing
                                              3 = adult_selection_add_spice(float spiceRatio)
                                              */
                                             false, //bool verbose
                                             false, //bool verboseplus
                                             true, //bool writeplotdata
                                             output //std::string plotOutputLocation
                                             //int tournamentSizeParameter (only for tournament parent selection)
                                             //float epsilonParameter (only for tournament parent selection)
                                             //float spiceRatio (only for adult_selection_add_spice)
                                             );
    
    
    int stabilizationCounter = stabilizationLimit;
    
    float lastFitness = 0;
startLabel:
    p->populationSize = populationSize;
    p->mutationRateParameter = mutationRate;
    p->rejectWorseGeneration = true;
    p->adultSelectionType = adultSelection;
    p->parentSelectionType = parentSelection;
    while (stabilizationCounter > 0) {
        p->do_evolution_cycle();
        
        if (p->adultBestIndividual->getFitness() > lastFitness) {
            stabilizationCounter = stabilizationLimit;
            lastFitness = p->adultBestIndividual->getFitness();
        }
        else {
            stabilizationCounter--;
            
            //cout << "  evolution stable for " << convergeLimit - convergeCounter << " generations.\n";
        }
        
    }
    
    if (postStabilization_exploration) {
        p->populationSize = populationSizeExploration;
        p->mutationRateParameter = mutationRateExploration;
        p->rejectWorseGeneration = true;
        p->adultSelectionType = adultSelectionExploration;
        p->parentSelectionType = parentSelectionExploration;
        
        EMTI_autoRun::explorationMode_adjustPopulation(p);
        
        stabilizationCounter = stabilizationLimit;
        
        while (stabilizationCounter > 0) {
            p->do_evolution_cycle();
            
            if (p->adultBestIndividual->getFitness() > lastFitness) {
                stabilizationCounter = stabilizationLimit;
                lastFitness = p->adultBestIndividual->getFitness();
                goto startLabel;
            }
            else {
                stabilizationCounter--;
                
                //cout << "  evolution stable for " << convergeLimit - convergeCounter << " generations.\n";
            }
        }
    }
    
    //cout << " cycles = " << p->cycleCount << " ";
    
    p->doPostCycleEndNodeOptimizations(k, false);
    
    EMTI_DecisionTree* tree = new EMTI_DecisionTree (*p->adultBestIndividual->tree);
    
    
    delete p;
    
    return tree;
    
}

void EMTI_autoRun::explorationMode_adjustPopulation (EMTI_Population* p) {
    for (EMTI_TreeCreature* child : *p->children) {
        delete child;
    }
    p->children->clear();
    delete p->children;
    p->children = new vector<EMTI_TreeCreature*> (*p->adults);
    
    p->adults->clear();
    
    std::sort(p->children->begin(), p->children->end(), compareFitness);
    
    for (vector<EMTI_TreeCreature*>::iterator it = p->children->begin()+ 50 ; it != p->children->end(); ++it) {
        delete *it;
    }
    p->children->erase(p->children->begin() + 50, p->children->end());
    
    for (int i = 0; i<50; i++) {
        
        EMTI_TreeCreature* ind = new EMTI_TreeCreature(p->fullAttributeScope);
        
        vector<vector<unsigned>>* ID3subSet = getNewRandomSubset(p->trainingSet, 0.85);
        EMTI_ID3Classifier* classifier = new EMTI_ID3Classifier(ID3subSet, p->fullAttributeScope);
        classifier->makeTree();
        
        ind->tree = new EMTI_DecisionTree (*classifier->tree);
        
        delete ID3subSet;
        delete classifier;
        
        ind->assignFitness(p->trainingSet, p->treeSize_fitnessWeight);
        
        p->children->push_back(ind);
        
        p->childrenBestIndividual = p->adultBestIndividual;
    }
    
    for (int i = 0; i<50; i++) {
        
        EMTI_TreeCreature* ind = new EMTI_TreeCreature(p->fullAttributeScope);
        
        ind->createRandomTree();
        
        ind->assignFitness(p->trainingSet, p->treeSize_fitnessWeight);
        
        p->children->push_back(ind);
        
        p->childrenBestIndividual = p->adultBestIndividual;
    }
    
    //    for (int i = 0; i<50; i++) {
    //
    //        EMTI_TreeCreature* ind = new EMTI_TreeCreature(p->fullAttributeScope);
    //
    //        ind->tree = EMTI_TreePruner::minimum_error_prune((*p->children)[i]->tree, p->trainingSet);
    //
    //        ind->assignFitness(p->trainingSet, p->treeSize_fitnessWeight);
    //
    //        p->children->push_back(ind);
    //
    //        p->childrenBestIndividual = p->adultBestIndividual;
    //    }
    
}

vector<vector<unsigned>>* EMTI_autoRun::getNewRandomSubset(vector<vector<unsigned>>* set, float size) {
    
    /* initialize random seed based on current time*/
    //srand ((int)time(NULL));
    
    vector<vector<unsigned>>* randomizedSubSet = new vector<vector<unsigned>>();
    
    while (randomizedSubSet->size() < set->size() * size) {
        
        int randomIndex = random_integer(0, (int)set->size()-1);
        
        randomizedSubSet->push_back((*set)[randomIndex]);
        
    }
    return randomizedSubSet;
    
}

int EMTI_autoRun::random_integer(int start, int end) {
    int range = end - start + 1;
    int number;
    
    if (end>start) {
        number = rand() % range + start;
    }
    else {
        number = start;
    }
    
    return number;
}


bool EMTI_autoRun::compareFitness (EMTI_TreeCreature* a, EMTI_TreeCreature* b) {
    return a->getFitness() > b->getFitness();
}



