//
//  EMTI_autoRun.hpp
//  EMTI_0.2_swift
//
//  Created by Anders Nikolai Fuglseth on 25.04.2016.
//

/* autoRun, the easiest way to use EMTI. 
 The training data needs to be abstractized to a vector of vectors of unsigned ints. 
 EMTI also needs the attribute scope; how many possible categories there are for each attribute, including the class (hence fullAttributeScope).
 postStabilization_exploration is a mode where the algorithm tries to expand the search space after reaching stabilization, by increasing mutation rate and changing selection methods. Probably not very effective. */
 
 

#ifndef EMTI_autoRun_hpp
#define EMTI_autoRun_hpp

#include <stdio.h>
#include "EMTI_Population.hpp"
#include "EMTI_ID3Classifier.hpp"
#include "EMTI_TreePruner.hpp"

class EMTI_autoRun {
public:
    static EMTI_DecisionTree* autoEMTI (vector<vector<unsigned>>* trainingSet, vector<int>* fullAttributeScope, bool postStabilization_exploration = false);
    
private:
    
    static void explorationMode_adjustPopulation (EMTI_Population*);
    
    static int random_integer(int start, int end);
    
    static vector<vector<unsigned>>* getNewRandomSubset(vector<vector<unsigned>>* set, float size);
    
    static bool compareFitness (EMTI_TreeCreature* a, EMTI_TreeCreature* b);
    
    static const int stabilizationLimit = 50;
};

#endif /* EMTI_autoRun_hpp */
