//
//  Agent.cpp
//  EMTI_Flatland
//
//  Created by aviVidi on 21.04.16.
//  Copyright (c) 2016 Anders Fuglseth. All rights reserved.
//


#include "Agent.h"

Agent::Agent(int numInputs, int numOutputs) {
    
    this->numInputs = numInputs;
    this->numOutputs = numOutputs;
    
    
    //std::cout << net->toString() << "\n";
    
}


void Agent::buildAI(GameBoard* board, bool dynamic) {
    
    string output = "/Users/avividi/Google Drive/masteroppgave/plotdata/plotData.dat";
    
    //create and clear output file
    std::ofstream myfile;
    myfile.open (output);
    myfile.close();
    
    
    vector<int>* fullAttributeScope = new vector<int>({3, 3, 3, 4});
    
    vector<vector<unsigned>>* trainSet = new vector<vector<unsigned>>({{0, 0, 0, 0, 0, 0}});
    
    const int populationSize = 100;
    const int cycles = 500;
    const float mutationRate = 1.0;
    const float crossoverRate = 1.0;
    const int parentSelection = 2;
    const int adultSelection = 2;
    const float spiceRatio = 0.1;
    
    p = new EMTI_Population(
                                     fullAttributeScope, //vector<int>* fullAttributeScope,
                                     trainSet, //vector<vector<unsigned>* trainingSet
                                     0, //int maxTreeSize (<= 0 for unlimited)
                                     populationSize, //int populationSize
                                     1000, //int treeSize_fitnessWeight
                                     mutationRate, //float mutationRate
                                     crossoverRate, //float crossOverRateParameter
                                     false, //bool rejectWorseGeneration
                                     parentSelection, //int parentSelectionType
                                     /*
                                      0 = parent_selection_fitness_proportionate
                                      1 = parent_selection_tournament
                                      2 = parent_selection_sigma_scaling
                                      */
                                     adultSelection, //int adultSelectionType
                                     /*
                                      0 = adult_selection_full_generational_replacement
                                      1 = adult_selection_overproduction
                                      2 = adult_selection_generational_mixing
                                      3 = adult_selection_add_spice(float spiceRatio)
                                      */
                                     false, //bool verbose
                                     false, //bool verboseplus
                                     false, //bool writeplotdata
                                     output, //std::string plotOutputLocation
                                     10, //int tournamentSizeParameter (only for tournament parent selection)
                                     0.7, //float epsilonParameter (only for tournament parent selection)
                                     spiceRatio //float spiceRatio (only for adult_selection_add_spice)
                                     );
    
    
    
    int keepDirection = agentDirection;
    
    std::vector<GameBoard*> testBoards;
    
    int tests = testScenarioCount;
    
    if (!dynamic) {
        tests = 1;
            
            GameBoard* testBoard = new GameBoard (*board);
            
            testBoards.push_back(testBoard);

    }
    
    else {
        for (int i = 0; i < tests; i++) {
            
            GameBoard* testBoard = new GameBoard (*board);
            
            testBoard->nullify();
            
            agentPosition = {randomNumber(0, testBoard->dimensions-1), randomNumber(0, testBoard->dimensions-1)};
            
            testBoard->agentPosition = agentPosition;
            
            testBoard->setPositionValue(agentPosition.first, agentPosition.second, 1);
            
            agentDirection = randomNumber(0, 3);
            
            testBoard->fillRandom();
            
            testBoards.push_back(testBoard);
        }
        
    }

    float currentBestFitness = std::numeric_limits<float>::min();
    for (EMTI_TreeCreature* child : *p->children) {
        
        decisionTree = child->tree;
        
        int individualFitness = 0;
        for (int j = 0; j < tests; j++) {
            individualFitness += evaluationRun(testBoards[j], evaluationRunMoveCount);
        }
        individualFitness = ((individualFitness / tests) * 1000 - child->tree->decisionNodesCount) / 1000;
        child->setFitness(individualFitness);

        
        if (child->getFitness() > currentBestFitness) {
            currentBestFitness = child->getFitness();
            p->childrenBestIndividual = child;
        }
    }
    
    while (p->cycleCount < cycles) {
        p->do_evolution_cycle();
        
//        if (dynamic) {
//            for (GameBoard* board : testBoards) {
//                delete board;
//            }
//            testBoards.clear();
//            for (int i = 0; i < tests; i++) {
//                
//                GameBoard* testBoard = new GameBoard (*board);
//                
//                testBoard->nullify();
//                
//                agentPosition = {randomNumber(0, testBoard->dimensions-1), randomNumber(0, testBoard->dimensions-1)};
//                
//                testBoard->agentPosition = agentPosition;
//                
//                testBoard->setPositionValue(agentPosition.first, agentPosition.second, 1);
//                
//                agentDirection = randomNumber(0, 3);
//                
//                testBoard->fillRandom();
//                
//                testBoards.push_back(testBoard);
//            }
//
//        }
        
        currentBestFitness = std::numeric_limits<float>::min();
        for (EMTI_TreeCreature* child : *p->children) {
            
            decisionTree = child->tree;
            
            float individualFitness = 0;
            for (int j = 0; j < tests; j++) {
                individualFitness += evaluationRun(testBoards[j], evaluationRunMoveCount);
            }
            individualFitness = ((individualFitness / tests) * 1000 - child->tree->decisionNodesCount) / 1000;
            child->setFitness(individualFitness);
            
            if (child->getFitness() > currentBestFitness) {
                currentBestFitness = child->getFitness();
                p->childrenBestIndividual = child;
            }
        }
        
        currentBestFitness = std::numeric_limits<float>::min();
        for (EMTI_TreeCreature* adult : *p->adults) {
            if (adult->getFitness() > currentBestFitness) {
                currentBestFitness = adult->getFitness();
                p->adultBestIndividual = adult;
            }
        }
        

        p->giveGenerationInfoOutput(true, false);
        
    }
    
    for (GameBoard* board : testBoards) {
        delete board;
    }
    
    agentPosition = board->agentPosition;
    agentDirection = keepDirection;
    
    decisionTree = new EMTI_DecisionTree(*p->adultBestIndividual->tree);
    
    cout << "Tree = " << decisionTree->to_string() << "\n";
    
    currentScore = 1000;
    
    delete p;
    
    string dotOutput = "/Users/avividi/Google Drive/masteroppgave/plotdata/trees/treeBig.dot";
    
    vector<map<int, string>> *attributeValues_intToString = new vector<map<int, string>>;
    map<int, string> attrs;
    attrs[0] = "Empty";
    attrs[1] = "Food";
    attrs[2] = "Poison";
    attributeValues_intToString->push_back(attrs);
    attributeValues_intToString->push_back(attrs);
    attributeValues_intToString->push_back(attrs);
    map<int, string> cats;
    cats[0] = "Nothing";
    cats[1] = "Front";
    cats[2] = "Right";
    cats[3] = "Left";
    attributeValues_intToString->push_back(cats);
    
    EMTI_DotTreeWrapper treewrapTrain(decisionTree, attributeValues_intToString);
    
    treewrapTrain.attributeNames = (vector<string>){"Front","Right","Left","Action"};
    
    treewrapTrain.exportToDot(dotOutput);
}

int Agent::evaluationRun(GameBoard* board, int steps) {
    
    int keepDirection = agentDirection;
    
    GameBoard* testBoard = new GameBoard(*board);
    
    currentScore = 1000;
    
    for (int i = 0; i < steps; i++) {
        int action = act(testBoard->getAgentInput(agentDirection));
        if (action == 0) currentScore--;
        agentDirection = testBoard->updateDirection(action, agentDirection);
        currentScore += testBoard->performAction(agentDirection, agentPosition);
        agentPosition = testBoard->agentPosition;
    }
    if (currentScore < 0) currentScore = 0;
    agentPosition = board->agentPosition;
    agentDirection = keepDirection;
    delete testBoard;
    return currentScore;
}

int Agent::act(std::vector<bool>* input) {

    vector<unsigned> entry = {0,0,0};
    
    for (int i = 0; i < 3; i++) {
        if ((*input)[i] == 1) {
            entry[i] = 1;
        }
        if ((*input)[i+3] == 1) {
            entry[i] = 2;
        }
    }
    
    
    currentAction = decisionTree->decide(entry);
    
    delete input;
    
    //cout << "currentAction = " << currentAction << "\n";
    
    return currentAction;
}

int Agent::randomNumber(int start, int end) {
    int range = end - start + 1;
    int number;
    
    if (end>start) {
        number = rand() % range + start;
    }
    else {
        number = start;
    }
    
    return number;
}









