//
//  EMTI_DecisionTree.hpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth on 30.10.2015.

#ifndef EMTI_DecisionTree_hpp
#define EMTI_DecisionTree_hpp

#include <stdio.h>
#include <iostream>

#include <vector>

using namespace std;

struct EMTI_TreeNode {
    
    EMTI_TreeNode();
    
    EMTI_TreeNode(EMTI_TreeNode* parent, int depth, int categoryScope, int value, int childIndex = -1, bool isLeafNode = false);
    
    ~EMTI_TreeNode();
    
    EMTI_TreeNode (const EMTI_TreeNode &b);
    
    void setDepth(int depth);
    
    EMTI_TreeNode* parent;
    
    vector<EMTI_TreeNode*> children;
    
    bool isEndNode = false;
    vector<int> endNodeAccuracies;
    void clearEndNodeAccuracies();
    
    int traffic = 0;
    
    int depth;
    
    int value;
    
    int childIndex;
};

struct EMTI_TreeIterator {
    
    EMTI_TreeIterator(EMTI_TreeNode* node);
    EMTI_TreeIterator operator++();
    bool operator!=(const EMTI_TreeIterator &rhs) const;
    
    EMTI_TreeNode* node;
};

class EMTI_DecisionTree {
public:
    
    EMTI_TreeNode* root;
    
    EMTI_DecisionTree(int rootVal, int childrenCount, int categoryScope);
    
    EMTI_DecisionTree(int categoryScope);
    
    ~EMTI_DecisionTree();
    
    EMTI_DecisionTree (const EMTI_DecisionTree &b);
    
    int decide(vector<unsigned> instance);
    
    EMTI_TreeNode* getEndNode(vector<unsigned> instance, EMTI_TreeNode* node = nullptr);
    
    void mergeSubTree (EMTI_TreeNode* subTreeRoot , int insertPosition);
    
    EMTI_TreeNode* getSubTree (int position);
    
    void changeNodeValue (int value, int position);
    
    int endNodesCount;
    
    int decisionNodesCount;
    
    int nodesCount;
    
    int categoryScope;
    
    string to_string();
    
    void updateNodeCountAndList ();
    
    EMTI_TreeIterator begin();
    EMTI_TreeIterator end();
    
    int random_number(int start, int end);
    
private:
    
    vector<EMTI_TreeNode*> nodeList;
    
    pair<int,int> countNodes_andUpdateTree (EMTI_TreeNode* node = nullptr, int depth = 0);
    
    int countDecisionNodes (EMTI_TreeNode* node = nullptr);
    
    int countEndNodes (EMTI_TreeNode* node = nullptr);
    
    string to_stringRecursive(EMTI_TreeNode* n);
    
};


#endif /* EMTI_DecisionTree_hpp */
