//
//  EMTI_DotTreeWrapper.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth on 04.11.2015.

#include "EMTI_DotTreeWrapper.hpp"

EMTI_DotTreeWrapper::EMTI_DotTreeWrapper(EMTI_DecisionTree* tree, vector<map<int, string>> *attributeValues_intToString) {
    
    this->tree = new EMTI_DecisionTree(*tree);
    
    this->attributeValues_intToString = new vector<map<int, string>>(*attributeValues_intToString);
    
    attributeNames = vector<string>();
    
}

EMTI_DotTreeWrapper::EMTI_DotTreeWrapper(vector<EMTI_TreeCreature*> pop, int generation, string directory) {
    
    
    int i = 0;
    for (EMTI_TreeCreature* t : pop) {
        this->tree = t->tree;
        exportToDotSmall(directory + "gen" + ::to_string(generation) + "individ" + ::to_string(i) + ".dot");
        i++;
    }
    
    this->attributeValues_intToString = new vector<map<int, string>>();
    this->tree = new EMTI_DecisionTree(*pop[0]->tree);
    
}

EMTI_DotTreeWrapper::~EMTI_DotTreeWrapper() {
    delete tree;
    delete attributeValues_intToString;
}

void writeTreeToFile (string path) {
}
EMTI_DecisionTree* readTreeFromFile (string path) {
    std::ofstream file;
    file.open (path);
    file.close();
    
    file.open(path, std::ios_base::app);
    
     return nullptr;
}

void EMTI_DotTreeWrapper::exportToDot(string path) {
    std::ofstream file;
    file.open (path);
    file.close();
    
    file.open(path, std::ios_base::app);
    
    file << "digraph G {\n";
    file << "  graph [pad=0.1, nodesep=0.05, ranksep=1];\n";
    file << "  edge [arrowhead=none];\n";
    file << "  node [margin=0.05 fontcolor=black fontsize=20 width=0.5, shape=plaintext];\n";
    
    for (EMTI_TreeIterator i = tree->begin(); i != tree->end(); ++i) {
        
        if (i.node->isEndNode) {
            file << "    \"" << i.node << "\" [label=\"" << (*attributeValues_intToString)[attributeValues_intToString->size()-1].at(i.node->value)  << "\"]\n";
        }
        else {
            if (i.node->value < attributeNames.size()) {
                file << "    \"" << i.node << "\" [label=\"" << attributeNames[i.node->value] << "\", shape=box]\n";
            }
            else {
                file << "    \"" << i.node << "\" [label=\"" << i.node->value << "\", shape=box]\n";
            }
        }
    }
    
    file << "\n";
    file << getTreeConnectionString(tree->root, 0).str();
    
    
    
    file << "}\n";
    file.close();
    
}

std::stringstream EMTI_DotTreeWrapper::getTreeConnectionString(EMTI_TreeNode* node, int depth) {
    std::stringstream s;
    
    string margin = string ((depth+1) * 4, ' ');
    
    for (EMTI_TreeNode* child : node->children) {
        
        s << margin << "\"" << node << "\" -> \"" << child << "\"";
        s << "[label = \" " << (*attributeValues_intToString)[node->value].at(child->childIndex) << " \"]\n";
        s << getTreeConnectionString(child, depth+1).str();
    }
    return s;
}


void EMTI_DotTreeWrapper::exportToDotSmall(string path) {
    std::ofstream file;
    file.open (path);
    file.close();
    
    file.open(path, std::ios_base::app);
    
    file << "digraph G {\n";
    file << "  graph [pad=0.025, nodesep=0.025, ranksep=0.1];\n";
    file << "  edge [arrowhead=none, style=filled, color=sienna];\n";
    file << "  node [label=\"\", margin=0.05 fontcolor=black fontsize=20 width=0.1, height=0.1 shape=plaintext];\n";
    
    for (EMTI_TreeIterator i = tree->begin(); i != tree->end(); ++i) {
        
        if (i.node->isEndNode) {
            file << "    \"" << i.node << "\"[width=0.05, height=0.05]\n";
        }
        else {
            string s = "cornsilk4";
            if (i.node->value < colorValues.size()) s = colorValues[i.node->value];
            file << "    \"" << i.node << "\" [style=filled, color=\"" << s << "\", shape=circle]\n";
        }
    }
    
    file << "\n";
    file << getTreeConnectionStringSmall(tree->root, 0).str();
    
    
    
    file << "}\n";
    file.close();
    
}

std::stringstream EMTI_DotTreeWrapper::getTreeConnectionStringSmall(EMTI_TreeNode* node, int depth) {
    std::stringstream s;
    
    string margin = string ((depth+1) * 4, ' ');
    
    for (EMTI_TreeNode* child : node->children) {
        
        s << margin << "\"" << node << "\" -> \"" << child << "\"\n";
        s << getTreeConnectionStringSmall(child, depth+1).str();
    }
    return s;
}

void EMTI_DotTreeWrapper::exportToDotTrafficAndAccu(string path) {
    std::ofstream file;
    file.open (path);
    file.close();
    
    file.open(path, std::ios_base::app);
    
    file << "digraph G {\n";
    file << "  graph [pad=0.025, nodesep=0.12, ranksep=0.25];\n";
    file << "  edge [arrowhead=none, style=filled,color=gray37, len = 0.01];\n";
    file << "  node [margin=0.05, width=0.17, height=0.17 shape=plaintext, label=\"\"];\n";
    
    for (EMTI_TreeIterator i = tree->begin(); i != tree->end(); ++i) {
        
        if (i.node->isEndNode) {
            file << "    \"" << i.node << "\"[style=filled, color=" << getAccuracyColor(i.node) <<", colorscheme = rdbu11, shape=diamond, width=0.07, height=0.07]\n";
        }
        else {
            string s = "cornsilk4";
            if (i.node->value < colorValues.size()) s = colorValues[i.node->value];
            file << "    \"" << i.node << "\" [style=filled, color=" << getAccuracyColor(i.node) <<", colorscheme = rdbu11, shape=circle]\n";
        }
    }
    
    file << "\n";
    int max = findMaxTraffic();
    file << getTreeConnectionStringTrafficAndAccu(tree->root, 0, max).str();
    
    
    
    file << "}\n";
    file.close();
    
}

int EMTI_DotTreeWrapper::findMaxTraffic() {
    int max = 0;
    for (EMTI_TreeNode* child : tree->root->children) {
        if (child->traffic > max) {
            max = child->traffic;
        }
    }
    return max;
}

int EMTI_DotTreeWrapper::getPenWidth(int maxTraffic, int traffic) {
    
    if (traffic == 1) return 1;
    if (traffic == 2) return 2;
    
    float relation = (float)traffic / (float)maxTraffic;
    
    if (relation > 7.5/8.0) return 8;
    else if (relation > 6.4/8.0) return 7;
    else if (relation > 4.6/8.0) return 6;
    else if (relation > 3.0/8.0) return 5;
    else if (relation > 1.6/8.0) return 4;
    else if (relation > 0.4/8.0) return 3;
    else return 2;
}

int EMTI_DotTreeWrapper::getAccuracyColor(EMTI_TreeNode* node) {
    
    if (node->traffic == 0) return 6;
    
    int bestEndNodeValue = 0;
    for (int value : node->endNodeAccuracies) {
        if (value > bestEndNodeValue) bestEndNodeValue = value;
    }
    
    float relation = (float)bestEndNodeValue / (float)node->traffic;
    
    if (relation > 9.9999/10.0) return 11;
    else if (relation > 9.0/10.0) return 10;
    else if (relation > 8.0/10.0) return 9;
    else if (relation > 7.0/10.0) return 8;
    else if (relation > 6.0/10.0) return 7;
    else if (relation > 5.0/10.0) return 5;
    else if (relation > 4.0/9.0) return 4;
    else if (relation > 3.0/9.0) return 3;
    else if (relation > 2.0/9.0) return 2;
    else return 1;
}


std::stringstream EMTI_DotTreeWrapper::getTreeConnectionStringTrafficAndAccu(EMTI_TreeNode* node, int depth, int trafficMax) {
    std::stringstream s;
    
    string margin = string ((depth+1) * 4, ' ');
    
    for (EMTI_TreeNode* child : node->children) {
        if (child->traffic > 0) {
            int penWidth = getPenWidth(trafficMax, child->traffic);
            s << margin << "\"" << node << "\" -> \"" << child << "\" [penwidth = " << penWidth << "]\n";
        }
        else {
            s << margin << "\"" << node << "\" -> \"" << child << "\" [penwidth = 1, color = gray86]\n";
        }
        s << getTreeConnectionStringTrafficAndAccu(child, depth+1, trafficMax).str();
    }
    return s;
}
