//
//  EMTI_ID3Classifier.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth on 07.11.2015.
//

#include "EMTI_ID3Classifier.hpp"

EMTI_ID3Classifier::EMTI_ID3Classifier (vector<vector<unsigned>>* trainingSet, vector<int>* fullAttributeScope) {
    this->trainingSet = trainingSet;
    
    this->fullAttributeScope = new vector<int>(*fullAttributeScope);
    
    this->categoryScope = (*this->fullAttributeScope)[this->fullAttributeScope->size()-1];
}

EMTI_ID3Classifier::~EMTI_ID3Classifier () {
    delete fullAttributeScope;
    
    //delete tree
}

void EMTI_ID3Classifier::makeTree() {
    
    tree = new EMTI_DecisionTree(categoryScope);
    
    set<int> attributes;
    for (int i = 0; i < fullAttributeScope->size()-1; i++) {
        attributes.insert(i);
    }
    
    tree->root = ID3Algorithm(attributes, trainingSet);
    
    tree->updateNodeCountAndList();
    
    //cout << " ID3 Size: " << ::to_string(tree->nodesCount) << " Tree: " <<tree->to_string();
    
    
}

EMTI_TreeNode* EMTI_ID3Classifier::ID3Algorithm (set<int> remainingAttributes, vector<vector<unsigned>>* subSet, EMTI_TreeNode* parent) {
    //cout << isPure(subSet) << " " << subSet << "\n";
    
    if (isPure(subSet) || remainingAttributes.size() == 0) {
        
        
        int value = -1;
        
        if (subSet->size() > 0) value = (*subSet)[0][(*subSet)[0].size()-1];
        //cout << "endNode" << value << " depth" << parent->depth+1 << "\n";
        return new EMTI_TreeNode(parent, parent->depth+1, categoryScope, value, (int)parent->children.size(), true);
    }
    
    
    double bestGain = - numeric_limits<double>::max();
    int chooseAttribute = -1;
    for (int i : remainingAttributes) {
        double g = gain(i, subSet);
        if (bestGain <= g) {
            bestGain = g;
            chooseAttribute = i;
        }
    }
    
    remainingAttributes.erase(chooseAttribute);
    
    EMTI_TreeNode* node;
    
    if (parent == NULL) {
        node = new EMTI_TreeNode(NULL, 0, categoryScope, chooseAttribute, -1, false);
        //cout << "root" << chooseAttribute << " depth" << 0 << "\n";
    }
    else {
        node = new EMTI_TreeNode(parent, parent->depth+1, categoryScope, chooseAttribute, (int)parent->children.size(), false);
        //cout << "decideNode" << chooseAttribute << " depth" << parent->depth+1 << "\n";
    }
    
    
    for (int i = 0; i < (*fullAttributeScope)[chooseAttribute]; i++) {
        vector<vector<unsigned>>* subSet2 = newSubset(subSet, chooseAttribute, i);
        node->children.push_back(ID3Algorithm(remainingAttributes,subSet2, node));
        delete subSet2;
    }
    return node;
}

int EMTI_ID3Classifier::classify(vector<unsigned>* entry) {
    return 0;
}

double EMTI_ID3Classifier::gain(int attribute, vector<vector<unsigned>>* subSet) {
    
    int attributeScope = (*fullAttributeScope)[attribute];
    
    double gain = entropy(categoryScope, subSet);
    
    vector<int> categoryDistribution (attributeScope, 0);
    
    for (vector<unsigned> entry : *subSet) {
        categoryDistribution[entry[attribute]]++;
    }
    
    for (int attributeValue = 0; attributeValue < attributeScope; attributeValue++) {
        vector<vector<unsigned>>* subSet2 = newSubset(subSet, attribute, attributeValue);
        //cout << "\n" << (double)categoryDistribution[attributeValue] << " / " << (double)subSet->size() << " * " << entropy(categoryScope, subSet2) <<"\n";
        gain -= ((double)categoryDistribution[attributeValue] / (double)subSet->size()) * entropy(categoryScope, subSet2);
        delete subSet2;
    }
    
    return gain;
}

double EMTI_ID3Classifier::entropy (int categoryScope, vector<vector<unsigned>>* subSet) {
    
    if (subSet->size() == 0) return 0;
    
    vector<int> categoryDistribution (categoryScope, 0);
    
    for (vector<unsigned> entry : *subSet) {
        categoryDistribution[entry[entry.size()-1]]++;
    }
    
    double entropy = 0;
    
    for (int occurences : categoryDistribution) {
        double dist = (double) occurences / (double) subSet->size();
        if (dist != 0) entropy -=  dist * log2(dist);
    }
    return entropy;
}

bool EMTI_ID3Classifier::isPure (vector<vector<unsigned>>* subset) {
    
    if (subset->size() == 0) {
        return true;
    }
    else {
        int category = (*subset)[0][(*subset)[0].size()-1];
        for (int i = 0; i < subset->size(); i++) {
            if (category != (*subset)[i][(*subset)[i].size()-1]) return false;
        }
    }
    
    return true;
}

vector<vector<unsigned>>* EMTI_ID3Classifier::newSubset (vector<vector<unsigned>>* set, int attribute, int attributeValue) {
    vector<vector<unsigned>>* subSet = new vector<vector<unsigned>>();
    
    for (vector<unsigned> entry : *set) {
        if (entry[attribute] == attributeValue) subSet->push_back(entry);
    }
    
    return subSet;
}





