//
//  EMTI_KNearestClassifier.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth on 29.09.15.

#include "EMTI_KNearestClassifier.hpp"

bool compareIntPairs (pair<int,int> a, pair<int,int> b) {
    return a.second > b.second;
}

inline vector<pair<int,int>> removeLowerSimilarityMeasures(vector<pair<int,int>> kNearestNeighbors, vector<vector<unsigned>>* trainingSet) {
    
    vector<pair<int,int>> kNearestNeighborsTrimmed;
    
    int highestSimilarity = 0;
    
    for (int i = 0; i < kNearestNeighbors.size(); i++) {
        if (kNearestNeighbors[i].second > highestSimilarity) {
            kNearestNeighborsTrimmed.clear();
            kNearestNeighborsTrimmed.push_back(kNearestNeighbors[i]);
            highestSimilarity = kNearestNeighbors[i].second;
        }
        else if (kNearestNeighbors[i].second == highestSimilarity) {
            kNearestNeighborsTrimmed.push_back(kNearestNeighbors[i]);
        }
    }
    
    
       return kNearestNeighborsTrimmed;
}


EMTI_KNearestClassifier::EMTI_KNearestClassifier(vector<vector<unsigned>>* trainingSet) {
    
    this->trainingSet = trainingSet;
    
}

int EMTI_KNearestClassifier::classify(vector<unsigned>* entry, int k, bool excludeLowerSimilarityMeasures, bool verbose) {
    
    vector<unsigned>* entryFormatted = new vector<unsigned>(*entry);
    if (entry->size() == (*trainingSet)[0].size()) {
        entryFormatted->pop_back();//bugfix 2016-4-17:Trim entry in case of the category label being present
    }
    
    std::chrono::high_resolution_clock::time_point t1 = std::chrono::high_resolution_clock::now();
     vector<pair<int,int>> kNearestNeighbors = findKNearestNeighbors(entryFormatted, k);
    delete entryFormatted;
    std::chrono::high_resolution_clock::time_point t2 = std::chrono::high_resolution_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::microseconds>( t2 - t1 ).count();
    if (verbose) std::cout << "\nRun time = " << duration << "ms \n";
    
    totalSimilarity = 0;
    if (verbose) cout << k << "-Nearest-Neighbors found: ";
    for (int i = 0; i < kNearestNeighbors.size(); i++) {
        if (verbose) cout << "(index=" << kNearestNeighbors[i].first << "|sim=" << kNearestNeighbors[i].second <<") ";
        totalSimilarity += kNearestNeighbors[i].second;
    }
    if (verbose) cout << " (total=" << totalSimilarity << ")\n";
    
    if (excludeLowerSimilarityMeasures) {
        kNearestNeighbors = removeLowerSimilarityMeasures(kNearestNeighbors, trainingSet);
    }
    
    //VOTING:

    map<int, int> voteMap;
    
    for (int i = 0; i < kNearestNeighbors.size(); i++) {
        voteMap[(*trainingSet)[kNearestNeighbors[i].first][(*trainingSet)[kNearestNeighbors[i].first].size()-1]] = 0;
    }
    for (int i = 0; i < kNearestNeighbors.size(); i++) {
        voteMap[(*trainingSet)[kNearestNeighbors[i].first][(*trainingSet)[kNearestNeighbors[i].first].size()-1]]++;
    }
    
    int highest = -1;
    int category = -1;
    for (std::map<int,int>::iterator it = voteMap.begin(); it != voteMap.end(); ++it) {
        if (it->second > highest) {
            highest = it->second;
            category = it->first;
        }
    }
    
    return category;
}

vector<pair<int,int>> EMTI_KNearestClassifier::findKNearestNeighbors(vector<unsigned>* entry, int k) {
    
    vector<pair<int,int>> kNearestNeighbors;
    
    for (int i = 0; i < trainingSet->size(); i++) {
        
        int sim = compareEntries(i, entry);
        
        //cout << sim << " " << (*trainingSet)[i][(*trainingSet)[i].size()-1] << "\n";
        
        if (kNearestNeighbors.size() == k) {
            
            if (sim > kNearestNeighbors.back().second) {
                kNearestNeighbors.pop_back();
                kNearestNeighbors.push_back(pair<int,int>(i, sim));
                std::sort(kNearestNeighbors.begin(), kNearestNeighbors.end(), compareIntPairs);
            }
        }
        else if (kNearestNeighbors.size() < k) {
            kNearestNeighbors.push_back(pair<int,int>(i, sim));
            std::sort(kNearestNeighbors.begin(), kNearestNeighbors.end(), compareIntPairs);
        }
    }

    return kNearestNeighbors;
}

int EMTI_KNearestClassifier::compareEntries(int trainingIndex, vector<unsigned>* entry) {
    int similarity = 0;
    
    for (int j = 0; j < entry->size(); j++) {
        
        if ((*trainingSet)[trainingIndex][j] == (*entry)[j]) {
            
            similarity++;
            
        }
    }
    
    return similarity;
}