//
//  EMTI_ParentSelection.cpp
//  EMTI0.2_OSX
//
//  Created by aviVidi on 04.02.2016.
//  Copyright © 2016 Anders Nikolai Fuglseth. All rights reserved.
//

#include "EMTI_ParentSelection.hpp"


pair<EMTI_TreeCreature*, EMTI_TreeCreature*> EMTI_ParentSelection::fitness_proportionate (vector<EMTI_TreeCreature*>* adults) {
    
    pair<EMTI_TreeCreature*, EMTI_TreeCreature*> parents;
    
    float totalFitness=0;
    for (EMTI_TreeCreature* adult : *adults) {
        totalFitness += adult->getFitness();
    }
    
    while (parents.second == NULL) {
        
        int winNumber = random_number(0, totalFitness);
        
        float findNumber = 0;
        
        for (EMTI_TreeCreature* adult : *adults) {
            
            if (winNumber < findNumber + adult->getFitness()) {
                
                if (parents.first == NULL) {
                    parents.first = adult;
                }
                else if (parents.first != adult) {
                    parents.second = adult;
                }
                break;
            }
            findNumber += adult->getFitness();
        }
        
    }
    
    return parents;
}

pair<EMTI_TreeCreature*, EMTI_TreeCreature*> EMTI_ParentSelection::tournament (vector<EMTI_TreeCreature*>* adults, int size, double e) {
    
    pair<EMTI_TreeCreature*, EMTI_TreeCreature*> parents;
    
    parents.first = single_tournament(adults, size, e);
    parents.second = single_tournament(adults, size, e);
    
    return  parents;
}

pair<EMTI_TreeCreature*, EMTI_TreeCreature*> EMTI_ParentSelection::sigma_scaling (vector<EMTI_TreeCreature*>* adults, float averageFitness, float sdFitness) {
    
    pair<EMTI_TreeCreature*, EMTI_TreeCreature*> parents;
    
    std::vector<float> scaledFitness;
    for (EMTI_TreeCreature* adult : *adults) {
        float scaled = 1 + ((float)adult->getFitness() - averageFitness) / (2*sdFitness);
        if (scaled > 0) {
            scaledFitness.push_back(1 + ((float)adult->getFitness() - averageFitness) / (2*sdFitness));
        }
        else {
            scaledFitness.push_back(0);
        }
    }
    
    float totalFitness=0;
    for (int i = 0; i < scaledFitness.size(); i++) {
        totalFitness += scaledFitness[i];
    }
    
    while (parents.second == NULL) {
        
        float winNumber = random_double(0, totalFitness);
        
        float findNumber = 0;
        for (int i = 0; i < scaledFitness.size(); i++) {
            if (winNumber < findNumber + scaledFitness[i]) {
                
                if (parents.first == NULL) {
                    parents.first = (*adults)[i];
                }
                else if (parents.first != (*adults)[i]) {
                    parents.second = (*adults)[i];
                }
                break;
            }
            
            findNumber += scaledFitness[i];
        }
    }
    return parents;
}

EMTI_TreeCreature* EMTI_ParentSelection::single_tournament (vector<EMTI_TreeCreature*>* adults, int size, double e) {
    
    EMTI_TreeCreature* initWinnerByPass = new EMTI_TreeCreature();/*doing this basically to stop compile warning*/
    EMTI_TreeCreature* winner = initWinnerByPass;
    
    std::set<EMTI_TreeCreature*> contestants;
    
    int bestFitness = 0;
    
    while (contestants.size() < size && contestants.size() < adults->size()) {
        int randomAdultIndex = random_number(0, (int)adults->size()-1);
        
        contestants.insert((*adults)[randomAdultIndex]);
        
        if ((*adults)[randomAdultIndex]->getFitness() > bestFitness) {
            bestFitness = (*adults)[randomAdultIndex]->getFitness();
            winner = (*adults)[randomAdultIndex];
        }
        
        
    }
    
    if (choice_is_random(e)) {
        return (*adults)[random_number(0, (int)adults->size()-1)];
    }
    
    if (winner != initWinnerByPass) delete initWinnerByPass;
    
    return  winner;
}



bool EMTI_ParentSelection::choice_is_random(double e) {
    
    double r = ((double) rand() / (RAND_MAX));
    
    if (r <= e) {
        return true;
    }
    return false;
}


int EMTI_ParentSelection::random_number(int start, int end) {
    int range = end - start + 1;
    int number;
    
    if (end>start) {
        number = rand() % range + start;
    }
    else {
        number = start;
    }
    
    return number;
}

double EMTI_ParentSelection::random_double(double start, double end) {
    double f = (double)rand() / RAND_MAX;
    return start + f * (end - start);
}