//
//  GameBoard.cpp
//  Solver2048
//
//  Created by Anders Fuglseth on 30.10.14.
//  Copyright (c) 2014 Anders Fuglseth. All rights reserved.
//

#include "GameBoard.h"

GameBoard::GameBoard() {}

GameBoard::~GameBoard() {
    delete[] board;
}

GameBoard::GameBoard (int dimensions, double foodDist, double poisonDist, int poisonEffect, int foodEffect) {
    
    this->dimensions = dimensions;
    
    this->foodEffect = foodEffect;
    this->poisonEffect = poisonEffect;
    
    board = new int[dimensions * dimensions];
    
    int i;
    for (i = 0; i < dimensions * dimensions; i++) {
        board[i] = 0;
    }
    //freeSpaceAmount = dimensions * dimensions - 1;
    
    this->foodDist = foodDist;
    this->poisonDist = poisonDist;
    
}

GameBoard::GameBoard (const GameBoard &b) {
    //freeSpaceAmount = b.freeSpaceAmount;
    dimensions = b.dimensions;
    board = new int[dimensions*dimensions];
    
    foodDist =b.foodDist;
    poisonDist = b.poisonDist;
    
    foodEffect = b.foodEffect;
    poisonEffect = b.poisonEffect;
    
    std::copy(b.board, b.board+dimensions*dimensions, board);
}

int GameBoard::getPositionValue(int positionX, int positionY) {
    
    if (positionX < dimensions && positionY < dimensions) {
        return board[positionX + positionY * dimensions];
    }
    return -1;
    
}

void GameBoard::nullify () {
    int i;
    for (i = 0; i < dimensions * dimensions; i++) {
        board[i] = 0;
    }
}

void GameBoard::setPositionValue(int positionX, int positionY, int value) {
    
    if (positionX < dimensions && positionY < dimensions) {
//        if (value == 0) {
//            freeSpaceAmount++;
//        }
//        
//        if (board[positionX + positionY * dimensions] == 0) {
//            freeSpaceAmount--;
//        }
        
        
        board[positionX + positionY * dimensions] = value;
    }
}

void GameBoard::setFoodAndPoisonDist(double foodDist, double poisonDist) {
    
    this->foodDist = foodDist;
    this->poisonDist = poisonDist;
}

void GameBoard::fillRandom () {
    
    enum boardTypes
    {
        empty,
        agent,
        food,
        poison,
    };
    
    
    for (int i = 0; i < dimensions*dimensions; i++) {
        
        if (fill(foodDist) && board[i] == empty) {
            board[i] = food;
        }
    }
    
    for (int i = 0; i < dimensions*dimensions; i++) {
        
        if (fill(poisonDist) && board[i] == empty) {
            board[i] = poison;
        }
    }
}

bool GameBoard::fill (double chance) {
    
    double r = ((double) rand() / (RAND_MAX));
    
    return chance > r;
}

int GameBoard::updateDirection (int action, int direction) {
    
    int newDirection = direction;
    
    switch (action) {
        case 0:
            return newDirection;
        case 1:
            return newDirection;
            break;
        case 2:
            newDirection++;
            if (newDirection > 3) newDirection = 0;
            return newDirection;
            break;
        case 3:
            newDirection--;
            if (newDirection < 0) newDirection = 3;
            return newDirection;
            break;
        default:
            return newDirection;
            break;
    }
    
}

int GameBoard::performAction (int direction, std::pair<int, int> position) {
    
    enum boardTypes
    {
        empty,
        agent,
        food,
        poison,
    };
    
    this->setPositionValue(position.first, position.second, 0);
    
    std::pair<int, int> newPosition = position;
    
    int effect = 0;
    
    switch (direction) {
        case 0:
            newPosition.second++;
            if (newPosition.second > dimensions-1) newPosition.second = 0;
            break;
        case 1:
            newPosition.first++;
            if (newPosition.first > dimensions-1) newPosition.first = 0;
            break;
        case 2:
            newPosition.second--;
            if (newPosition.second < 0) newPosition.second = dimensions-1;
            break;
        case 3:
            newPosition.first--;
            if (newPosition.first < 0) newPosition.first = dimensions-1;
            break;
        default:
            break;
    }
    if (this->getPositionValue(newPosition.first, newPosition.second) == food) {
        effect = foodEffect;
    }
    else if (this->getPositionValue(newPosition.first, newPosition.second) == poison) {
        effect = poisonEffect;
    }
    
    this->setPositionValue(newPosition.first, newPosition.second, 1);
    
    agentPosition = newPosition;
    
    
    return effect;
}

std::vector<bool>* GameBoard::getAgentInput(int direction) {
    std::pair<int, int> position = agentPosition;
    enum directions
    {
        north,
        east,
        south,
        west,
    };
    
    enum boardTypes
    {
        empty,
        agent,
        food,
        poison,
    };
    
    int xFront = position.first;
    int yFront = position.second;
    int xRight = position.first;
    int yRight = position.second;
    int xLeft = position.first;
    int yLeft = position.second;
    
    
    
    switch (direction) {
        case north:
            yFront+=1;
            xRight+=1;
            xLeft-=1;
            break;
        case east:
            xFront+=1;
            yRight-=1;
            yLeft+=1;
            break;
        case south:
            yFront-=1;
            xRight-=1;
            xLeft-=1;
            break;
        case west:
            xFront-=1;
            yRight+=1;
            yLeft-=1;
            break;
        default:
            break;
    }
    
    /*calculate wrapping: */
    if (xFront == dimensions) xFront = 0;
    else if (xFront < 0) xFront = dimensions -1;
    if (yFront == dimensions) yFront = 0;
    else if (yFront < 0) yFront = dimensions -1;
    if (xRight == dimensions) xRight = 0;
    else if (xRight < 0) xRight = dimensions -1;
    if (yRight == dimensions) yRight = 0;
    else if (yRight < 0) yRight = dimensions -1;
    if (xLeft == dimensions) xLeft = 0;
    else if (xLeft < 0) xLeft = dimensions -1;
    if (yLeft == dimensions) yLeft = 0;
    else if (yLeft < 0) yLeft = dimensions -1;
    
    std::vector<bool>* inputs = new std::vector<bool>();
    
    /* food input */
    if (getPositionValue(xFront, yFront) == food) inputs->push_back(true);
    else inputs->push_back(false);
    if (getPositionValue(xRight, yRight) == food) inputs->push_back(true);
    else inputs->push_back(false);
    if (getPositionValue(xLeft, yLeft) == food) inputs->push_back(true);
    else inputs->push_back(false);
    
    /* poison input */
    if (getPositionValue(xFront, yFront) == poison) inputs->push_back(true);
    else inputs->push_back(false);
    if (getPositionValue(xRight, yRight) == poison) inputs->push_back(true);
    else inputs->push_back(false);
    if (getPositionValue(xLeft, yLeft) == poison) inputs->push_back(true);
    else inputs->push_back(false);
    
    
    return inputs;
}



std::string GameBoard::toString() {
    std::string toString = "";
    
//    int i, j;
//    for (j = dimensions - 1; j >= 0; j--) {
//        toString+="[ ";
//        for (i = 0; i < dimensions; i++) {
//            toString+="[ ";
//            toString+=std::to_string(board[i + j * dimensions]);
//            toString+=" ]";
//        }
//        toString+=" ]\n";
//    }
    
    for (int i = 0; i < dimensions*dimensions; i++) {
        toString += "board[" + std::to_string(i) + "]=" + std::to_string(board[i]) + ";\n";
    }
    
    return toString;
}

void GameBoard::setStaticScenario (int scenario) {
    switch (scenario) {
        case 1:
            board[0]=0;
            board[1]=3;
            board[2]=3;
            board[3]=0;
            board[4]=3;
            board[5]=2;
            board[6]=2;
            board[7]=2;
            board[8]=0;
            board[9]=0;
            board[10]=3;
            board[11]=3;
            board[12]=0;
            board[13]=3;
            board[14]=0;
            board[15]=0;
            board[16]=1;
            board[17]=3;
            board[18]=2;
            board[19]=2;
            board[20]=2;
            board[21]=0;
            board[22]=3;
            board[23]=0;
            board[24]=2;
            board[25]=0;
            board[26]=0;
            board[27]=0;
            board[28]=2;
            board[29]=0;
            board[30]=3;
            board[31]=2;
            board[32]=2;
            board[33]=3;
            board[34]=0;
            board[35]=0;
            board[36]=0;
            board[37]=0;
            board[38]=2;
            board[39]=3;
            board[40]=0;
            board[41]=0;
            board[42]=0;
            board[43]=2;
            board[44]=2;
            board[45]=0;
            board[46]=2;
            board[47]=0;
            board[48]=3;
            board[49]=3;
            board[50]=0;
            board[51]=2;
            board[52]=2;
            board[53]=2;
            board[54]=0;
            board[55]=0;
            board[56]=2;
            board[57]=3;
            board[58]=0;
            board[59]=2;
            board[60]=2;
            board[61]=3;
            board[62]=3;
            board[63]=0;
            board[64]=0;
            board[65]=2;
            board[66]=0;
            board[67]=0;
            board[68]=0;
            board[69]=2;
            board[70]=0;
            board[71]=2;
            board[72]=3;
            board[73]=2;
            board[74]=0;
            board[75]=3;
            board[76]=3;
            board[77]=0;
            board[78]=0;
            board[79]=0;
            board[80]=2;
            board[81]=0;
            board[82]=2;
            board[83]=3;
            board[84]=3;
            board[85]=0;
            board[86]=3;
            board[87]=2;
            board[88]=2;
            board[89]=0;
            board[90]=3;
            board[91]=3;
            board[92]=0;
            board[93]=0;
            board[94]=3;
            board[95]=0;
            board[96]=3;
            board[97]=0;
            board[98]=0;
            board[99]=2;
            break;
        case 2:
            board[0]=0;
            board[1]=2;
            board[2]=2;
            board[3]=2;
            board[4]=3;
            board[5]=0;
            board[6]=2;
            board[7]=3;
            board[8]=3;
            board[9]=0;
            board[10]=0;
            board[11]=0;
            board[12]=3;
            board[13]=0;
            board[14]=0;
            board[15]=3;
            board[16]=2;
            board[17]=2;
            board[18]=0;
            board[19]=2;
            board[20]=0;
            board[21]=2;
            board[22]=0;
            board[23]=2;
            board[24]=2;
            board[25]=0;
            board[26]=0;
            board[27]=0;
            board[28]=3;
            board[29]=0;
            board[30]=2;
            board[31]=2;
            board[32]=2;
            board[33]=0;
            board[34]=0;
            board[35]=2;
            board[36]=0;
            board[37]=0;
            board[38]=2;
            board[39]=2;
            board[40]=0;
            board[41]=0;
            board[42]=2;
            board[43]=1;
            board[44]=2;
            board[45]=3;
            board[46]=0;
            board[47]=3;
            board[48]=0;
            board[49]=2;
            board[50]=2;
            board[51]=2;
            board[52]=3;
            board[53]=3;
            board[54]=0;
            board[55]=2;
            board[56]=3;
            board[57]=2;
            board[58]=3;
            board[59]=3;
            board[60]=0;
            board[61]=0;
            board[62]=3;
            board[63]=0;
            board[64]=0;
            board[65]=0;
            board[66]=0;
            board[67]=0;
            board[68]=2;
            board[69]=2;
            board[70]=0;
            board[71]=2;
            board[72]=0;
            board[73]=0;
            board[74]=3;
            board[75]=2;
            board[76]=3;
            board[77]=2;
            board[78]=0;
            board[79]=3;
            board[80]=2;
            board[81]=0;
            board[82]=3;
            board[83]=0;
            board[84]=2;
            board[85]=0;
            board[86]=2;
            board[87]=2;
            board[88]=3;
            board[89]=0;
            board[90]=3;
            board[91]=3;
            board[92]=0;
            board[93]=0;
            board[94]=2;
            board[95]=2;
            board[96]=0;
            board[97]=2;
            board[98]=0;
            board[99]=3;
            break;
        default:
            break;
    }
    }






