//
//  EMTI_DotTreeWrapper.hpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth on 04.11.2015.
//
// Used for exporting decision tree layout to a dot format (graph description language) for visualization

#ifndef EMTI_DotTreeWrapper_hpp
#define EMTI_DotTreeWrapper_hpp

#include <stdio.h>
#include "EMTI_DecisionTree.hpp"
#include "EMTI_TreeCreature.hpp"

#include <map>
#include <fstream>
#include <sstream>

class EMTI_DotTreeWrapper {
public:
    
    EMTI_DotTreeWrapper(EMTI_DecisionTree* tree, vector<map<int, string>> *attributeValues_intToString);
    
    EMTI_DotTreeWrapper(vector<EMTI_TreeCreature*> pop, int generation, string directory);
    
    ~EMTI_DotTreeWrapper();
    
    EMTI_DecisionTree* tree;
    
    vector<map<int, string>> *attributeValues_intToString;
    
    vector<string> attributeNames;
    
    void writeTreeToFile (string path);
    EMTI_DecisionTree* readTreeFromFile (string path);
    
    void exportToDot (string path);
    void exportToDotSmall(string path);
    
    void exportToDotTrafficAndAccu(string path);
    
    void exportToDotTreelike(string path) ;
    
    const vector<string> colorValues = {"firebrick", "darkorange2", "chocolate3", "darkgoldenrod2", "brown3", "burlywood3", "deeppink4"};
    
private:
    std::stringstream getTreeConnectionString(EMTI_TreeNode* node, int depth);
    std::stringstream getTreeConnectionStringSmall(EMTI_TreeNode* node, int depth);
    std::stringstream getTreeConnectionStringTrafficAndAccu(EMTI_TreeNode* node, int depth, int trafficMax);
    int findMaxTraffic();
    int getPenWidth(int maxTraffic, int traffic);
    int getAccuracyColor(EMTI_TreeNode* node);
};

#endif /* EMTI_DotTreeWrapper_hpp */
