//
//  EMTI_Population.hpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth, 2015.

#ifndef EMTI_Population_hpp
#define EMTI_Population_hpp

#include <stdio.h>
#include <vector>
#include <time.h>
#include <iostream>
#include <fstream>
#include <math.h>
#include <limits>
#include <sstream>

#include <map>
#include <set>

#include "EMTI_TreeCreature.hpp"
#include "EMTI_KNearestClassifier.hpp"

#include "EMTI_ParentSelection.hpp"
#include "EMTI_AdultSelection.hpp"

class EMTI_Population {
public:
    
    EMTI_Population (vector<int>* fullAttributeScope, vector<vector<unsigned>>* trainingSet, int maxTreeSize, int populationSize, int treeSize_fitnessWeight, float mutationRate, float crossOverRateParameter, bool rejectWorseGeneration, int parentSelectionType, int adultSelectionType, bool verbose = false, bool verboseplus = false, bool writePlotData = false, std::string plotOutputLocation = "", int tournamentSizeParameter = 1, float epsilonParameter = 1, float spiceRatio = 0.1);
    
    ~EMTI_Population ();
    
    virtual void do_evolution_cycle ();
    
    void giveGenerationInfoOutput (bool verbose, bool verboseplus, bool writePlotData = false);
    
    void doPostCycleEndNodeOptimizations (int k, bool verbose);
    /* Assigning good values for endNodes not visited in the training set */
    
    vector<EMTI_TreeCreature*>* children;
    
    vector<EMTI_TreeCreature*>* adults;
    
    vector<vector<unsigned>>* trainingSet;//NOT COPIED
    
    vector<vector<unsigned>>* getTrainingSetPart (vector<vector<unsigned>>* trainingSet);
    int subSetSize = 10;
    int currentPart = 0;
    
    vector<int>* fullAttributeScope;
    int maxTreeSize = -1;
    bool maintainMaxTreeSize = false;
    
    int populationSize = 10;
    
    int treeSize_fitnessWeight = 100;
    
    bool rejectWorseGeneration = false;
    
    int parentSelectionType = 0;
    
    int adultSelectionType = 0;
    
    int cycleCount = 0;
    
    int mutationsInCycleCount = 0;
    int crossoversInCycleCount = 0;//NOT USED
    
    EMTI_TreeCreature* childrenBestIndividual;
    EMTI_TreeCreature* adultBestIndividual;
    
    int tournamentSizeParameter;
    
    float mutationRateParameter = 1;
    
    float crossOverRateParameter = 1;
    
    float epsilonParameter;
    
    float spiceRatio;
    
    bool verbose = false;
    bool verboseplus = false;
    
    bool writePlotData = false;
    string plotOutputLocation = "~/plotData.dat";
    
    bool setCurrentGenerationInfoOutput = true;
    string currentGenerationInfoOutput = "";
    
    double sdFitness = 1;
    
    double averageFitness = 0;
    
    double diversity = 0;
    
    double averageSize = 0;
    double sdSize = 1;
    
private:
    
    float measurePopulationDiversity (float sdSize, float sdAccu);
    
    static vector<vector<unsigned>>* getNewRandomSize_RandomTrainingSubset (vector<vector<unsigned>>* trainingSet);
    
    static int random_number (int start, int end);
    static double random_double(double start, double end);
    
    static bool compareFitness (EMTI_TreeCreature* a, EMTI_TreeCreature* b);
    
};

#endif /* EMTI_Data_hpp */
