//
//  EMTI_PopulationES2.cpp
//  EMTI_0.2_swift
//
//  Created by Anders Nikolai Fuglseth on 07.04.2016.
//

#include "EMTI_PopulationES2.hpp"

EMTI_PopulationES2::EMTI_PopulationES2 (vector<int>* fullAttributeScope, vector<vector<unsigned>>* trainingSet, int maxTreeSize, int populationSize, int treeSize_fitnessWeight, float mutationRate, float crossOverRateParameter, bool rejectWorseGeneration, int parentSelectionType, int adultSelectionType, bool verbose, bool verboseplus, bool writePlotData, std::string plotOutputLocation, int tournamentSizeParameter, float epsilonParameter, float spiceRatio)

: EMTI_Population (fullAttributeScope, trainingSet, maxTreeSize, populationSize, treeSize_fitnessWeight, mutationRate, crossOverRateParameter, rejectWorseGeneration, parentSelectionType, adultSelectionType, verbose, verboseplus, writePlotData, plotOutputLocation, tournamentSizeParameter, epsilonParameter, spiceRatio)

{
    
    this->rejectWorseGeneration = true;
    
    this->validationSet = getNewDataSubset(trainingSet, 0.75, 1.0);
    this->trainingSet = getNewDataSubset(trainingSet, 0.0, 0.75);
    
    previousAdults = new vector<EMTI_TreeCreature*>;
    
    //bugfix: reassign fitness for the new trainingset
    for (EMTI_TreeCreature* child : *children) {
        child->assignFitness(this->trainingSet, treeSize_fitnessWeight);
    }
}

EMTI_PopulationES2::~EMTI_PopulationES2 () {
    
    delete trainingSet;
    
    delete validationSet;
}

void EMTI_PopulationES2::do_evolution_cycle ()  {
    
    
    EMTI_Population::do_evolution_cycle();
    
    
    previousValidationSetAccuracy = currentValidationSetAccuracy;
    currentValidationSetAccuracy = calculateValidationSetAccuracy();
    statisticsValidationSetAccuracy = currentValidationSetAccuracy;
    
    //cout << "currentValidationSetAccuracy = " << currentValidationSetAccuracy << " vs previousValidationSetAccuracy =" << previousValidationSetAccuracy << "\n";
    
    if (currentValidationSetAccuracy < previousValidationSetAccuracy) {
        
        rollBackCounter++;
        currentValidationSetAccuracy = previousValidationSetAccuracy;
        //cout << "roll back count = " << rollBackCounter << "\n";
        
        if (rollBackCounter >= rollBackThreshold) {
            
            rollBackCounter = 0;
            rollBackCount++;
            
            //cout << "reject and roll back\n";
            for (EMTI_TreeCreature* adult : *adults) {
                delete adult;
            }
            adults->clear();
            delete adults;
            adults = new vector<EMTI_TreeCreature*>;
            for (EMTI_TreeCreature* creature : *previousAdults) {
                EMTI_TreeCreature* newCreature = new EMTI_TreeCreature(*creature);
                adults->push_back(newCreature);
                
                if (creature == previousAdultBestIndividual) {
                    adultBestIndividual = newCreature;
                }
            }
            childrenBestIndividual->setFitness(-1); //Dirty trick to make generation rejected
            giveGenerationInfoOutput(false, false, false);
        }
    }
    else {
        
        rollBackCounter = 0;
        
        //cout << "accept and continue\n";
        for (EMTI_TreeCreature* adult : *previousAdults) {
            delete adult;
        }
        previousAdults->clear();
        delete previousAdults;
        previousAdults = new vector<EMTI_TreeCreature*>;
        
        for (EMTI_TreeCreature* creature : *adults) {
            EMTI_TreeCreature* newCreature = new EMTI_TreeCreature(*creature);
            previousAdults->push_back(newCreature);
            
            if (creature == adultBestIndividual) {
                previousAdultBestIndividual = newCreature;
            }
        }
    }
    
}

vector<vector<unsigned>>* EMTI_PopulationES2::getNewDataSubset(vector<vector<unsigned>>* dataSet, double start, double end) {
    
    vector<vector<unsigned>>* result;
    
    if (start < 1 && start >= 0 && end <= 1 && end > 0) {
        
        int startIndex = dataSet->size() * start;
        int endIndex = dataSet->size() * end;
        
        vector<vector<unsigned>>::const_iterator first = dataSet->begin() + startIndex;
        vector<vector<unsigned>>::const_iterator last = dataSet->begin() + endIndex;
        
        result = new vector<vector<unsigned>>(first, last);
        
    }
    else {
        result = new vector<vector<unsigned>>();
    }
    
    return result;
}

bool EMTI_PopulationES2::doPostCycle_rollBackCheck() {
    if (rollBackCounter > 0) {
        for (EMTI_TreeCreature* adult : *adults) {
            delete adult;
        }
        adults->clear();
        delete adults;
        adults = new vector<EMTI_TreeCreature*>;
        for (EMTI_TreeCreature* creature : *previousAdults) {
            EMTI_TreeCreature* newCreature = new EMTI_TreeCreature(*creature);
            adults->push_back(newCreature);
            
            if (creature == previousAdultBestIndividual) {
                adultBestIndividual = newCreature;
            }
        }
        return true;
    }
    else return false;
}

double EMTI_PopulationES2::calculateValidationSetAccuracy() {
    
    int right = 0;
    int wrong = 0;
    EMTI_TreeCreature* best = this->adultBestIndividual;
    for (int i = 0; i < validationSet->size(); i++) {
        if ((*validationSet)[i].size() > 0 ) {
            if (best->tree->decide((*validationSet)[i]) == (*validationSet)[i][(*validationSet)[i].size()-1]) {
                right++;
            }
            else wrong ++;
        }
    }
    return (float) right / (float)validationSet->size();
}







