//
//  EMTI_Population_ID3PopInit.cpp
//  EMTI_0.2_swift
//
//  Created by 2016 Anders Nikolai Fuglseth on 22.04.2016.
//

#include "EMTI_Population_ID3PopInit.hpp"

EMTI_Population_ID3PopInit::EMTI_Population_ID3PopInit (vector<int>* fullAttributeScope, vector<vector<unsigned>>* trainingSet, int maxTreeSize, int populationSize, int treeSize_fitnessWeight, float mutationRate, float crossOverRateParameter, bool rejectWorseGeneration, int parentSelectionType, int adultSelectionType, bool verbose, bool verboseplus, bool writePlotData, std::string plotOutputLocation, int tournamentSizeParameter, float epsilonParameter, float spiceRatio)

: EMTI_Population (fullAttributeScope, trainingSet, maxTreeSize, populationSize, treeSize_fitnessWeight, mutationRate, crossOverRateParameter, rejectWorseGeneration, parentSelectionType, adultSelectionType, verbose, verboseplus, writePlotData, plotOutputLocation, tournamentSizeParameter, epsilonParameter, spiceRatio)

{
    
    for (EMTI_TreeCreature* child : *children) {
        delete child;
    }
    children->clear();
    delete children;
    
    for (EMTI_TreeCreature* adult : *adults) {
        delete adult;
    }
    adults->clear();
    delete adults;
    
    /* create new random population: */
    children = new vector<EMTI_TreeCreature*>;
    adults = new vector<EMTI_TreeCreature*>;
    
    float currentBestFitness = std::numeric_limits<float>::min();;
    for (int i = 0; i<populationSize; i++) {
        
        EMTI_TreeCreature* ind = new EMTI_TreeCreature(fullAttributeScope);
        
        vector<vector<unsigned>>* ID3subSet = getNewRandomSubset(EMTI_Population::trainingSet, trainingSetRatio);
        EMTI_ID3Classifier* classifier = new EMTI_ID3Classifier(ID3subSet, EMTI_Population::fullAttributeScope);
        classifier->makeTree();
        
        ind->tree = new EMTI_DecisionTree (*classifier->tree);
        
        delete ID3subSet;
        delete classifier;
        
        ind->assignFitness(trainingSet, treeSize_fitnessWeight);
        if (ind->getFitness() > currentBestFitness) {
            currentBestFitness = ind->getFitness();
            childrenBestIndividual = ind;
        }
        children->push_back(ind);
    }
    adultBestIndividual = childrenBestIndividual;
}

void EMTI_Population_ID3PopInit::initPopulation() {
    
    
}

vector<vector<unsigned>>* EMTI_Population_ID3PopInit::getNewRandomSubset(vector<vector<unsigned>>* set, float size) {
    
    /* initialize random seed based on current time*/
    //srand ((int)time(NULL));
    
    vector<vector<unsigned>>* randomizedSubSet = new vector<vector<unsigned>>();
    
    while (randomizedSubSet->size() < set->size() * size) {
        
        int randomIndex = random_integer(0, (int)set->size()-1);
        
        randomizedSubSet->push_back((*set)[randomIndex]);
        
    }
    return randomizedSubSet;

}

int EMTI_Population_ID3PopInit::random_integer(int start, int end) {
    int range = end - start + 1;
    int number;
    
    if (end>start) {
        number = rand() % range + start;
    }
    else {
        number = start;
    }
    
    return number;
}








