//
//  EMTI_TreeCreature.hpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth on 30.10.2015.
//
// An individual.

#ifndef EMTI_TreeCreature_hpp
#define EMTI_TreeCreature_hpp

#include <stdio.h>
#include <unordered_set>

#include "EMTI_DecisionTree.hpp"

#include <pthread.h>



class EMTI_TreeCreature {
    
public:
    
    EMTI_TreeCreature(vector<int>* fullAttributeScope);
    
    EMTI_TreeCreature(vector<int>* attributeScope, int categoryScope);
    
    ~EMTI_TreeCreature();
    
    EMTI_TreeCreature();
    
    EMTI_TreeCreature (const EMTI_TreeCreature &b);
    
    void createRandomTree ();
    
    EMTI_TreeNode* createRandomSubTree (EMTI_TreeNode* parent, int childIndex);
    
    int mutate (float mutationRate); //Returns how many mutations were performed
    
    static pair<EMTI_TreeCreature*, EMTI_TreeCreature*> crossover (pair<EMTI_TreeCreature*, EMTI_TreeCreature*> parents, float crossoverRate, vector<int>* fullAttributeScope);
    
    void assignFitness (vector<vector<unsigned>>* trainingSet, int sizeWeight);
    
    void setFitness (float fitness);
    
    float getFitness ();
    
    void determineEndNodeAccuraciesAndTreeTraffic (vector<vector<unsigned>>* trainingSet);
    
    string to_string();
    
    EMTI_DecisionTree* tree;
    
    vector<int>* attributeScope;
    
    int categoryScope;
    
    int classificationAccuracy = 0;
    
private:
    
    float fitness = 0;
    
    static int random_number(int start, int end);
    static double random_double(double start, double end);
    
    void assignAccuracyUpToRoot (EMTI_TreeNode* n);
};

#endif /* EMTI_TreeCreature_hpp */
