//
//  ExperimentSuite2.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth 2016.

#include "ExperimentSuite2.hpp"

Experiment2::Experiment2(EMTI_Data* d) {
    this->data = d;
}

void Experiment2::runCrossvalidation(int n, int folds) {
    
    double rightAverage;
    double SD;
    int total;
    int rightTotal;
    vector<double> rights;
    double sizeAverage;
    
    chrono::high_resolution_clock::time_point t1;
    chrono::high_resolution_clock::time_point t2;
    long long duration;
    
    srand ((int)time(NULL));
    
    
    const int populationSize = 100;
    const int cycles = 600;
    const float mutationRate = 1.0;
    const float crossoverRate = 1.0;
    const int parentSelection = 2;
    const int adultSelection = 2;
    const float spiceRatio = 0.1;
    const int k = 3;
    
    cout << "-- EMTI experiment run " << folds << " folds " << n << " times, on " << currentDateTime() << " : --\n";
    
    /* ------------- ID3 ----------------------------------------------------------------------------------*/
    total = 0;
    rightTotal = 0;
    
    t1 = chrono::high_resolution_clock::now();
    
    for (int j = 0; j < n; j++) {
        
        data->randomizeDataOrder();
        vector<vector<unsigned>>* trainingSet;
        vector<vector<unsigned>>* testSet;
        
        for (int i = 0; i < folds; i++) {
            
            double testStart = i * (1 / (double)folds);
            double testEnd = (i + 1) * (1 / (double)folds);
            
            //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
            
            trainingSet = data->getNewTrainingSubset(0.0, testStart);
            vector<vector<unsigned>>* trainingSetTail = data->getNewTrainingSubset(testEnd, 1.0);
            
            trainingSet->insert(trainingSet->end(), trainingSetTail->begin(), trainingSetTail->end());
            
            testSet = data->getNewTrainingSubset(testStart, testEnd);
            
            vector<int>* fattrScope = data->getNewFullAttributeScope();
            
            /*----INIT CLASSIFIER ------------------------------------------------------------*/
            EMTI_ID3Classifier ID3(trainingSet, fattrScope);
            ID3.makeTree();
            /*--------------------------------------------------------------------------------*/
            
            int right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (ID3.tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal++;
                    }
                }
            }
            rights.push_back((double)right / testSet->size());
            total += testSet->size();
            sizeAverage += ID3.tree->nodesCount;
            
            delete fattrScope;
            delete testSet;
            delete trainingSet;
            delete trainingSetTail;
            
            cout << ".";
        }
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights.clear();
    
    
    cout << "ID3: " << rightTotal << "/" << total << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal / ((double)total / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage /(folds * n) << ")\n"
    << "      (Duration=" << duration << " microseconds " << folds << " folds " << n << " times)\n";

    
    
    /* ------------- KNN ----------------------------------------------------------------------------------*/
    
    for (int k = 1; k < 20; k+=2) {
        total = 0;
        rightTotal = 0;
        
        t1 = chrono::high_resolution_clock::now();
        
        for (int j = 0; j < n; j++) {
            
            data->randomizeDataOrder();
            vector<vector<unsigned>>* trainingSet;
            vector<vector<unsigned>>* testSet;
            vector<vector<unsigned>>* testSetNoCategory;
            
            for (int i = 0; i < folds; i++) {
                
                double testStart = i * (1 / (double)folds);
                double testEnd = (i + 1) * (1 / (double)folds);
                
                //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
                
                trainingSet = data->getNewTrainingSubset(0.0, testStart);
                vector<vector<unsigned>>* trainingSetTail = data->getNewTrainingSubset(testEnd, 1.0);
                
                trainingSet->insert(trainingSet->end(), trainingSetTail->begin(), trainingSetTail->end());
                
                testSet = data->getNewTrainingSubset(testStart, testEnd);
                testSetNoCategory = data->getNewTestSubset(testStart, testEnd);
                
                vector<int>* fattrScope = data->getNewFullAttributeScope();
                
                /*----INIT CLASSIFIER ------------------------------------------------------------*/
                EMTI_KNearestClassifier kNN(trainingSet);
                /*--------------------------------------------------------------------------------*/
                
                int right = 0;
                for (int i = 0; i < testSet->size(); i++) {
                    if ((*testSet)[i].size() > 0 ) {
                        if (kNN.classify(&(*testSetNoCategory)[i], k) == (*testSet)[i][(*testSet)[i].size()-1]) {
                            right++;
                            rightTotal++;
                        }
                    }
                }
                rights.push_back((double)right / testSet->size());
                total += testSet->size();
                //sizeAverage += ID3.tree->nodesCount;
                
                delete fattrScope;
                delete testSet;
                delete testSetNoCategory;
                delete trainingSet;
                delete trainingSetTail;
                
                cout << ".";
            }
        }
        
        t2 = chrono::high_resolution_clock::now();
        duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
        
        cout << "\n";
        
        rightAverage = 0;
        
        for (int i = 0; i < folds * n; i++) {
            rightAverage += rights[i];
        }
        rightAverage = rightAverage / (folds * n);
        
        SD = 0;
        for (int i = 0; i < folds * n; i++) {
            SD += pow(rights[i] - rightAverage, 2);
        }
        
        SD = sqrt(SD/(folds * n));
        
        rights.clear();
        
        
        cout << k << "NN: " << rightTotal << "/" << total << " avg=" << rightAverage << " SD=" << SD << "\n"
        << "      (" << (double) rightTotal / ((double)total / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
        << "      (avg.size=N/A)\n"
        << "      (Duration=" << duration << " microseconds " << folds << " folds " << n << " times)\n";
    }
    
    
    
    
    /* ------------- EMTI BASIC------------------------------------------------------------------------------*/
    
    int trainingRightTotal = 0;
    long trainingSetSizeTotal = 0;
    
    int total_PRE = 0;
    int rightTotal_PRE = 0;
    double sizeAverage_PRE = 0;
    vector<double> rights_PRE;
    
    int total_OPT = 0;
    int rightTotal_OPT = 0;
    double sizeAverage_OPT = 0;
    vector<double> rights_OPT;
    
    int trainingRightTotal_PRUN1 = 0;
    
    int total_PRUN1 = 0;
    int rightTotal_PRUN1 = 0;
    double sizeAverage_PRUN1 = 0;
    vector<double> rights_PRUN1;
    
    t1 = chrono::high_resolution_clock::now();
    
    for (int j = 0; j < n; j++) {
        
        data->randomizeDataOrder();
        vector<vector<unsigned>>* trainingSet;
        vector<vector<unsigned>>* testSet;
        
        for (int i = 0; i < folds; i++) {
            
            double testStart = i * (1 / (double)folds);
            double testEnd = (i + 1) * (1 / (double)folds);
            
            //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
            
            trainingSet = data->getNewTrainingSubset(0.0, testStart);
            vector<vector<unsigned>>* trainingSetTail = data->getNewTrainingSubset(testEnd, 1.0);
            
            trainingSet->insert(trainingSet->end(), trainingSetTail->begin(), trainingSetTail->end());
            
            testSet = data->getNewTrainingSubset(testStart, testEnd);
            
            vector<int>* fattrScope = data->getNewFullAttributeScope();
            
            string output = "/Users/avividi/Dropbox/fordypningsprosjekt_Anders_Fuglseth_2015/plotdata/plotData.dat";
            std::ofstream myfile;
            myfile.open (output);
            myfile.close();
            
            /*----INIT CLASSIFIER ------------------------------------------------------------*/
            EMTI_Population p(
                                      fattrScope, //vector<int>* fullAttributeScope,
                                      trainingSet, //vector<vector<int>* trainingSet
                                      0, //int maxTreeSize (< 0 for unlimited)
                                      populationSize, //int populationSize
                                      1000, //int treeSize_fitnessWeight
                                      mutationRate, //float mutationRate
                                      crossoverRate, //float crossOverRateParameter
                                      true, //bool rejectWorseGeneration
                                      parentSelection, //int parentSelectionType
                                      /*
                                       0 = parent_selection_fitness_proportionate
                                       1 = parent_selection_tournament
                                       2 = parent_selection_sigma_scaling
                                       */
                                      adultSelection, //int adultSelectionType
                                      /*
                                       0 = adult_selection_full_generational_replacement
                                       1 = adult_selection_overproduction
                                       2 = adult_selection_generational_mixing
                                       3 = adult_selection_add_spice(float spiceRatio)
                                       */
                                      false, //bool verbose
                                      false, //bool verboseplus
                                      false, //bool writeplotdata
                                      output, //std::string plotOutputLocation
                                      20, //int tournamentSizeParameter (only for tournament parent selection)
                                      0.8, //float epsilonParameter (only for tournament parent selection)
                                      spiceRatio //float spiceRatio (only for adult_selection_add_spice)
                                      );
            

            
            while (p.cycleCount < cycles) {
                p.do_evolution_cycle();
            }
            
            trainingRightTotal += p.adultBestIndividual->classificationAccuracy;
            
            EMTI_TreeCreature* best = p.adultBestIndividual;
            
            trainingSetSizeTotal += trainingSet->size();
            
            /*--------------------------------------------------------------------------------*/
            
            int right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_PRE++;
                    }
                }
            }
            
            rights_PRE.push_back((double)right / testSet->size());
            total_PRE += testSet->size();
            sizeAverage_PRE += best->tree->nodesCount;
            
            /*--------------------------------------------------------------------------------*/
            
            p.doPostCycleEndNodeOptimizations(k, false);
            
            right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_OPT++;
                    }
                }
            }
            
            rights_OPT.push_back((double)right / testSet->size());
            total_OPT += testSet->size();
            sizeAverage_OPT += best->tree->nodesCount;
            
            /*--------------------------------------------------------------------------------*/
            
            best->tree = EMTI_TreePruner::minimum_error_prune(best->tree, trainingSet);
            
            int trainingRight = 0;
            for (int i = 0; i < trainingSet->size(); i++) {
                if ((*trainingSet)[i].size() > 0 ) {
                    if (best->tree->decide((*trainingSet)[i]) == (*trainingSet)[i][(*trainingSet)[i].size()-1]) {
                        trainingRight++;
                        trainingRightTotal_PRUN1++;
                    }
                }
            }
            
            right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_PRUN1++;
                    }
                }
            }
            
            rights_PRUN1.push_back((double)right / testSet->size());
            total_PRUN1 += testSet->size();
            sizeAverage_PRUN1 += best->tree->nodesCount;

            
            delete fattrScope;
            delete testSet;
            delete trainingSet;
            delete trainingSetTail;
            
            cout << ".";
        }
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_PRE[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_PRE[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights_PRE.clear();
    
    
    cout
    << "EMTI_Basic_PRE: " << rightTotal_PRE << "/" << total_PRE << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_PRE / ((double)total_PRE / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_PRE /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal / trainingSetSizeTotal << ")\n"
    << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_OPT[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_OPT[i] - rightAverage, 2);
    }
   
    SD = sqrt(SD/(folds * n));
    
    rights_OPT.clear();
    
    
    cout
    << "EMTI_Basic_OPT: " << rightTotal_OPT << "/" << total_OPT << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_OPT / ((double)total_OPT / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_OPT /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal / trainingSetSizeTotal << ")\n"
    << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_PRUN1[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_PRUN1[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights_PRUN1.clear();
    
    
    cout
    << "EMTI_Basic_Min-Err-pruned: " << rightTotal_PRUN1 << "/" << total_PRUN1 << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_PRUN1 / ((double)total_PRUN1 / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_PRUN1 /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal_PRUN1 / trainingSetSizeTotal << ")\n"
    << "\n";
    
    cout << "      (Duration=" << duration << " microseconds " << folds << " folds " << n << " times)\n"
    << "      (Params: cycles=" << cycles << " popSize=" << populationSize << " mutation=" << mutationRate << " crossover=" << crossoverRate << " \nparentSelection=" << parentSelection << " adultSelection=" << adultSelection << " spice=" << spiceRatio << " k=" << k << ")\n";
    
    
    
    /* ------------- EMTI ES1: Single rollback -----------------------------------------------------------------------*/
    
    int trainingRightTotal_PRE = 0;
    int trainingSetSizeTotal_PRE = 0;
    
    total_PRE = 0;
    rightTotal_PRE = 0;
    sizeAverage_PRE = 0;
    
    int trainingRightTotal_OPT = 0;
    int trainingSetSizeTotal_OPT = 0;
    
    total_OPT = 0;
    rightTotal_OPT = 0;
    sizeAverage_OPT = 0;
    
    trainingRightTotal_PRUN1 = 0;
    
    total_PRUN1 = 0;
    rightTotal_PRUN1 = 0;
    sizeAverage_PRUN1 = 0;
    
    t1 = chrono::high_resolution_clock::now();
    
    for (int j = 0; j < n; j++) {
        
        data->randomizeDataOrder();
        vector<vector<unsigned>>* trainingSet;
        vector<vector<unsigned>>* testSet;
        
        for (int i = 0; i < folds; i++) {
            
            double testStart = i * (1 / (double)folds);
            double testEnd = (i + 1) * (1 / (double)folds);
            
            //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
            
            trainingSet = data->getNewTrainingSubset(0.0, testStart);
            vector<vector<unsigned>>* trainingSetTail = data->getNewTrainingSubset(testEnd, 1.0);
            
            trainingSet->insert(trainingSet->end(), trainingSetTail->begin(), trainingSetTail->end());
            
            testSet = data->getNewTrainingSubset(testStart, testEnd);
            
            vector<int>* fattrScope = data->getNewFullAttributeScope();
            
            string output = "/Users/avividi/Dropbox/fordypningsprosjekt_Anders_Fuglseth_2015/plotdata/plotData.dat";
            std::ofstream myfile;
            myfile.open (output);
            myfile.close();
            
            /*----INIT CLASSIFIER ------------------------------------------------------------*/
            EMTI_PopulationES1 p(
                              fattrScope, //vector<int>* fullAttributeScope,
                              trainingSet, //vector<vector<int>* trainingSet
                              0, //int maxTreeSize (< 0 for unlimited)
                              populationSize, //int populationSize
                              1000, //int treeSize_fitnessWeight
                              mutationRate, //float mutationRate
                              crossoverRate, //float crossOverRateParameter
                              true, //bool rejectWorseGeneration
                              parentSelection, //int parentSelectionType
                              /*
                               0 = parent_selection_fitness_proportionate
                               1 = parent_selection_tournament
                               2 = parent_selection_sigma_scaling
                               */
                              adultSelection, //int adultSelectionType
                              /*
                               0 = adult_selection_full_generational_replacement
                               1 = adult_selection_overproduction
                               2 = adult_selection_generational_mixing
                               3 = adult_selection_add_spice(float spiceRatio)
                               */
                              false, //bool verbose
                              false, //bool verboseplus
                              false, //bool writeplotdata
                              output, //std::string plotOutputLocation
                              20, //int tournamentSizeParameter (only for tournament parent selection)
                              0.8, //float epsilonParameter (only for tournament parent selection)
                              spiceRatio //float spiceRatio (only for adult_selection_add_spice)
                              );
            
            
            
            while (p.cycleCount < cycles) {
                p.do_evolution_cycle();
            }
            
            p.doPostCycleEndNodeOptimizations(k, false);
            
            trainingRightTotal_PRE += p.adultBestIndividual->classificationAccuracy;
            
            trainingSetSizeTotal_PRE += p.trainingSet->size();
            
            EMTI_TreeCreature* best = p.adultBestIndividual;
            
            /*--------------------------------------------------------------------------------*/
            
            int right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_PRE++;
                    }
                }
            }
            
            rights_PRE.push_back((double)right / testSet->size());
            total_PRE += testSet->size();
            sizeAverage_PRE += best->tree->nodesCount;
            
            
            /*--------------------------------------------------------------------------------*/
            
            EMTI_DecisionTree* pruneTree = EMTI_TreePruner::error_complexity_prune(best->tree, p.trainingSet, p.validationSet);
            
            int trainingRight = 0;
            for (int i = 0; i < p.trainingSet->size(); i++) {
                if ((*p.trainingSet)[i].size() > 0 ) {
                    if (pruneTree->decide((*p.trainingSet)[i]) == (*p.trainingSet)[i][(*p.trainingSet)[i].size()-1]) {
                        trainingRight++;
                        trainingRightTotal_PRUN1++;
                    }
                }
            }
            
            right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (pruneTree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_PRUN1++;
                    }
                }
            }
            
            rights_PRUN1.push_back((double)right / testSet->size());
            total_PRUN1 += testSet->size();
            sizeAverage_PRUN1 += pruneTree->nodesCount;
            
            delete pruneTree;
            
            /*--------------------------------------------------------------------------------*/
            p.doPostCycle_rollBackCheck();
            
            trainingRightTotal_OPT += p.adultBestIndividual->classificationAccuracy;
            
            trainingSetSizeTotal_OPT += p.trainingSet->size();
            
            best = p.adultBestIndividual;
            
            right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_OPT++;
                    }
                }
            }
            
            rights_OPT.push_back((double)right / testSet->size());
            total_OPT += testSet->size();
            sizeAverage_OPT += best->tree->nodesCount;
            
            delete fattrScope;
            delete testSet;
            delete trainingSet;
            delete trainingSetTail;
            
            cout << ".";
        }
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_PRE[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_PRE[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights_PRE.clear();
    
    
    cout
    << "EMTI_Valset_OPT_Norollback: " << rightTotal_PRE << "/" << total_PRE << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_PRE / ((double)total_PRE / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_PRE /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal_PRE / trainingSetSizeTotal_PRE << ")\n"
    << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_PRUN1[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_PRUN1[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights_PRUN1.clear();
    
    
    cout
    << "EMTI_Basic_Comp-err-pruned: " << rightTotal_PRUN1 << "/" << total_PRUN1 << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_PRUN1 / ((double)total_PRUN1 / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_PRUN1 /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal_PRUN1 / trainingSetSizeTotal_PRE << ")\n"
    << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_OPT[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_OPT[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights_OPT.clear();
    
    
    cout
    << "EMTI_Valset_OPT_Rollback: " << rightTotal_OPT << "/" << total_OPT << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_OPT / ((double)total_OPT / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_OPT /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal_OPT / trainingSetSizeTotal_OPT << ")\n"
    << "\n";
    
    cout << "      (Duration=" << duration << " microseconds " << folds << " folds " << n << " times)\n"
    << "      (Params: cycles=" << cycles << " popSize=" << populationSize << " mutation=" << mutationRate << " crossover=" << crossoverRate << " \nparentSelection=" << parentSelection << " adultSelection=" << adultSelection << " spice=" << spiceRatio << " k=" << k << ")\n";
    
    
    
    
    
    /* ------------- EMTI ES2: Multi rollback -----------------------------------------------------------------------*/
//    
//    trainingRightTotal_PRE = 0;
//    trainingSetSizeTotal_PRE = 0;
//    
//    total_PRE = 0;
//    rightTotal_PRE = 0;
//    sizeAverage_PRE = 0;
//    
//    trainingRightTotal_OPT = 0;
//    trainingSetSizeTotal_OPT = 0;
//    
//    total_OPT = 0;
//    rightTotal_OPT = 0;
//    sizeAverage_OPT = 0;
//    
//    t1 = chrono::high_resolution_clock::now();
//    
//    for (int j = 0; j < n; j++) {
//        
//        data->randomizeDataOrder();
//        vector<vector<unsigned>>* trainingSet;
//        vector<vector<unsigned>>* testSet;
//        
//        for (int i = 0; i < folds; i++) {
//            
//            double testStart = i * (1 / (double)folds);
//            double testEnd = (i + 1) * (1 / (double)folds);
//            
//            //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
//            
//            trainingSet = data->getNewTrainingSubset(0.0, testStart);
//            vector<vector<unsigned>>* trainingSetTail = data->getNewTrainingSubset(testEnd, 1.0);
//            
//            trainingSet->insert(trainingSet->end(), trainingSetTail->begin(), trainingSetTail->end());
//            
//            testSet = data->getNewTrainingSubset(testStart, testEnd);
//            
//            vector<int>* fattrScope = data->getNewFullAttributeScope();
//            
//            string output = "/Users/avividi/Dropbox/fordypningsprosjekt_Anders_Fuglseth_2015/plotdata/plotData.dat";
//            std::ofstream myfile;
//            myfile.open (output);
//            myfile.close();
//            
//            /*----INIT CLASSIFIER ------------------------------------------------------------*/
//            EMTI_PopulationES2 p(
//                                 fattrScope, //vector<int>* fullAttributeScope,
//                                 trainingSet, //vector<vector<int>* trainingSet
//                                 0, //int maxTreeSize (< 0 for unlimited)
//                                 populationSize, //int populationSize
//                                 1000, //int treeSize_fitnessWeight
//                                 mutationRate, //float mutationRate
//                                 crossoverRate, //float crossOverRateParameter
//                                 true, //bool rejectWorseGeneration
//                                 parentSelection, //int parentSelectionType
//                                 /*
//                                  0 = parent_selection_fitness_proportionate
//                                  1 = parent_selection_tournament
//                                  2 = parent_selection_sigma_scaling
//                                  */
//                                 adultSelection, //int adultSelectionType
//                                 /*
//                                  0 = adult_selection_full_generational_replacement
//                                  1 = adult_selection_overproduction
//                                  2 = adult_selection_generational_mixing
//                                  3 = adult_selection_add_spice(float spiceRatio)
//                                  */
//                                 false, //bool verbose
//                                 false, //bool verboseplus
//                                 false, //bool writeplotdata
//                                 output, //std::string plotOutputLocation
//                                 20, //int tournamentSizeParameter (only for tournament parent selection)
//                                 0.8, //float epsilonParameter (only for tournament parent selection)
//                                 spiceRatio //float spiceRatio (only for adult_selection_add_spice)
//                                 );
//            
//            
//            
//            while (p.cycleCount < cycles) {
//                p.do_evolution_cycle();
//            }
//            
//            p.doPostCycleEndNodeOptimizations(k, false);
//            
//            trainingRightTotal_PRE += p.adultBestIndividual->classificationAccuracy;
//            
//            trainingSetSizeTotal_PRE += p.trainingSet->size();
//            
//            EMTI_TreeCreature* best = p.adultBestIndividual;
//            
//            /*--------------------------------------------------------------------------------*/
//            
//            int right = 0;
//            for (int i = 0; i < testSet->size(); i++) {
//                if ((*testSet)[i].size() > 0 ) {
//                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
//                        right++;
//                        rightTotal_PRE++;
//                    }
//                }
//            }
//            
//            rights_PRE.push_back((double)right / testSet->size());
//            total_PRE += testSet->size();
//            sizeAverage_PRE += best->tree->nodesCount;
//            
//            /*--------------------------------------------------------------------------------*/
//            p.doPostCycle_rollBackCheck();
//            
//            trainingRightTotal_OPT += p.adultBestIndividual->classificationAccuracy;
//            
//            trainingSetSizeTotal_OPT += p.trainingSet->size();
//            
//            best = p.adultBestIndividual;
//            
//            right = 0;
//            for (int i = 0; i < testSet->size(); i++) {
//                if ((*testSet)[i].size() > 0 ) {
//                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
//                        right++;
//                        rightTotal_OPT++;
//                    }
//                }
//            }
//            
//            rights_OPT.push_back((double)right / testSet->size());
//            total_OPT += testSet->size();
//            sizeAverage_OPT += best->tree->nodesCount;
//            
//            delete fattrScope;
//            delete testSet;
//            delete trainingSet;
//            delete trainingSetTail;
//            
//            cout << ".";
//        }
//    }
//    
//    t2 = chrono::high_resolution_clock::now();
//    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
//    
//    cout << "\n";
//    
////    rightAverage = 0;
////    
////    for (int i = 0; i < folds * n; i++) {
////        rightAverage += rights_PRE[i];
////    }
////    rightAverage = rightAverage / (folds * n);
////    
////    SD = 0;
////    for (int i = 0; i < folds * n; i++) {
////        SD += pow(rights_PRE[i] - rightAverage, 2);
////    }
////    
////    SD = sqrt(SD/(folds * n));
////    
////    rights_PRE.clear();
////    
////    
////    cout
////    << "EMTI_Valset_OPT_Norollback: " << rightTotal_PRE << "/" << total_PRE << " avg=" << rightAverage << " SD=" << SD << "\n"
////    << "      (" << (double) rightTotal_PRE / ((double)total_PRE / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
////    << "      (avg.size=" << sizeAverage_PRE /(folds * n) << ")\n"
////    << "      (trainingaccuracy reached=" << (double)trainingRightTotal_PRE / trainingSetSizeTotal_PRE << ")\n"
////    << "\n";
//    
//    rightAverage = 0;
//    
//    for (int i = 0; i < folds * n; i++) {
//        rightAverage += rights_OPT[i];
//    }
//    rightAverage = rightAverage / (folds * n);
//    
//    SD = 0;
//    for (int i = 0; i < folds * n; i++) {
//        SD += pow(rights_OPT[i] - rightAverage, 2);
//    }
//    
//    SD = sqrt(SD/(folds * n));
//    
//    rights_OPT.clear();
//    
//    
//    cout
//    << "EMTI_Valset_OPT_MultiRollback: " << rightTotal_OPT << "/" << total_OPT << " avg=" << rightAverage << " SD=" << SD << "\n"
//    << "      (" << (double) rightTotal_OPT / ((double)total_OPT / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
//    << "      (avg.size=" << sizeAverage_OPT /(folds * n) << ")\n"
//    << "      (trainingaccuracy reached=" << (double)trainingRightTotal_OPT / trainingSetSizeTotal_OPT << ")\n"
//    << "\n";
//    
//    cout << "      (Duration=" << duration << " microseconds " << folds << " folds " << n << " times)\n"
//    << "      (Params: cycles=" << cycles << " popSize=" << populationSize << " mutation=" << mutationRate << " crossover=" << crossoverRate << " \nparentSelection=" << parentSelection << " adultSelection=" << adultSelection << " spice=" << spiceRatio << " k=" << k << ")\n";
//    
    
    
    /* ------------- EMTI ES2: Max tree size adjustment  --------------------------------------------------------------*/
    
    trainingRightTotal_PRE = 0;
    trainingSetSizeTotal_PRE = 0;
    
    total_PRE = 0;
    rightTotal_PRE = 0;
    sizeAverage_PRE = 0;
    
    trainingRightTotal_OPT = 0;
    trainingSetSizeTotal_OPT = 0;
    
    total_OPT = 0;
    rightTotal_OPT = 0;
    sizeAverage_OPT = 0;
    
    t1 = chrono::high_resolution_clock::now();
    
    for (int j = 0; j < n; j++) {
        
        data->randomizeDataOrder();
        vector<vector<unsigned>>* trainingSet;
        vector<vector<unsigned>>* testSet;
        
        for (int i = 0; i < folds; i++) {
            
            double testStart = i * (1 / (double)folds);
            double testEnd = (i + 1) * (1 / (double)folds);
            
            //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
            
            trainingSet = data->getNewTrainingSubset(0.0, testStart);
            vector<vector<unsigned>>* trainingSetTail = data->getNewTrainingSubset(testEnd, 1.0);
            
            trainingSet->insert(trainingSet->end(), trainingSetTail->begin(), trainingSetTail->end());
            
            testSet = data->getNewTrainingSubset(testStart, testEnd);
            
            vector<int>* fattrScope = data->getNewFullAttributeScope();
            
            string output = "/Users/avividi/Dropbox/fordypningsprosjekt_Anders_Fuglseth_2015/plotdata/plotData.dat";
            std::ofstream myfile;
            myfile.open (output);
            myfile.close();
            
            /*----INIT CLASSIFIER ------------------------------------------------------------*/
            EMTI_PopulationES3 p(
                                 fattrScope, //vector<int>* fullAttributeScope,
                                 trainingSet, //vector<vector<int>* trainingSet
                                 0, //int maxTreeSize (< 0 for unlimited)
                                 populationSize, //int populationSize
                                 1000, //int treeSize_fitnessWeight
                                 mutationRate, //float mutationRate
                                 crossoverRate, //float crossOverRateParameter
                                 true, //bool rejectWorseGeneration
                                 parentSelection, //int parentSelectionType
                                 /*
                                  0 = parent_selection_fitness_proportionate
                                  1 = parent_selection_tournament
                                  2 = parent_selection_sigma_scaling
                                  */
                                 adultSelection, //int adultSelectionType
                                 /*
                                  0 = adult_selection_full_generational_replacement
                                  1 = adult_selection_overproduction
                                  2 = adult_selection_generational_mixing
                                  3 = adult_selection_add_spice(float spiceRatio)
                                  */
                                 false, //bool verbose
                                 false, //bool verboseplus
                                 false, //bool writeplotdata
                                 output, //std::string plotOutputLocation
                                 20, //int tournamentSizeParameter (only for tournament parent selection)
                                 0.8, //float epsilonParameter (only for tournament parent selection)
                                 spiceRatio //float spiceRatio (only for adult_selection_add_spice)
                                 );
            
            
            
            while (p.cycleCount < cycles) {
                p.do_evolution_cycle();
            }
            
            p.doPostCycleEndNodeOptimizations(k, false);
            
            trainingRightTotal_PRE += p.adultBestIndividual->classificationAccuracy;
            
            trainingSetSizeTotal_PRE += p.trainingSet->size();
            
            EMTI_TreeCreature* best = p.adultBestIndividual;
            
            /*--------------------------------------------------------------------------------*/
            
            int right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_PRE++;
                    }
                }
            }
            
            rights_PRE.push_back((double)right / testSet->size());
            total_PRE += testSet->size();
            sizeAverage_PRE += best->tree->nodesCount;
            
            /*--------------------------------------------------------------------------------*/
            p.doPostCycle_rollBackCheck();
            
            trainingRightTotal_OPT += p.adultBestIndividual->classificationAccuracy;
            
            trainingSetSizeTotal_OPT += p.trainingSet->size();
            
            best = p.adultBestIndividual;
            
            right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_OPT++;
                    }
                }
            }
            
            rights_OPT.push_back((double)right / testSet->size());
            total_OPT += testSet->size();
            sizeAverage_OPT += best->tree->nodesCount;
            
            delete fattrScope;
            delete testSet;
            delete trainingSet;
            delete trainingSetTail;
            
            cout << ".";
        }
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
        rightAverage = 0;
    
        for (int i = 0; i < folds * n; i++) {
            rightAverage += rights_PRE[i];
        }
        rightAverage = rightAverage / (folds * n);
    
        SD = 0;
        for (int i = 0; i < folds * n; i++) {
            SD += pow(rights_PRE[i] - rightAverage, 2);
        }
    
        SD = sqrt(SD/(folds * n));
    
        rights_PRE.clear();
    
    
        cout
        << "EMTI_Valset_OPT_MTSize_norollback: " << rightTotal_PRE << "/" << total_PRE << " avg=" << rightAverage << " SD=" << SD << "\n"
        << "      (" << (double) rightTotal_PRE / ((double)total_PRE / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
        << "      (avg.size=" << sizeAverage_PRE /(folds * n) << ")\n"
        << "      (trainingaccuracy reached=" << (double)trainingRightTotal_PRE / trainingSetSizeTotal_PRE << ")\n"
        << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_OPT[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_OPT[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights_OPT.clear();
    
    
    cout
    << "EMTI_Valset_OPT_MTSize_rollback: " << rightTotal_OPT << "/" << total_OPT << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_OPT / ((double)total_OPT / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_OPT /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal_OPT / trainingSetSizeTotal_OPT << ")\n"
    << "\n";
    
    cout << "      (Duration=" << duration << " microseconds " << folds << " folds " << n << " times)\n"
    << "      (Params: cycles=" << cycles << " popSize=" << populationSize << " mutation=" << mutationRate << " crossover=" << crossoverRate << " \nparentSelection=" << parentSelection << " adultSelection=" << adultSelection << " spice=" << spiceRatio << " k=" << k << ")\n";
    
}

// Get current date/time, format is YYYY-MM-DD.HH:mm:ss
const std::string Experiment2::currentDateTime() {
    time_t     now = time(0);
    struct tm  tstruct;
    char       buf[80];
    tstruct = *localtime(&now);
    strftime(buf, sizeof(buf), "%Y-%m-%d.%X", &tstruct);
    
    return buf;
}








