//
//  ExperimentSuite3.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth 2016.

#include "ExperimentSuite3.hpp"

Experiment3::Experiment3() {
//    this->traindata = new EMTI_Data("/Users/avividi/monks/monks-2.train.txt", ' ', {7}, 0);
//    
//    
//    this->testdata = new EMTI_Data("/Users/avividi/monks/monks-2.test.txt",
//                                   traindata->attributeValues_int,
//                                   traindata->attributeValues_intToString,
//                                   traindata->attributeValues_stringToInt,
//                                   ' ', {7}, 0);
    
 
    this->traindata = new EMTI_Data("/Users/avividi/c45testing/mofn.data");
    
    
    this->testdata = new EMTI_Data("/Users/avividi/c45testing/mofntest.data",
                                   traindata->attributeValues_int,
                                   traindata->attributeValues_intToString,
                                   traindata->attributeValues_stringToInt);

}

void Experiment3::run(int n) {
    
    double rightAverage;
    double SD;
    int total;
    int rightTotal;
    vector<double> rights;
    double sizeAverage;
    
    chrono::high_resolution_clock::time_point t1;
    chrono::high_resolution_clock::time_point t2;
    long long duration;
    
    srand ((int)time(NULL));
    
    cout << "-- EMTI experiment run " << n << " times, on " << currentDateTime() << " : --\n";
    
    vector<vector<unsigned>>* trainingSet;
    vector<vector<unsigned>>* testSet;
    
    //    traindata->randomizeDataOrder();
    //    testdata->randomizeDataOrder();
    
    trainingSet = traindata->dataIntVector;
    testSet = testdata->dataIntVector;
    
    vector<int>* fattrScope = traindata->getNewFullAttributeScope();
    
    /* ------------- KNN ----------------------------------------------------------------------------------*/
    
    for (int k = 1; k < 20; k+=2) {
        total = 0;
        rightTotal = 0;
        
        t1 = chrono::high_resolution_clock::now();
        
        for (int j = 0; j < n; j++) {
                
                //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
                
                /*----INIT CLASSIFIER ------------------------------------------------------------*/
                EMTI_KNearestClassifier kNN(trainingSet);
                /*--------------------------------------------------------------------------------*/
                
                int right = 0;
                for (int i = 0; i < testSet->size(); i++) {
                    if ((*testSet)[i].size() > 0 ) {
                        if (kNN.classify(&(*testSet)[i], k, true) == (*testSet)[i][(*testSet)[i].size()-1]) {
                            right++;
                            rightTotal++;
                        }
                    }
                }
                rights.push_back((double)right / testSet->size());
                total += testSet->size();
                //sizeAverage += ID3.tree->nodesCount;
                
                cout << ".";
        }
        
        t2 = chrono::high_resolution_clock::now();
        duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
        
        cout << "\n";
        
        rightAverage = 0;
        
        for (int i = 0; i < n; i++) {
            rightAverage += rights[i];
        }
        rightAverage = rightAverage / (n);
        
        SD = 0;
        for (int i = 0; i < n; i++) {
            SD += pow(rights[i] - rightAverage, 2);
        }
        
        SD = sqrt(SD/(n));
        
        rights.clear();
        
        
        cout << k << "NN: " << rightTotal << "/" << total << " avg=" << rightAverage << " SD=" << SD << "\n"
        << "      (" << (double) rightTotal / ((double)total / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
        << "      (avg.size=N/A)\n"
        << "      (Duration=" << duration << " microseconds " << n << " times)\n";
    }
    
    /* ------------- ID3 ----------------------------------------------------------------------------------*/
    total = 0;
    rightTotal = 0;
    
    t1 = chrono::high_resolution_clock::now();
    
    for (int j = 0; j < n; j++) {
        
        
        /*----INIT CLASSIFIER ------------------------------------------------------------*/
        EMTI_ID3Classifier ID3(trainingSet, fattrScope);
        ID3.makeTree();
        /*--------------------------------------------------------------------------------*/
        
        int right = 0;
        for (int i = 0; i < testSet->size(); i++) {
            if ((*testSet)[i].size() > 0 ) {
                
                if (ID3.tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                    right++;
                    rightTotal++;
                }
            }
        }
        rights.push_back((double)right / testSet->size());
        total += testSet->size();
        sizeAverage += ID3.tree->nodesCount;
        
        cout << ".";
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < n; i++) {
        rightAverage += rights[i];
    }
    rightAverage = rightAverage / (n);
    
    SD = 0;
    for (int i = 0; i < n; i++) {
        SD += pow(rights[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(n));
    
    rights.clear();
    
    
    cout << "ID3: " << rightTotal << "/" << total << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal / ((double)total / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage /(n) << ")\n"
    << "      (Duration=" << duration << " microseconds " << n << " times)\n";
    
    
    /* ------------- EMTI ----------------------------------------------------------------------------------*/
    const int populationSize = 100;
    const int cycles = 1000;
    const float mutationRate = 1.0;
    const float crossoverRate = 1.0;
    const int parentSelection = 2;
    const int adultSelection = 2;
    const float spiceRatio = 0.1;
    const int k = 7;
    
    int trainingRightTotal = 0;
    long trainingSetSizeTotal = 0;
    
    int total_PRE = 0;
    int rightTotal_PRE = 0;
    double sizeAverage_PRE = 0;
    vector<double> rights_PRE;
    
    int total_OPT = 0;
    int rightTotal_OPT = 0;
    double sizeAverage_OPT = 0;
    vector<double> rights_OPT;
    
    int trainingRightTotal_PRUN1 = 0;
    
    int total_PRUN1 = 0;
    int rightTotal_PRUN1 = 0;
    double sizeAverage_PRUN1 = 0;
    vector<double> rights_PRUN1;
    
    t1 = chrono::high_resolution_clock::now();
    
    const int feedbackTimeNumber = (cycles * n) / 60;
    int feedbackIncrement = 0;
    cout << "|";
    for (int f = 0; f < 58; f++) {
        cout << " ";
    }
    cout << "|\n";
    
    for (int j = 0; j < n; j++) {
        
        
        //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
        
        string output = "/Users/avividi/Google Drive/masteroppgave/plotdata/plotData.dat";
        std::ofstream myfile;
        myfile.open (output);
        myfile.close();
        
        /*----INIT CLASSIFIER ------------------------------------------------------------*/
        EMTI_Population p(
                          fattrScope, //vector<int>* fullAttributeScope,
                          trainingSet, //vector<vector<int>* trainingSet
                          0, //int maxTreeSize (< 0 for unlimited)
                          populationSize, //int populationSize
                          1000, //int treeSize_fitnessWeight
                          mutationRate, //float mutationRate
                          crossoverRate, //float crossOverRateParameter
                          false, //bool rejectWorseGeneration
                          parentSelection, //int parentSelectionType
                          /*
                           0 = parent_selection_fitness_proportionate
                           1 = parent_selection_tournament
                           2 = parent_selection_sigma_scaling
                           */
                          adultSelection, //int adultSelectionType
                          /*
                           0 = adult_selection_full_generational_replacement
                           1 = adult_selection_overproduction
                           2 = adult_selection_generational_mixing
                           3 = adult_selection_add_spice(float spiceRatio)
                           */
                          false, //bool verbose
                          false, //bool verboseplus
                          true, //bool writeplotdata
                          output, //std::string plotOutputLocation
                          20, //int tournamentSizeParameter (only for tournament parent selection)
                          0.8, //float epsilonParameter (only for tournament parent selection)
                          spiceRatio //float spiceRatio (only for adult_selection_add_spice)
                          );
        
        
        while (p.cycleCount < cycles) {
            p.do_evolution_cycle();
            
            feedbackIncrement++;
            if (feedbackIncrement>=feedbackTimeNumber) {
                feedbackIncrement = 0;
                cout << "-";
            }
        }
        
        trainingRightTotal += p.adultBestIndividual->classificationAccuracy;
        
        EMTI_TreeCreature* best = p.adultBestIndividual;
        
        trainingSetSizeTotal += trainingSet->size();
        
        /*--------------------------------------------------------------------------------*/
        
        int right = 0;
        for (int i = 0; i < testSet->size(); i++) {
            if ((*testSet)[i].size() > 0 ) {
                if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                    right++;
                    rightTotal_PRE++;
                }
            }
        }
        
        rights_PRE.push_back((double)right / testSet->size());
        total_PRE += testSet->size();
        sizeAverage_PRE += best->tree->nodesCount;
        
        /*--------------------------------------------------------------------------------*/
        
        p.doPostCycleEndNodeOptimizations(k, false);
        
        right = 0;
        for (int i = 0; i < testSet->size(); i++) {
            if ((*testSet)[i].size() > 0 ) {
                if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                    right++;
                    rightTotal_OPT++;
                }
            }
        }
        
        rights_OPT.push_back((double)right / testSet->size());
        total_OPT += testSet->size();
        sizeAverage_OPT += best->tree->nodesCount;
        
        /*--------------------------------------------------------------------------------*/
        
        best->tree = EMTI_TreePruner::minimum_error_prune(best->tree, trainingSet);
        
        int trainingRight = 0;
        for (int i = 0; i < trainingSet->size(); i++) {
            if ((*trainingSet)[i].size() > 0 ) {
                if (best->tree->decide((*trainingSet)[i]) == (*trainingSet)[i][(*trainingSet)[i].size()-1]) {
                    trainingRight++;
                    trainingRightTotal_PRUN1++;
                }
            }
        }
        
        right = 0;
        for (int i = 0; i < testSet->size(); i++) {
            if ((*testSet)[i].size() > 0 ) {
                if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                    right++;
                    rightTotal_PRUN1++;
                }
            }
        }
        
        rights_PRUN1.push_back((double)right / testSet->size());
        total_PRUN1 += testSet->size();
        sizeAverage_PRUN1 += best->tree->nodesCount;
        
        
        //cout << ".";
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < n; i++) {
        rightAverage += rights_PRE[i];
    }
    rightAverage = rightAverage / (n);
    
    SD = 0;
    for (int i = 0; i < n; i++) {
        SD += pow(rights_PRE[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(n));
    
    rights_PRE.clear();
    
    
    cout
    << "EMTI_Basic_PRE: " << rightTotal_PRE << "/" << total_PRE << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_PRE / ((double)total_PRE / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_PRE /(n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal / trainingSetSizeTotal << ")\n"
    << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < n; i++) {
        rightAverage += rights_OPT[i];
    }
    rightAverage = rightAverage / (n);
    
    SD = 0;
    for (int i = 0; i < n; i++) {
        SD += pow(rights_OPT[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(n));
    
    rights_OPT.clear();
    
    
    cout
    << "EMTI_Basic_OPT: " << rightTotal_OPT << "/" << total_OPT << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_OPT / ((double)total_OPT / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_OPT /(n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal / trainingSetSizeTotal << ")\n"
    << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < n; i++) {
        rightAverage += rights_PRUN1[i];
    }
    rightAverage = rightAverage / (n);
    
    SD = 0;
    for (int i = 0; i < n; i++) {
        SD += pow(rights_PRUN1[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(n));
    
    rights_PRUN1.clear();
    
    
    cout
    << "EMTI_Basic_Min-Err-pruned: " << rightTotal_PRUN1 << "/" << total_PRUN1 << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_PRUN1 / ((double)total_PRUN1 / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_PRUN1 /(n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal_PRUN1 / trainingSetSizeTotal << ")\n"
    << "\n";
    
    cout << "      (Duration=" << duration << " microseconds " << n << " times)\n"
    << "      (Params: cycles=" << cycles << " popSize=" << populationSize << " mutation=" << mutationRate << " crossover=" << crossoverRate << " \nparentSelection=" << parentSelection << " adultSelection=" << adultSelection << " spice=" << spiceRatio << " k=" << k << ")\n";
    
    delete fattrScope;
    
}

// Get current date/time, format is YYYY-MM-DD.HH:mm:ss
const std::string Experiment3::currentDateTime() {
    time_t     now = time(0);
    struct tm  tstruct;
    char       buf[80];
    tstruct = *localtime(&now);
    strftime(buf, sizeof(buf), "%Y-%m-%d.%X", &tstruct);
    
    return buf;
}








