//
//  ExperimentSuite2.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth 2016.

#include "ExperimentSuite4.hpp"

Experiment4::Experiment4(EMTI_Data* d) {
    this->data = d;
}

void Experiment4::runCrossvalidation(int n, int folds) {
    
    double rightAverage;
    double SD;
    //int total;
    //int rightTotal;
    vector<double> rights;
    //double sizeAverage;
    
    chrono::high_resolution_clock::time_point t1;
    chrono::high_resolution_clock::time_point t2;
    long long duration;
    
    srand ((int)time(NULL));
    
    
    const int populationSize = 100;
    const float mutationRate = 1.0;
    const float crossoverRate = 0.0;
    const int parentSelection = 2;
    const int adultSelection = 2;
    const float spiceRatio = 0.1;
    const int k = 5;
    
    cout << "-- EMTI experiment run " << folds << " folds " << n << " times, on " << currentDateTime() << " : --\n";
    
    
    
    /* ------------- EMTI BASIC------------------------------------------------------------------------------*/
    
    int trainingRightTotal = 0;
    long trainingSetSizeTotal = 0;
    
    int total_PRE = 0;
    int rightTotal_PRE = 0;
    double sizeAverage_PRE = 0;
    vector<double> rights_PRE;
    
    int cycleCount = 0;
    vector<double> cycleCounts;
    
    int total_OPT = 0;
    int rightTotal_OPT = 0;
    double sizeAverage_OPT = 0;
    vector<double> rights_OPT;
    
    t1 = chrono::high_resolution_clock::now();
    
    
    for (int j = 0; j < n; j++) {
        
        data->randomizeDataOrder();
        vector<vector<unsigned>>* trainingSet;
        vector<vector<unsigned>>* testSet;
        
        for (int i = 0; i < folds; i++) {
            
            double testStart = i * (1 / (double)folds);
            double testEnd = (i + 1) * (1 / (double)folds);
            
            //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
            
            trainingSet = data->getNewTrainingSubset(0.0, testStart);
            vector<vector<unsigned>>* trainingSetTail = data->getNewTrainingSubset(testEnd, 1.0);
            
            trainingSet->insert(trainingSet->end(), trainingSetTail->begin(), trainingSetTail->end());
            
            testSet = data->getNewTrainingSubset(testStart, testEnd);
            
            vector<int>* fattrScope = data->getNewFullAttributeScope();
            
            string output = "/Users/avividi/Dropbox/fordypningsprosjekt_Anders_Fuglseth_2015/plotdata/plotData.dat";
            std::ofstream myfile;
            myfile.open (output);
            myfile.close();
            
            /*----INIT CLASSIFIER ------------------------------------------------------------*/
            EMTI_Population p(
                              fattrScope, //vector<int>* fullAttributeScope,
                              trainingSet, //vector<vector<int>* trainingSet
                              0, //int maxTreeSize (< 0 for unlimited)
                              populationSize, //int populationSize
                              1000, //int treeSize_fitnessWeight
                              mutationRate, //float mutationRate
                              crossoverRate, //float crossOverRateParameter
                              true, //bool rejectWorseGeneration
                              parentSelection, //int parentSelectionType
                              /*
                               0 = parent_selection_fitness_proportionate
                               1 = parent_selection_tournament
                               2 = parent_selection_sigma_scaling
                               */
                              adultSelection, //int adultSelectionType
                              /*
                               0 = adult_selection_full_generational_replacement
                               1 = adult_selection_overproduction
                               2 = adult_selection_generational_mixing
                               3 = adult_selection_add_spice(float spiceRatio)
                               */
                              false, //bool verbose
                              false, //bool verboseplus
                              false, //bool writeplotdata
                              output, //std::string plotOutputLocation
                              20, //int tournamentSizeParameter (only for tournament parent selection)
                              0.8, //float epsilonParameter (only for tournament parent selection)
                              spiceRatio //float spiceRatio (only for adult_selection_add_spice)
                              );
            
            
            
            while (p.adultBestIndividual->classificationAccuracy <= p.trainingSet->size()-9 && cycleCount < 1000) {
                p.do_evolution_cycle();
                
                cycleCount++;
                
            }
            cout << i + folds*j << ": cycleCount = " << cycleCount << "\n";
            
            cycleCounts.push_back(cycleCount);
            cycleCount = 0;
            
            trainingRightTotal += p.adultBestIndividual->classificationAccuracy;
            
            EMTI_TreeCreature* best = p.adultBestIndividual;
            
            trainingSetSizeTotal += trainingSet->size();
            
            /*--------------------------------------------------------------------------------*/
            
            int right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_PRE++;
                    }
                }
            }
            
            rights_PRE.push_back((double)right / testSet->size());
            total_PRE += testSet->size();
            sizeAverage_PRE += best->tree->nodesCount;
            
            /*--------------------------------------------------------------------------------*/
            
            p.doPostCycleEndNodeOptimizations(k, false);
            
            right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_OPT++;
                    }
                }
            }
            
            rights_OPT.push_back((double)right / testSet->size());
            total_OPT += testSet->size();
            sizeAverage_OPT += best->tree->nodesCount;
            
            delete fattrScope;
            delete testSet;
            delete trainingSet;
            delete trainingSetTail;
            
        }
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_PRE[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_PRE[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights_PRE.clear();
    
    double averageCycles = 0;
    
    for (int i = 0; i < folds * n; i++) {
        averageCycles += cycleCounts[i];
    }
    averageCycles = averageCycles / (folds * n);
    
    double SDcycles = 0;
    for (int i = 0; i < folds * n; i++) {
        SDcycles += pow(cycleCounts[i] - averageCycles, 2);
    }
    SDcycles = sqrt(SDcycles/(folds * n));
    
    cout << "Average cycle count = " << averageCycles << " SD = " << SDcycles << "\n\n";
    
    cout
    << "EMTI_Basic_PRE: " << rightTotal_PRE << "/" << total_PRE << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_PRE / ((double)total_PRE / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_PRE /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal / trainingSetSizeTotal << ")\n"
    << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_OPT[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_OPT[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights_OPT.clear();
    
    
    cout
    << "EMTI_Basic_OPT: " << rightTotal_OPT << "/" << total_OPT << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_OPT / ((double)total_OPT / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_OPT /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal / trainingSetSizeTotal << ")\n"
    << "\n";
    
    cout << "      (Duration=" << duration << " microseconds " << folds << " folds " << n << " times)\n"
    << " popSize=" << populationSize << " mutation=" << mutationRate << " crossover=" << crossoverRate << " \nparentSelection=" << parentSelection << " adultSelection=" << adultSelection << " spice=" << spiceRatio << " k=" << k << ")\n";
    
    
}

// Get current date/time, format is YYYY-MM-DD.HH:mm:ss
const std::string Experiment4::currentDateTime() {
    time_t     now = time(0);
    struct tm  tstruct;
    char       buf[80];
    tstruct = *localtime(&now);
    strftime(buf, sizeof(buf), "%Y-%m-%d.%X", &tstruct);
    
    return buf;
}








