//
//  ExperimentSuite5_pruning.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth 2016.

#include "ExperimentSuite5_pruning.hpp"

Experiment5::Experiment5() {
    this->traindata = new EMTI_Data("/Users/avividi/monks/monks-3.train.txt", ' ', {7}, 0);
    
    
    this->testdata = new EMTI_Data("/Users/avividi/monks/monks-3.test.txt",
                                   traindata->attributeValues_int,
                                   traindata->attributeValues_intToString,
                                   traindata->attributeValues_stringToInt,
                                   ' ', {7}, 0);
    
    
}

void Experiment5::run(int n) {
    
    double rightAverage;
    double SD;
    int total;
    int rightTotal;
    vector<double> rights;
    double sizeAverage;
    
    chrono::high_resolution_clock::time_point t1;
    chrono::high_resolution_clock::time_point t2;
    long long duration;
    
    srand ((int)time(NULL));
    
    cout << "-- EMTI experiment run " << n << " times, on " << currentDateTime() << " : --\n";
    
    vector<vector<unsigned>>* trainingSet;
    vector<vector<unsigned>>* testSet;
    
    //    traindata->randomizeDataOrder();
    //    testdata->randomizeDataOrder();
    
    trainingSet = traindata->dataIntVector;
    testSet = testdata->dataIntVector;
    
    vector<int>* fattrScope = traindata->getNewFullAttributeScope();
    
    /* ------------- KNN ----------------------------------------------------------------------------------*/
    
    for (int k = 1; k < 20; k+=2) {
        total = 0;
        rightTotal = 0;
        
        t1 = chrono::high_resolution_clock::now();
        
        for (int j = 0; j < n; j++) {
            
            //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
            
            /*----INIT CLASSIFIER ------------------------------------------------------------*/
            EMTI_KNearestClassifier kNN(trainingSet);
            /*--------------------------------------------------------------------------------*/
            
            int right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (kNN.classify(&(*testSet)[i], k, true) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal++;
                    }
                }
            }
            rights.push_back((double)right / testSet->size());
            total += testSet->size();
            //sizeAverage += ID3.tree->nodesCount;
            
            cout << ".";
        }
        
        t2 = chrono::high_resolution_clock::now();
        duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
        
        cout << "\n";
        
        rightAverage = 0;
        
        for (int i = 0; i < n; i++) {
            rightAverage += rights[i];
        }
        rightAverage = rightAverage / (n);
        
        SD = 0;
        for (int i = 0; i < n; i++) {
            SD += pow(rights[i] - rightAverage, 2);
        }
        
        SD = sqrt(SD/(n));
        
        rights.clear();
        
        
        cout << k << "NN: " << rightTotal << "/" << total << " avg=" << rightAverage << " SD=" << SD << "\n"
        << "      (" << (double) rightTotal / ((double)total / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
        << "      (avg.size=N/A)\n"
        << "      (Duration=" << duration << " microseconds " << n << " times)\n";
    }
    
    /* ------------- ID3 ----------------------------------------------------------------------------------*/
    total = 0;
    rightTotal = 0;
    
    t1 = chrono::high_resolution_clock::now();
    
    for (int j = 0; j < n; j++) {
        
        
        /*----INIT CLASSIFIER ------------------------------------------------------------*/
        EMTI_ID3Classifier ID3(trainingSet, fattrScope);
        ID3.makeTree();
        /*--------------------------------------------------------------------------------*/
        
        int right = 0;
        for (int i = 0; i < testSet->size(); i++) {
            if ((*testSet)[i].size() > 0 ) {
                
                if (ID3.tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                    right++;
                    rightTotal++;
                }
            }
        }
        rights.push_back((double)right / testSet->size());
        total += testSet->size();
        sizeAverage += ID3.tree->nodesCount;
        
        cout << ".";
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < n; i++) {
        rightAverage += rights[i];
    }
    rightAverage = rightAverage / (n);
    
    SD = 0;
    for (int i = 0; i < n; i++) {
        SD += pow(rights[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(n));
    
    rights.clear();
    
    
    cout << "ID3: " << rightTotal << "/" << total << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal / ((double)total / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage /(n) << ")\n"
    << "      (Duration=" << duration << " microseconds " << n << " times)\n";
    
    
    /* ------------- EMTI ----------------------------------------------------------------------------------*/
    const int populationSize = 100;
    const int cycles = 1000;
    const float mutationRate = 1.0;
    const float crossoverRate = 1.0;
    const int parentSelection = 2;
    const int adultSelection = 2;
    const float spiceRatio = 0.1;
    const int k = 3;
    
    int trainingRightTotal = 0;
    long trainingSetSizeTotal = 0;
    total = 0;
    rightTotal = 0;
    sizeAverage = 0;
    
    t1 = chrono::high_resolution_clock::now();
    
    const int feedbackTimeNumber = (cycles * n) / 60;
    int feedbackIncrement = 0;
//    cout << "|";
//    for (int f = 0; f < 58; f++) {
//        cout << " ";
//    }
//    cout << "|\n";
//    
//    for (int j = 0; j < n; j++) {
//        
//        
//        //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
//        
//        string output = "/Users/avividi/Google Drive/masteroppgave/plotdata/plotData.dat";
//        std::ofstream myfile;
//        myfile.open (output);
//        myfile.close();
//        
//        /*----INIT CLASSIFIER ------------------------------------------------------------*/
//        EMTI_Population p(
//                          fattrScope, //vector<int>* fullAttributeScope,
//                          trainingSet, //vector<vector<int>* trainingSet
//                          0, //int maxTreeSize (< 0 for unlimited)
//                          populationSize, //int populationSize
//                          1000, //int treeSize_fitnessWeight
//                          mutationRate, //float mutationRate
//                          crossoverRate, //float crossOverRateParameter
//                          true, //bool rejectWorseGeneration
//                          parentSelection, //int parentSelectionType
//                          /*
//                           0 = parent_selection_fitness_proportionate
//                           1 = parent_selection_tournament
//                           2 = parent_selection_sigma_scaling
//                           */
//                          adultSelection, //int adultSelectionType
//                          /*
//                           0 = adult_selection_full_generational_replacement
//                           1 = adult_selection_overproduction
//                           2 = adult_selection_generational_mixing
//                           3 = adult_selection_add_spice(float spiceRatio)
//                           */
//                          false, //bool verbose
//                          false, //bool verboseplus
//                          false, //bool writeplotdata
//                          output, //std::string plotOutputLocation
//                          20, //int tournamentSizeParameter (only for tournament parent selection)
//                          0.8, //float epsilonParameter (only for tournament parent selection)
//                          spiceRatio //float spiceRatio (only for adult_selection_add_spice)
//                          );
//        
//        
//        while (p.cycleCount < cycles) {
//            p.do_evolution_cycle();
//            
//            feedbackIncrement++;
//            if (feedbackIncrement>=feedbackTimeNumber) {
//                feedbackIncrement = 0;
//                cout << "-";
//            }
//        }
//        
//        p.doPostCycleEndNodeOptimizations(k, false);
//        
//        //p.doPostCycle_rollBackCheck();
//        
//        trainingRightTotal += p.adultBestIndividual->classificationAccuracy;
//        
//        EMTI_TreeCreature* best = p.adultBestIndividual;
//        trainingSetSizeTotal += p.trainingSet->size();
//        /*--------------------------------------------------------------------------------*/
//        
//        int right = 0;
//        for (int i = 0; i < testSet->size(); i++) {
//            if ((*testSet)[i].size() > 0 ) {
//                if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
//                    right++;
//                    rightTotal++;
//                }
//            }
//        }
//        
//        rights.push_back((double)right / testSet->size());
//        total += testSet->size();
//        sizeAverage += best->tree->nodesCount;
//        
//        
//        //cout << ".";
//    }
//    
//    t2 = chrono::high_resolution_clock::now();
//    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
//    
//    cout << "\n";
//    
//    rightAverage = 0;
//    
//    for (int i = 0; i < n; i++) {
//        rightAverage += rights[i];
//    }
//    rightAverage = rightAverage / (n);
//    
//    SD = 0;
//    for (int i = 0; i <  n; i++) {
//        SD += pow(rights[i] - rightAverage, 2);
//    }
//    
//    SD = sqrt(SD/( n));
//    
//    rights.clear();
//    
//    
//    cout << "EMTI: " << rightTotal << "/" << total << " avg=" << rightAverage << " SD=" << SD << "\n"
//    << "      (" << (double) rightTotal / ((double)total / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
//    << "      (avg.size=" << sizeAverage /( n) << ")\n"
//    << "      (Duration=" << duration << " microseconds " << n << " times)\n"
//    << "      (trainingaccuracy reached=" << (double)trainingRightTotal / trainingSetSizeTotal << ")\n"
//    << "      (Params: cycles=" << cycles << " popSize=" << populationSize << " mutation=" << mutationRate << " crossover=" << crossoverRate << " \nparentSelection=" << parentSelection << " adultSelection=" << adultSelection << " spice=" << spiceRatio << " k=" << k << ")\n";
    
    
    /* ------------- EMTI pruning----------------------------------------------------------------------------*/
    
    t1 = chrono::high_resolution_clock::now();
    
    feedbackIncrement = 0;
    cout << "|";
    for (int f = 0; f < 58; f++) {
        cout << " ";
    }
    cout << "|\n";
    
    int total_PRE = 0;
    int rightTotal_PRE = 0;
    double sizeAverage_PRE = 0;
    vector<double> rights_PRE;
    
    int total_OPT = 0;
    int rightTotal_OPT = 0;
    double sizeAverage_OPT = 0;
    vector<double> rights_OPT;
    
    for (int j = 0; j < n; j++) {
        
        
        //traindata->randomizeDataOrder();
        
        //trainingSet = traindata->getNewTrainingSubset(0.0, 0.66);
        //vector<vector<unsigned>>* validationSet = traindata->getNewTrainingSubset(0.66, 1.0);
        
        string output = "/Users/avividi/Dropbox/fordypningsprosjekt_Anders_Fuglseth_2015/plotdata/plotData.dat";
        std::ofstream myfile;
        myfile.open (output);
        myfile.close();
        
        /*----INIT CLASSIFIER ------------------------------------------------------------*/
        EMTI_Population p(
                          fattrScope, //vector<int>* fullAttributeScope,
                          trainingSet, //vector<vector<int>* trainingSet
                          0, //int maxTreeSize (< 0 for unlimited)
                          populationSize, //int populationSize
                          1000, //int treeSize_fitnessWeight
                          mutationRate, //float mutationRate
                          crossoverRate, //float crossOverRateParameter
                          true, //bool rejectWorseGeneration
                          parentSelection, //int parentSelectionType
                          /*
                           0 = parent_selection_fitness_proportionate
                           1 = parent_selection_tournament
                           2 = parent_selection_sigma_scaling
                           */
                          adultSelection, //int adultSelectionType
                          /*
                           0 = adult_selection_full_generational_replacement
                           1 = adult_selection_overproduction
                           2 = adult_selection_generational_mixing
                           3 = adult_selection_add_spice(float spiceRatio)
                           */
                          false, //bool verbose
                          false, //bool verboseplus
                          false, //bool writeplotdata
                          output, //std::string plotOutputLocation
                          20, //int tournamentSizeParameter (only for tournament parent selection)
                          0.8, //float epsilonParameter (only for tournament parent selection)
                          spiceRatio //float spiceRatio (only for adult_selection_add_spice)
                          );
        
        
        
        while (p.cycleCount < cycles) {
            p.do_evolution_cycle();
            
            feedbackIncrement++;
            if (feedbackIncrement>=feedbackTimeNumber) {
                feedbackIncrement = 0;
                cout << "-";
            }
        }
        
        trainingRightTotal += p.adultBestIndividual->classificationAccuracy;
        
        EMTI_TreeCreature* best = p.adultBestIndividual;
        
        trainingSetSizeTotal += trainingSet->size();
        
        /*--------------------------------------------------------------------------------*/
        
        int right = 0;
        for (int i = 0; i < testSet->size(); i++) {
            if ((*testSet)[i].size() > 0 ) {
                if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                    right++;
                    rightTotal_PRE++;
                }
            }
        }
        
        rights_PRE.push_back((double)right / testSet->size());
        total_PRE += testSet->size();
        sizeAverage_PRE += best->tree->nodesCount;
        
        /*--------------------------------------------------------------------------------*/
        
        p.doPostCycleEndNodeOptimizations(k, false);
        
        best->tree = EMTI_TreePruner::minimum_error_prune(best->tree, trainingSet);
        
        right = 0;
        for (int i = 0; i < testSet->size(); i++) {
            if ((*testSet)[i].size() > 0 ) {
                if (best->tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                    right++;
                    rightTotal_OPT++;
                }
            }
        }
        
        rights_OPT.push_back((double)right / testSet->size());
        total_OPT += testSet->size();
        sizeAverage_OPT += best->tree->nodesCount;
        
        //delete trainingSet;
        //delete validationSet;
        
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < n; i++) {
        rightAverage += rights_PRE[i];
    }
    rightAverage = rightAverage / (n);
    
    SD = 0;
    for (int i = 0; i < n; i++) {
        SD += pow(rights_PRE[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(n));
    
    rights_PRE.clear();
    
    
    cout
    << "EMTI_Basic_PRE: " << rightTotal_PRE << "/" << total_PRE << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_PRE / ((double)total_PRE / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_PRE /(n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal / trainingSetSizeTotal << ")\n"
    << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < n; i++) {
        rightAverage += rights_OPT[i];
    }
    rightAverage = rightAverage / (n);
    
    SD = 0;
    for (int i = 0; i < n; i++) {
        SD += pow(rights_OPT[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(n));
    
    rights_OPT.clear();
    
    
    cout
    << "EMTI_Basic_OPT: " << rightTotal_OPT << "/" << total_OPT << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_OPT / ((double)total_OPT / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_OPT /(n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal / trainingSetSizeTotal << ")\n"
    << "\n";
    
    cout << "      (Duration=" << duration << " microseconds " << n << " times)\n"
    << "      (Params: cycles=" << cycles << " popSize=" << populationSize << " mutation=" << mutationRate << " crossover=" << crossoverRate << " \nparentSelection=" << parentSelection << " adultSelection=" << adultSelection << " spice=" << spiceRatio << " k=" << k << ")\n";
    
    
    delete fattrScope;
    
}

// Get current date/time, format is YYYY-MM-DD.HH:mm:ss
const std::string Experiment5::currentDateTime() {
    time_t     now = time(0);
    struct tm  tstruct;
    char       buf[80];
    tstruct = *localtime(&now);
    strftime(buf, sizeof(buf), "%Y-%m-%d.%X", &tstruct);
    
    return buf;
}








