//
//  ExperimentSuite7.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth 2016.

#include "ExperimentSuite7.hpp"

Experiment7::Experiment7(EMTI_Data* d) {
    this->data = d;
}

void Experiment7::runCrossvalidation(int n, int folds) {
    
    double rightAverage;
    double SD;
    int total;
    int rightTotal;
    vector<double> rights;
    double sizeAverage;
    
    chrono::high_resolution_clock::time_point t1;
    chrono::high_resolution_clock::time_point t2;
    long long duration;
    
    srand ((int)time(NULL));
    
    
    const int populationSize = 100;
    const int cycles = 500;
    const float mutationRate = 1.0;
    const float crossoverRate = 1.0;
    const int parentSelection = 2;
    const int adultSelection = 2;
    const float spiceRatio = 0.1;
    const int k = 5;
    
    cout << "-- EMTI experiment run " << folds << " folds " << n << " times, on " << currentDateTime() << " : --\n";
    
    /* ------------- ID3 ----------------------------------------------------------------------------------*/
    total = 0;
    rightTotal = 0;
    
    t1 = chrono::high_resolution_clock::now();
    
    for (int j = 0; j < n; j++) {
        
        data->randomizeDataOrder();
        vector<vector<unsigned>>* trainingSet;
        vector<vector<unsigned>>* testSet;
        
        for (int i = 0; i < folds; i++) {
            
            double testStart = i * (1 / (double)folds);
            double testEnd = (i + 1) * (1 / (double)folds);
            
            //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
            
            trainingSet = data->getNewTrainingSubset(0.0, testStart);
            vector<vector<unsigned>>* trainingSetTail = data->getNewTrainingSubset(testEnd, 1.0);
            
            trainingSet->insert(trainingSet->end(), trainingSetTail->begin(), trainingSetTail->end());
            
            testSet = data->getNewTrainingSubset(testStart, testEnd);
            
            vector<int>* fattrScope = data->getNewFullAttributeScope();
            
            /*----INIT CLASSIFIER ------------------------------------------------------------*/
            EMTI_ID3Classifier ID3(trainingSet, fattrScope);
            ID3.makeTree();
            /*--------------------------------------------------------------------------------*/
            
            int right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (ID3.tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal++;
                    }
                }
            }
            rights.push_back((double)right / testSet->size());
            total += testSet->size();
            sizeAverage += ID3.tree->nodesCount;
            
            delete fattrScope;
            delete testSet;
            delete trainingSet;
            delete trainingSetTail;
            
            cout << ".";
        }
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights.clear();
    
    
    cout << "ID3: " << rightTotal << "/" << total << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal / ((double)total / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage /(folds * n) << ")\n"
    << "      (Duration=" << duration << " microseconds " << folds << " folds " << n << " times)\n";
    
    
    
    /* ------------- KNN ----------------------------------------------------------------------------------*/
    
    for (int k = 1; k < 20; k+=2) {
        total = 0;
        rightTotal = 0;
        
        t1 = chrono::high_resolution_clock::now();
        
        for (int j = 0; j < n; j++) {
            
            data->randomizeDataOrder();
            vector<vector<unsigned>>* trainingSet;
            vector<vector<unsigned>>* testSet;
            vector<vector<unsigned>>* testSetNoCategory;
            
            for (int i = 0; i < folds; i++) {
                
                double testStart = i * (1 / (double)folds);
                double testEnd = (i + 1) * (1 / (double)folds);
                
                //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
                
                trainingSet = data->getNewTrainingSubset(0.0, testStart);
                vector<vector<unsigned>>* trainingSetTail = data->getNewTrainingSubset(testEnd, 1.0);
                
                trainingSet->insert(trainingSet->end(), trainingSetTail->begin(), trainingSetTail->end());
                
                testSet = data->getNewTrainingSubset(testStart, testEnd);
                testSetNoCategory = data->getNewTestSubset(testStart, testEnd);
                
                vector<int>* fattrScope = data->getNewFullAttributeScope();
                
                /*----INIT CLASSIFIER ------------------------------------------------------------*/
                EMTI_KNearestClassifier kNN(trainingSet);
                /*--------------------------------------------------------------------------------*/
                
                int right = 0;
                for (int i = 0; i < testSet->size(); i++) {
                    if ((*testSet)[i].size() > 0 ) {
                        if (kNN.classify(&(*testSetNoCategory)[i], k) == (*testSet)[i][(*testSet)[i].size()-1]) {
                            right++;
                            rightTotal++;
                        }
                    }
                }
                rights.push_back((double)right / testSet->size());
                total += testSet->size();
                //sizeAverage += ID3.tree->nodesCount;
                
                delete fattrScope;
                delete testSet;
                delete testSetNoCategory;
                delete trainingSet;
                delete trainingSetTail;
                
                cout << ".";
            }
        }
        
        t2 = chrono::high_resolution_clock::now();
        duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
        
        cout << "\n";
        
        rightAverage = 0;
        
        for (int i = 0; i < folds * n; i++) {
            rightAverage += rights[i];
        }
        rightAverage = rightAverage / (folds * n);
        
        SD = 0;
        for (int i = 0; i < folds * n; i++) {
            SD += pow(rights[i] - rightAverage, 2);
        }
        
        SD = sqrt(SD/(folds * n));
        
        rights.clear();
        
        
        cout << k << "NN: " << rightTotal << "/" << total << " avg=" << rightAverage << " SD=" << SD << "\n"
        << "      (" << (double) rightTotal / ((double)total / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
        << "      (avg.size=N/A)\n"
        << "      (Duration=" << duration << " microseconds " << folds << " folds " << n << " times)\n";
    }
    
    
    
    
    /* ------------- EMTI BASIC------------------------------------------------------------------------------*/
    
    int trainingRightTotal = 0;
    long trainingSetSizeTotal = 0;
    
    int total_OPT = 0;
    int rightTotal_OPT = 0;
    double sizeAverage_OPT = 0;
    vector<double> rights_OPT;
    
    int trainingRightTotal_PRUN1 = 0;
    
    int total_PRUN1 = 0;
    int rightTotal_PRUN1 = 0;
    double sizeAverage_PRUN1 = 0;
    vector<double> rights_PRUN1;
    
    t1 = chrono::high_resolution_clock::now();
    
    for (int j = 0; j < n; j++) {
        
        data->randomizeDataOrder();
        vector<vector<unsigned>>* trainingSet;
        vector<vector<unsigned>>* testSet;
        
        for (int i = 0; i < folds; i++) {
            
            double testStart = i * (1 / (double)folds);
            double testEnd = (i + 1) * (1 / (double)folds);
            
            //cout << i << " t1=" << testStart << " t2=" << testEnd << "\n";
            
            trainingSet = data->getNewTrainingSubset(0.0, testStart);
            vector<vector<unsigned>>* trainingSetTail = data->getNewTrainingSubset(testEnd, 1.0);
            
            trainingSet->insert(trainingSet->end(), trainingSetTail->begin(), trainingSetTail->end());
            
            testSet = data->getNewTrainingSubset(testStart, testEnd);
            
            vector<int>* fattrScope = data->getNewFullAttributeScope();
            
            string output = "/Users/avividi/Dropbox/fordypningsprosjekt_Anders_Fuglseth_2015/plotdata/plotData.dat";
            std::ofstream myfile;
            myfile.open (output);
            myfile.close();
            
            EMTI_DecisionTree* tree = EMTI_autoRun::autoEMTI (trainingSet, fattrScope);
            
            /*----INIT CLASSIFIER ------------------------------------------------------------*/
            
            
            int trainingRight = 0;
            for (int i = 0; i < trainingSet->size(); i++) {
                if ((*trainingSet)[i].size() > 0 ) {
                    if (tree->decide((*trainingSet)[i]) == (*trainingSet)[i][(*trainingSet)[i].size()-1]) {
                        trainingRight++;
                        trainingRightTotal++;
                    }
                }
            }
            trainingSetSizeTotal += trainingSet->size();
            
            /*--------------------------------------------------------------------------------*/
            
            int right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_OPT++;
                    }
                }
            }
            
            rights_OPT.push_back((double)right / testSet->size());
            total_OPT += testSet->size();
            sizeAverage_OPT += tree->nodesCount;
            
            /*--------------------------------------------------------------------------------*/
            
            tree = EMTI_TreePruner::minimum_error_prune(tree, trainingSet);
            
            trainingRight = 0;
            for (int i = 0; i < trainingSet->size(); i++) {
                if ((*trainingSet)[i].size() > 0 ) {
                    if (tree->decide((*trainingSet)[i]) == (*trainingSet)[i][(*trainingSet)[i].size()-1]) {
                        trainingRight++;
                        trainingRightTotal_PRUN1++;
                    }
                }
            }
            
            right = 0;
            for (int i = 0; i < testSet->size(); i++) {
                if ((*testSet)[i].size() > 0 ) {
                    if (tree->decide((*testSet)[i]) == (*testSet)[i][(*testSet)[i].size()-1]) {
                        right++;
                        rightTotal_PRUN1++;
                    }
                }
            }
            
            rights_PRUN1.push_back((double)right / testSet->size());
            total_PRUN1 += testSet->size();
            sizeAverage_PRUN1 += tree->nodesCount;
            
            
            delete fattrScope;
            delete testSet;
            delete trainingSet;
            delete trainingSetTail;
            
            cout << ".";
        }
    }
    
    t2 = chrono::high_resolution_clock::now();
    duration = chrono::duration_cast<chrono::microseconds>( t2 - t1 ).count();
    
    cout << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_OPT[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_OPT[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights_OPT.clear();
    
    
    cout
    << "EMTI_Basic_OPT: " << rightTotal_OPT << "/" << total_OPT << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_OPT / ((double)total_OPT / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_OPT /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal / trainingSetSizeTotal << ")\n"
    << "\n";
    
    rightAverage = 0;
    
    for (int i = 0; i < folds * n; i++) {
        rightAverage += rights_PRUN1[i];
    }
    rightAverage = rightAverage / (folds * n);
    
    SD = 0;
    for (int i = 0; i < folds * n; i++) {
        SD += pow(rights_PRUN1[i] - rightAverage, 2);
    }
    
    SD = sqrt(SD/(folds * n));
    
    rights_PRUN1.clear();
    
    
    cout
    << "EMTI_Basic_Min-Err-pruned: " << rightTotal_PRUN1 << "/" << total_PRUN1 << " avg=" << rightAverage << " SD=" << SD << "\n"
    << "      (" << (double) rightTotal_PRUN1 / ((double)total_PRUN1 / 100.0) << "\% SD=" << SD * 100.0 << "\%)\n"
    << "      (avg.size=" << sizeAverage_PRUN1 /(folds * n) << ")\n"
    << "      (trainingaccuracy reached=" << (double)trainingRightTotal_PRUN1 / trainingSetSizeTotal << ")\n"
    << "\n";
    
    cout << "      (Duration=" << duration << " microseconds " << folds << " folds " << n << " times)\n"
    << "      (Params: cycles=" << cycles << " popSize=" << populationSize << " mutation=" << mutationRate << " crossover=" << crossoverRate << " \nparentSelection=" << parentSelection << " adultSelection=" << adultSelection << " spice=" << spiceRatio << " k=" << k << ")\n";


}

// Get current date/time, format is YYYY-MM-DD.HH:mm:ss
const std::string Experiment7::currentDateTime() {
    time_t     now = time(0);
    struct tm  tstruct;
    char       buf[80];
    tstruct = *localtime(&now);
    strftime(buf, sizeof(buf), "%Y-%m-%d.%X", &tstruct);
    
    return buf;
}








