//
//  EMTI_AdultSelection.cpp
//  EMTI0.2_OSX
//
//  Created by Anders Nikolai Fuglseth on 04.02.2016.
//

#include "EMTI_AdultSelection.hpp"

vector<EMTI_TreeCreature*>* EMTI_AdultSelection::full_generational_replacement(vector<EMTI_TreeCreature*>* children, vector<EMTI_TreeCreature*>* adults) {
    
    for (EMTI_TreeCreature* adult : *adults) {
        delete adult;
    }
    adults->clear();
    delete adults;
    
    return children;
    
}

vector<EMTI_TreeCreature*>* EMTI_AdultSelection::overproduction (vector<EMTI_TreeCreature*>* children, vector<EMTI_TreeCreature*>* adults, int populationSize) {
    
    std::sort(children->begin(), children->end(), compareFitness);
    
    for (vector<EMTI_TreeCreature*>::iterator it = children->begin()+ populationSize ; it != children->end(); ++it) {
        delete *it;
    }
    children->erase(children->begin() + populationSize, children->end());
    
    return EMTI_AdultSelection::full_generational_replacement(children, adults);
    
}

vector<EMTI_TreeCreature*>* EMTI_AdultSelection::generational_mixing (vector<EMTI_TreeCreature*>* children, vector<EMTI_TreeCreature*>* adults) {
    
    long popSize = children->size();
    
    children->insert(children->end(), adults->begin(), adults->end());
    std::sort(children->begin(), children->end(), compareFitness);
    
    for (vector<EMTI_TreeCreature*>::iterator it = children->begin()+ popSize ; it != children->end(); ++it) {
        delete *it;
    }
    children->erase(children->begin() + popSize, children->end());
    
    adults->clear();
    delete adults;
    
    return children;
}

vector<EMTI_TreeCreature*>* EMTI_AdultSelection::generational_mixing_add_spice (vector<EMTI_TreeCreature*>* children, vector<EMTI_TreeCreature*>* adults, float spiceRatio, vector<int>* fullAttributeScope, vector<vector<unsigned>>* trainingSet, int sizeWeight) {
    
    long popSize = children->size();
    long cutOff = popSize * (1-spiceRatio);
    
    children->insert(children->end(), adults->begin(), adults->end());
    std::sort(children->begin(), children->end(), compareFitness);
    
    for (vector<EMTI_TreeCreature*>::iterator it = children->begin()+ cutOff ; it != children->end(); ++it) {
        delete *it;
    }
    children->erase(children->begin() + cutOff, children->end());
    
        while (children->size() < popSize) {
            EMTI_TreeCreature* ind = new EMTI_TreeCreature(fullAttributeScope);
            ind->createRandomTree();
            ind->assignFitness(trainingSet, sizeWeight);
            children->push_back(ind);
        }
    adults->clear();
    delete adults;
    
    return children;
}

bool EMTI_AdultSelection::compareFitness (EMTI_TreeCreature* a, EMTI_TreeCreature* b) {
    return a->getFitness() > b->getFitness();
}