//
//  EMTI_Population_Wrapper.m
//  EMTI_0.2_swift
//
//  Created by 2016 Anders Nikolai Fuglseth on 09.02.2016.
//

#import "EMTI_Swift_Wrapper.h"

#import "EMTI_Data.hpp"
#import "EMTI_DecisionTree.hpp"
#import "EMTI_TreeCreature.hpp"
#import "EMTI_Population.hpp"
#import "EMTI_KNearestClassifier.hpp"
#import "EMTI_ID3Classifier.hpp"

#import "EMTI_PopulationES.hpp"
#import "EMTI_Population_ID3PopInit.hpp"

#import "ExperimentSuite1.hpp"
#import "ExperimentSuite2.hpp"
#import "ExperimentSuite3.hpp"
#import "ExperimentSuite4.hpp"
#import "ExperimentSuite5_pruning.hpp"
#import "ExperimentSuite6_pruning.hpp"
#import "ExperimentSuite7.hpp"
#import "EMTI_PopulationES1.hpp"
#import "EMTI_PopulationES2.hpp"
#import "EMTI_PopulationES3.hpp"

#import "EMTI_autoRun.hpp"
#import "EMTI_TreePruner.hpp"

#import "EMTI_DotTreeWrapper.hpp"


#include <sstream>

@implementation EMTI_Swift_Wrapper {
    
@private
    EMTI_Population* population;
    
    EMTI_Data* data;
    EMTI_TreeCreature* bestTree;
    vector<vector<unsigned>> *trainingData;
    vector<vector<unsigned>>* validationData;
    vector<vector<unsigned>> *testDataNoCategory;
    vector<vector<unsigned>> *testData;
}

- (void)data_init:(const char*) filePath {
    
    if (data != NULL) delete data;
    
    data = new EMTI_Data(filePath);
//    vector<int> s;
//    s.push_back(0);
//    data = new EMTI_Data(filePath, ',', s);
//    data->saveToFile("");
    
    _dataAttributesCountText = [NSString stringWithFormat:@"%ld",data->attributes];
    _dataInstancesCountText = [NSString stringWithFormat:@"%ld",data->instances];
    _dataAttributeValuesText = [NSString stringWithUTF8String:data->attributeValues_to_string().c_str()];
    
    _cyclecount = 0;
}

- (bool)populationInit_populationSize:(int)psize sizeWeight:(int)sW mutationRate:(float)mR crossoverRate: (float)cR adultSelectionType: (int)aT parentSelectionType: (int)pT rejectWorseGeneration: (int)rwg spiceRation: (float)sR tournamentSize: (int)tsize epsilonParameter: (float)eP trainingSetSize: (float) trainingSetSize

{
    srand ((int)time(NULL));
    //srand (24101990);
    
    if (data == NULL || data->dataIntVector->size() == 0) return false;
    
    delete bestTree;
    
    data->randomizeDataOrder();
    
    trainingData = data->getNewTrainingSubset(0.0, trainingSetSize);
    validationData = data->getNewTrainingSubset(0.0, 0.0);
    testDataNoCategory = data->getNewTestSubset(trainingSetSize, 1.0);
    testData = data->getNewTrainingSubset(trainingSetSize, 1.0);
    
    
//    trainingData = data->getNewTrainingSubset(0.0, 0.6);
//    validationData = data->getNewTrainingSubset(0.6, 0.9);
//    testDataNoCategory = data->getNewTestSubset(0.9, 1.0);
//    testData = data->getNewTrainingSubset(0.9, 1.0);
    
//        trainingData = data->getNewTrainingSubset(0, 86);
//        validationData = data->getNewTrainingSubset(87, 123);
//        testDataNoCategory = data->getNewTestSubset(123, (int)data->dataIntVector->size()-1);
//        testData = data->getNewTrainingSubset(123, (int)data->dataIntVector->size()-1);
    
//    trainingData = data->getNewTrainingSubset(0.0, 1.0);
//    validationData = data->getNewTrainingSubset(0.0, 0.0);
//    testDataNoCategory = data->getNewTestSubset(0.0, 1.0);
//    testData = data->getNewTrainingSubset(0.0, 1.0);
    
    
    vector<int>* fAttrScope = data->getNewFullAttributeScope();
    
    const char* s = [[NSBundle mainBundle] bundlePath].fileSystemRepresentation;
    
    string output = string(s) + "/../plotData.dat";
    std::ofstream myfile;
    myfile.open (output);
    myfile.close();
    
    population = new EMTI_Population(
                                     fAttrScope, //vector<int>* fullAttributeScope,
                                     trainingData, //vector<vector<int>* trainingSet
                                     0, //int maxTreeSize (<= 0 for unlimited)
                                     psize, //int populationSize
                                     sW, //int treeSize_fitnessWeight
                                     mR, //float mutationRate
                                     cR, //float crossOverRateParameter
                                     rwg, //bool rejectWorseGeneration
                                     pT, //int parentSelectionType
                                     /*
                                      0 = parent_selection_fitness_proportionate
                                      1 = parent_selection_tournament
                                      2 = parent_selection_sigma_scaling
                                      */
                                     aT, //int adultSelectionType
                                     /*
                                      0 = adult_selection_full_generational_replacement
                                      1 = adult_selection_overproduction
                                      2 = adult_selection_generational_mixing
                                      3 = adult_selection_add_spice(float spiceRatio)
                                      */
                                     false, //bool verbose
                                     false, //bool verboseplus
                                     true, //bool writeplotdata
                                     output, //std::string plotOutputLocation
                                     tsize, //int tournamentSizeParameter (only for tournament parent selection)
                                     eP, //float epsilonParameter (only for tournament parent selection)
                                     sR //float spiceRatio (only for adult_selection_add_spice)
                                     );
    
    delete fAttrScope;
    
    return true;
}

- (void)performanceFeedbackOutput {
    
    //*ES
//    bool postCycle_rollBack = population->doPostCycle_rollBackCheck();
//    cout << "Rollback count = " << population->rollBackCount << ", postCycleRollBack = " << postCycle_rollBack << "\n";
    //ES*
    
    
    vector<int>* fAttrScope = data->getNewFullAttributeScope();
    std::stringstream resultStream;
    
    //p.giveGenerationInfoOutput(true, true, false);
    population->doPostCycleEndNodeOptimizations(/*k*/3, /*verbose*/false);
    
    EMTI_TreeCreature* best = population->adultBestIndividual;
    int right = 0;
    int wrong = 0;
    for (int i = 0; i < testData->size(); i++) {
        if ((*testData)[i].size() > 0 ) {
            if (best->tree->decide((*testDataNoCategory)[i]) == (*testData)[i][(*testData)[i].size()-1]) {
                right++;
            }
            else wrong ++;
        }
    }
    resultStream << "EMTI OPT: " << right << "/" << testData->size()  << " (" << (float) right / ((float)testData->size() / 100.0) << "\%)"
    << " (size=" << best->tree->nodesCount << ")  ";
    
    //best->tree = EMTI_TreePruner::error_complexity_prune(best->tree, trainingData, validationData);
    best->tree = EMTI_TreePruner::minimum_error_prune(best->tree, trainingData);
    
    right = 0;
    wrong = 0;
    for (int i = 0; i < testData->size(); i++) {
        if ((*testData)[i].size() > 0 ) {
            if (best->tree->decide((*testDataNoCategory)[i]) == (*testData)[i][(*testData)[i].size()-1]) {
                right++;
            }
            else wrong ++;
        }
    }
    resultStream << "Pruned: " << right << "/" << testData->size()  << " (" << (float) right / ((float)testData->size() / 100.0) << "\%)"
    << " (size=" << best->tree->nodesCount << ")\n";
    
    
    
    EMTI_ID3Classifier ID3(trainingData, fAttrScope);
    ID3.makeTree();
    right = 0;
    wrong = 0;
    for (int i = 0; i < testData->size(); i++) {
        if ((*testData)[i].size() > 0 ) {
            if (ID3.tree->decide((*testDataNoCategory)[i]) == (*testData)[i][(*testData)[i].size()-1]) {
                right++;
            }
            else wrong ++;
        }
    }
    resultStream << "ID3: " << right << "/" << testData->size()  << " (" << (float) right / ((float)testData->size() / 100.0) << "\%)"
    << " (size=" << ID3.tree->nodesCount << ")  ";
    
    ID3.tree = EMTI_TreePruner::minimum_error_prune(ID3.tree, trainingData);
    right = 0;
    wrong = 0;
    for (int i = 0; i < testData->size(); i++) {
        if ((*testData)[i].size() > 0 ) {
            if (ID3.tree->decide((*testDataNoCategory)[i]) == (*testData)[i][(*testData)[i].size()-1]) {
                right++;
            }
            else wrong ++;
        }
    }
    resultStream << "Pruned: " << right << "/" << testData->size()  << " (" << (float) right / ((float)testData->size() / 100.0) << "\%)"
    << " (size=" << ID3.tree->nodesCount << ")\n";
    
    EMTI_KNearestClassifier kN(trainingData);
    vector<pair<int,int>> rightwrongks (10, pair<int,int>(0,0));
    for (int i = 0; i < testData->size(); i++) {
        if ((*testData)[i].size() > 0 ) {
            
            for (int k = 1; k <= rightwrongks.size(); ++k) {
                
                int category = kN.classify(&(*testDataNoCategory)[i], /*k*/k+(k-1), /*excludeLowerSimilarityMeasures*/false, /*verbose*/false);
                
                if ((*testData)[i][(*testData)[i].size()-1] == category)
                    rightwrongks[k-1].first++;
                else {
                    rightwrongks[k-1].second++;
                }
            }
        }
    }
    int k = 1;
    bool parapgraph = false;
    for (pair<int,int> rightwrong : rightwrongks) {
        resultStream << k << "-Nearest: " << rightwrong.first << "/" << testData->size()  << " (" << (float) rightwrong.first / ((float)testData->size() / 100.0) << "\%) ";
        if (parapgraph) {
            resultStream << "\n";
        }
        k+=2;
        parapgraph = !parapgraph;
    }
    
    _resultsInfoText = [NSString stringWithUTF8String: resultStream.str().c_str()];
    
    delete fAttrScope;
}

- (void)reset_cycleCount {
    _cyclecount = 0;
    
    
}

- (void) population_delete {
    if (population != NULL) {
        delete population;
        delete trainingData;
        delete testDataNoCategory;
        delete testData;
        
    }
}

- (void)doEvolutionCycle {
    
    //EMTI_DotTreeWrapper wrapper(*population->children, population->cycleCount, "/Users/avividi/Google Drive/masteroppgave/plotdata/trees/forest/");
    
    population->do_evolution_cycle();
    _cyclecount++;
    _treeSizeBest = population->adultBestIndividual->tree->nodesCount;
    _treeSizeAverage = population->averageSize;
    _fitnessBest = population->adultBestIndividual->getFitness();
    _fitnessAverage = population->averageFitness;
    _currentGenerationInfoText = [NSString stringWithUTF8String:population->currentGenerationInfoOutput.c_str()];
    
    
    //TEST
    int right = 0;
    EMTI_TreeCreature* best = population->adultBestIndividual;
    for (int i = 0; i < testData->size(); i++) {
        if ((*testData)[i].size() > 0 ) {
            right += (best->tree->decide((*testDataNoCategory)[i]) == (*testData)[i][(*testData)[i].size()-1]);
        }
    }
    _TESTvalidationAccuracy = (float) right / (float)testData->size();
    //TEST
    if (validationData->size() > 0) {
    right = 0;
    for (int i = 0; i < validationData->size(); i++) {
        if ((*validationData)[i].size() > 0 ) {
            right += (best->tree->decide((*validationData)[i]) == (*validationData)[i][(*validationData)[i].size()-1]);
        }
    }
    _TESTinternalValidationAccuracy = (float) right / (float)validationData->size();
    }
    else {
        _TESTinternalValidationAccuracy = 0;
    }
    _TESTmaxTreeSize = population->maxTreeSize;
    
}

- (void)exportTreeToDot {
    if (population == NULL) return;
    if (population->adultBestIndividual == NULL) return;
    EMTI_DotTreeWrapper wrapper(population->adultBestIndividual->tree, data->attributeValues_intToString);
    
    wrapper.exportToDot(string([[NSBundle mainBundle] bundleURL].fileSystemRepresentation) + "/../treeLarge.dot");
    wrapper.exportToDotSmall(string([[NSBundle mainBundle] bundlePath].fileSystemRepresentation) + "/../treeSmall.dot");
    wrapper.exportToDotTreelike(string([[NSBundle mainBundle] bundlePath].fileSystemRepresentation) + "/../treeVisual.dot");
    
    string outputTrafficTrain =string([[NSBundle mainBundle] bundlePath].fileSystemRepresentation) +  "/../treeTrafficTrain.dot";
    string outputTrafficTest =string([[NSBundle mainBundle] bundlePath].fileSystemRepresentation) +  "/../treeTrafficTest.dot";
    
    population->adultBestIndividual->determineEndNodeAccuraciesAndTreeTraffic(trainingData);
    wrapper.exportToDotTrafficAndAccu(outputTrafficTrain);
    
    population->adultBestIndividual->determineEndNodeAccuraciesAndTreeTraffic(testData);
    EMTI_DotTreeWrapper treewrapTest(population->adultBestIndividual->tree, data->attributeValues_intToString);
    treewrapTest.exportToDotTrafficAndAccu(outputTrafficTest);
}

- (void) conductExperiment {
//    Experiment3 experiment = Experiment3();
//    experiment.run(50);

    
    if (data == NULL || data->dataIntVector->size() == 0) return;
    Experiment1 experiment = Experiment1(data);
    
    experiment.runCrossvalidation(10, 10);
    
    //EMTI_autoRun::autoEMTI(data->dataIntVector, data->getNewFullAttributeScope());
}


@end
