//
//  Front_End.swift
//  EMTI_0.2_swift
//
//  Created by 2016 Anders Nikolai Fuglseth on 09.02.2016.
//

import Cocoa

class Front_End: NSObject {
    
    let EMTI : EMTI_Swift_Wrapper;
    
    var cycleCount:Int32 = 300;
    var populationSize:Int32 = 100;
    var treeSize_fitnessWeight:Int32 = 1000;
    var mutationRate:Float = 1.0;
    var crossoverRate:Float = 1.0;
    var spiceRatio:Float = 0.1;
    var tournamentSize:Int32 = 10;
    var epsilon:Float = 0.8;
    var parentSelection:Int32 = 2;
    var adultSelection:Int32 = 2;
    var rejectWorseGenerations:Int32 = 1;
    var trainingSetSize:Float = 0.8;
    
    @IBOutlet weak var cycleCountField: NSTextField!
    @IBOutlet weak var populationSizeField: NSTextField!
    @IBOutlet weak var treeSize_fitnessWeightField: NSTextField!
    @IBOutlet weak var mutationRateField: NSTextField!
    @IBOutlet weak var crossoverRateField: NSTextField!
    @IBOutlet weak var spiceRatioField: NSTextField!
    @IBOutlet weak var tournamentSizeField: NSTextField!
    @IBOutlet weak var epsilonField: NSTextField!
    
    @IBOutlet weak var parentSelectionPopup: NSPopUpButton!
    @IBOutlet weak var adultSelectionPopup: NSPopUpButton!
    
    @IBOutlet weak var rejectWorseGenerationsRadioBox: NSButton!
    
    @IBOutlet weak var trainingSetSizeField: NSTextField!
    
    @IBOutlet weak var tournamentSizeLabel: NSTextField!
    @IBOutlet weak var epsilonLabel: NSTextField!
    
    @IBOutlet weak var dataPathTextField: NSTextField!
    @IBOutlet weak var dataAttributesTextField: NSTextField!
    @IBOutlet weak var dataInstancesTextField: NSTextField!
    @IBOutlet weak var dataAttributesScrollView: NSScrollView!
    
    @IBOutlet weak var goButton: NSButton!
    @IBOutlet weak var clearGraphsButton: NSButton!
    @IBOutlet weak var clearPopulationButton: NSButton!
    
    @IBOutlet weak var graphViewSize: GraphView!
    @IBOutlet weak var graphViewFitness: GraphView!
    
    @IBOutlet weak var generationInfoOutputField: NSTextField!
    @IBOutlet weak var resultsInfoOutputField: NSTextField!
    
    
    var fitnessGraphFirstVal:Float = -1.0;
    
    var makeNewPopulation:Bool = true;
    
    override init() {
        
        EMTI = EMTI_Swift_Wrapper();
        
        
        super.init();
    }
    
    func UI_setup() {
        self.cycleCountField.integerValue = Int(cycleCount);
        self.populationSizeField.integerValue = Int(populationSize);
        self.treeSize_fitnessWeightField.integerValue = Int(treeSize_fitnessWeight);
        self.mutationRateField.floatValue = Float(mutationRate);
        self.crossoverRateField.floatValue = Float(crossoverRate);
        self.spiceRatioField.floatValue = Float(spiceRatio);
        self.tournamentSizeField.integerValue = Int(tournamentSize);
        self.epsilonField.floatValue = Float(epsilon);
        
        parentSelectionPopup.selectItemAtIndex(Int(parentSelection));
        if (parentSelection != 1) {
            tournamentSizeField.hidden = true;
            epsilonField.hidden = true;
            tournamentSizeLabel.hidden = true;
            epsilonLabel.hidden = true;
        }
        
        adultSelectionPopup.selectItemAtIndex(Int(adultSelection));
        if (adultSelection != 3) {
            spiceRatioField.hidden = true;
        }
        
        self.rejectWorseGenerationsRadioBox.state = Int(rejectWorseGenerations);
        
        self.trainingSetSizeField.floatValue = trainingSetSize;
        
        self.graphViewSize.graphPoints1 = [];
        self.graphViewSize.graphPoints2 = [];
        self.graphViewSize.graphPoints3 = [];
        
        self.graphViewFitness.graphPoints1 = [];
        self.graphViewFitness.graphPoints2 = [];
        self.graphViewFitness.graphPoints3 = [];
        
        fitnessGraphFirstVal = -1.0;
        
    }
    
    @IBAction func parentSelectionPopupAction (pID:AnyObject) {
        if (parentSelectionPopup.indexOfSelectedItem == 1) {
            tournamentSizeField.hidden = false;
            epsilonField.hidden = false;
            tournamentSizeLabel.hidden = false;
            epsilonLabel.hidden = false;
        }
        else {
            tournamentSizeField.hidden = true;
            epsilonField.hidden = true;
            tournamentSizeLabel.hidden = true;
            epsilonLabel.hidden = true;
        }
    }
    
    @IBAction func adultSelectionPopup (pID:AnyObject) {
        if (adultSelectionPopup.indexOfSelectedItem == 3) {
            spiceRatioField.hidden = false;
        }
        else {
            spiceRatioField.hidden = true;
        }
    }
    
    @IBAction func clearGraphs (pID:AnyObject) {
        self.graphViewSize.graphPoints1 = [];
        self.graphViewSize.graphPoints2 = [];
        self.graphViewSize.graphPoints3 = [];
        
        self.graphViewFitness.graphPoints1 = [];
        self.graphViewFitness.graphPoints2 = [];
        self.graphViewFitness.graphPoints3 = [];
        
        fitnessGraphFirstVal = -1.0;
        
        self.graphViewSize.setNeedsDisplayInRect(self.graphViewSize.bounds);
        self.graphViewFitness.setNeedsDisplayInRect(self.graphViewFitness.bounds);
    }
    
    @IBAction func doOpen (pID:AnyObject) {
        let openPanel = NSOpenPanel()
        openPanel.allowsMultipleSelection = false
        openPanel.canChooseDirectories = false
        openPanel.canCreateDirectories = false
        openPanel.canChooseFiles = true
        
        openPanel.allowedFileTypes = ["DATA", "data"];
        
        openPanel.beginWithCompletionHandler { (result) -> Void in
            if result == NSFileHandlingPanelOKButton {
                self.EMTI.data_init(openPanel.URLs[0].fileSystemRepresentation);
                
                self.dataPathTextField.stringValue = openPanel.URLs[0].absoluteString;
                
                self.dataAttributesTextField.stringValue = self.EMTI.dataAttributesCountText;
                self.dataInstancesTextField.stringValue = self.EMTI.dataInstancesCountText;
                (self.dataAttributesScrollView.contentView.documentView as! NSTextView).string = self.EMTI.dataAttributeValuesText;
                
                self.clearPopulation(self);
                
            }
        }
    }
    
    @IBAction func go (pID:AnyObject) {
        
        cycleCount = Int32(self.cycleCountField.integerValue);
        populationSize = Int32(self.populationSizeField.integerValue);
        treeSize_fitnessWeight = Int32(self.treeSize_fitnessWeightField.integerValue);
        mutationRate = self.mutationRateField.floatValue;
        crossoverRate = self.crossoverRateField.floatValue;
        spiceRatio = self.spiceRatioField.floatValue;
        tournamentSize = Int32(self.tournamentSizeField.integerValue);
        epsilon = self.epsilonField.floatValue;
        
        parentSelection = Int32(parentSelectionPopup.indexOfSelectedItem);
        adultSelection = Int32(adultSelectionPopup.indexOfSelectedItem);
        
        rejectWorseGenerations = Int32(self.rejectWorseGenerationsRadioBox.state);
        
        trainingSetSize = self.trainingSetSizeField.floatValue;
        
        
        if (!makeNewPopulation || self.EMTI.populationInit_populationSize(populationSize, sizeWeight: treeSize_fitnessWeight, mutationRate: mutationRate, crossoverRate: crossoverRate, adultSelectionType: adultSelection, parentSelectionType: parentSelection,rejectWorseGeneration: rejectWorseGenerations, spiceRation: spiceRatio, tournamentSize: tournamentSize, epsilonParameter: epsilon, trainingSetSize: trainingSetSize)) {
            
            makeNewPopulation = false;
            
            self.goButton.enabled = false;
            self.clearGraphsButton.enabled = false;
            self.clearPopulationButton.enabled = false;
            
            dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0)) {
                while (self.EMTI.cyclecount < self.cycleCount) {
                    
                    self.EMTI.doEvolutionCycle();
                    
                    dispatch_async(dispatch_get_main_queue()) {
                        
                        
                        self.graphViewSize.graphPoints1.append(Float(self.EMTI.treeSizeBest));
                        self.graphViewSize.graphPoints2.append(Float(self.EMTI.treeSizeAverage));
                        self.graphViewSize.graphPoints3.append(Float(self.EMTI.TESTmaxTreeSize));
                        
                        if (self.fitnessGraphFirstVal == -1) {
                            self.fitnessGraphFirstVal = Float(self.EMTI.fitnessBest);
                        }
                        
                        self.graphViewFitness.color2 = NSColor(red: CGFloat(0xd4)/0xff, green: CGFloat(0x12)/0xff, blue: CGFloat(0x12)/0xff, alpha: 1.0); //dark red
                        self.graphViewFitness.graphPoints1.append(Float(self.EMTI.fitnessBest)-self.fitnessGraphFirstVal+0.05);
                        self.graphViewFitness.graphPoints2.append(Float(self.EMTI.TESTvalidationAccuracy)-self.fitnessGraphFirstVal+0.05);
                        self.graphViewFitness.graphPoints3.append(Float(self.EMTI.TESTinternalValidationAccuracy)-self.fitnessGraphFirstVal+0.05);
                        
                        self.graphViewSize.setNeedsDisplayInRect(self.graphViewSize.bounds);
                        self.graphViewFitness.setNeedsDisplayInRect(self.graphViewFitness.bounds);
                        
                        self.generationInfoOutputField.stringValue = self.EMTI.currentGenerationInfoText;
                    }
                }
                self.EMTI.performanceFeedbackOutput();
                self.resultsInfoOutputField.stringValue = self.EMTI.resultsInfoText;
                self.EMTI.reset_cycleCount();
                
                self.goButton.enabled = true;
                self.clearGraphsButton.enabled = true;
                self.clearPopulationButton.enabled = true;
                
            }
        }
    }
    
    @IBAction func clearPopulation (pID:AnyObject) {
        if (!makeNewPopulation) {
            self.EMTI.reset_cycleCount();
            self.EMTI.population_delete();
            makeNewPopulation = true;
        }
    }
    
    @IBAction func exportTreeToDot (pID:AnyObject) {
        self.EMTI.exportTreeToDot();
    }
    
    @IBAction func runExperiment (pID:AnyObject) {
        self.EMTI.conductExperiment();
    }
}
